//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#include <boost/algorithm/string.hpp>
#include <boost/optional.hpp>
#include <glibmm.h>
#include <raptor.h>
#include <src/util.hh>
#include <src/util-file.hh>
#include <src/uri++.hh>
#include <src/vfs.hh>
#include <src/x_library.hh>

namespace
{
  const char* operator_names[] =
  {
      "fuzzy",
      "exact",
      "greater-than",
      "lesser-than",  
      "greater-than-or-equal",
      "lesser-than-or-equal",
  };

  boost::optional<Glib::ustring> value;
  boost::optional<Glib::ustring> id;
  boost::optional<Glib::ustring> match_type;
  boost::optional<Glib::ustring> value_type;

  Bmp::DB::AttributeV attrs;

#define RQ_SEQ    "http://beep-media-player.org/rq/rq-1#Seq"
#define RQ_QUERY  "http://beep-media-player.org/rq/rq-1#Query"
#define RQ_ATTR   "http://beep-media-player.org/rq/rq-1#Attribute"

#define RQ_TYPE_STRING    "string"
#define RQ_TYPE_INTEGRAL  "integral"

#define RQ_ATTR_PARTICLE_VALUE      "value"
#define RQ_ATTR_PARTICLE_ID         "id"
#define RQ_ATTR_PARTICLE_OPERATOR   "operator"
#define RQ_ATTR_PARTICLE_TYPE       "type"

  void
  add_attribute (void* data, const raptor_statement* triple)
  {
    raptor_identifier_type s = triple->subject_type;
    raptor_identifier_type p = triple->predicate_type;
    raptor_identifier_type o = triple->object_type;

    if (s == 2 && p == 1 && o == 1)
    {
      if (!strcmp ((char*)triple->object, RQ_SEQ)) 
        {
          value.reset();
          id.reset();
          match_type.reset();
          value_type.reset();
        }
    }
    else if (s == 2 && p == 1 && o == 5)
    {
      using namespace boost::algorithm;
      using namespace std; 

      std::string predicate = (const char*)triple->predicate;
      vector<string> subs; 
      split (subs, predicate, is_any_of("#"));

      if (subs[1] == RQ_ATTR_PARTICLE_VALUE)
        {
          value = (char*)triple->object;
        }
      else
      if (subs[1] == RQ_ATTR_PARTICLE_ID) 
        {
          id = (char*)triple->object;
        }
      else
      if (subs[1] == RQ_ATTR_PARTICLE_OPERATOR) 
        {
          match_type = (char*)triple->object;
        }
      else
      if (subs[1] == RQ_ATTR_PARTICLE_TYPE) 
        {
          value_type = (char*)triple->object;
        }
    }
    else if (s == 2 && p == 1 && o == 2)
    {
      if (!strcmp ((char*)triple->predicate, RQ_ATTR) && (match_type && value_type && id && value)) 
        {
          unsigned int n;
          for (n = 0; n < G_N_ELEMENTS(operator_names); ++n)
            {
              if (match_type.get() == operator_names[n]) break;
            }

          if (value_type.get() == RQ_TYPE_STRING)
          {
            attrs.push_back (Bmp::DB::Attribute (Bmp::DB::MatchStyle(n), id.get(), value.get()));
          }
          else
          if (value_type.get() == RQ_TYPE_INTEGRAL) 
          {
            attrs.push_back (Bmp::DB::Attribute (Bmp::DB::MatchStyle(n), id.get(), int(atoi(value.get().c_str()))));
          }
        }
    }
  }
}

namespace Bmp
{
  namespace VFS
  {
      class PluginContainerRLQ
        : public Bmp::VFS::PluginContainerBase
      {
        public:

          virtual bool
          can_process (const Glib::ustring& uri)
          {
            return (Bmp::Util::str_has_suffix_nocase (uri.c_str(), "rlq"));
          }

          virtual bool 
          handle_read	(Bmp::VFS::Handle   &handle,
                       UriList            &list)
          {
              using namespace Bmp::DB;
              using namespace Bmp::Library;

              attrs.clear ();
              raptor_init ();
              raptor_parser *rdf_parser = raptor_new_parser ("rdfxml");
              raptor_uri *uri = raptor_new_uri ((const unsigned char*)handle.get_uri().c_str());
              raptor_set_statement_handler (rdf_parser, NULL, add_attribute); 
              raptor_start_parse (rdf_parser, uri); 
              Glib::ustring buffer ((const char*)handle.get_buffer());
              raptor_parse_chunk (rdf_parser, (const unsigned char*)buffer.c_str(), strlen(buffer.c_str()), 0); 
              raptor_parse_chunk (rdf_parser, NULL, 0, 1); 
              raptor_free_parser (rdf_parser);
              raptor_free_uri (uri);
              raptor_finish ();

              RowV vector;

              if (attrs.empty())
                {
                  std::cerr << "Empty attribute list.";
                  return false;
                }
    

              library->query (attrs, vector);
              
              for (RowV::iterator row = vector.begin (),
                                      end = vector.end   (); row != end; ++row)
              {
                Row& r = (*row);
                Row::iterator rIter = r.find (MetadatumId (DATUM_LOCATION));
                list.push_back (boost::get<Glib::ustring>(rIter->second));
              }

              return true;
          }

          virtual bool
          can_write ()
          {
            return false;
          }

          virtual bool
          handle_write (Bmp::VFS::Handle& handle, const UriList& uri_list)
          {
            return false;
          }

          virtual Bmp::VFS::ExportData
          get_export_data ()
          {
            static Bmp::VFS::ExportData export_data ("RLQ (RDF Library Query)", "rlq"); 
            return export_data;
          }
      };
  }
}
  
extern "C" Bmp::VFS::PluginContainerBase* plugin_create (void)
{
  return new Bmp::VFS::PluginContainerRLQ;
}

extern "C" void plugin_delete (Bmp::VFS::PluginContainerRLQ* p)
{
  delete p;
}
