//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtkmm.h>
#include <glibmm.h>
#include <glibmm/i18n.h>
#include <glib/gstdio.h>

#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <unistd.h>
#include <sys/stat.h>

#include "util.hh"
#include "uri++.hh"
#include "util-file.hh"
#include "vfs.hh"
#include "network.hh"
#include "debug.hh"

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xmlreader.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>

#include <boost/format.hpp>
#include <boost/algorithm/string.hpp>

#include "main.hh"
#include "paths.hh"
#include "mbxml.hh"
#include "amazon.hh"
#include "xml.hh"

#include "x_library.hh"
#include "x_vfs.hh"

namespace
{
  const char* amazon_urls[] =
  {
      "http://images.amazon.com/images/P/%s.01.MZZZZZZZ.jpg",
      "http://images-eu.amazon.com/images/P/%s.01.MZZZZZZZ.jpg",
      "http://images.amazon.co.uk/images/P/%s.01.MZZZZZZZ.jpg",
      "http://images.amazon.jp/images/P/%s.01.MZZZZZZZ.jpg"
  };
}

namespace Bmp
{
  namespace Amazon
  {
   void 
    get_cover_by_asin (Glib::ustring const& asin, Glib::RefPtr<Gdk::Pixbuf>& cover, bool only_cached)
    {
      using namespace Gdk;
      using namespace Glib;
      using namespace Bmp;
      using namespace std;

      string basename = (boost::format ("%s.png") % asin).str ();
      string thumb_path = Glib::build_filename (BMP_PATH_COVER_THUMB_DIR, basename);
      if (Glib::file_test (thumb_path, Glib::FILE_TEST_EXISTS))
        {
          cover = Gdk::Pixbuf::create_from_file (thumb_path); 
          return;
        }

      if (only_cached)
        {
          throw NoCoverError("No cached cover present");
        }

      if (!Network::check_connected ())
        {
          throw NoCoverError("No network connection");
        }

      VFS::Handle handle;
      for (unsigned int n = 0; n < G_N_ELEMENTS(amazon_urls); ++n)
        {
          URI uri ((boost::format (amazon_urls[n]) % asin).str(), true);
          handle = VFS::Handle (ustring(uri));
          vfs->read (handle, VFS::TRANSPORT);
          if (handle.get_buffer()) break;
        }

      if (!handle.get_buffer())
        {
          throw NoCoverError("No cover on amazon found");
        }

      RefPtr<Gdk::PixbufLoader> loader = PixbufLoader::create ("jpeg");
      loader->write (reinterpret_cast<const guint8*>(handle.get_buffer()), handle.get_buffer_size());
      loader->close ();
      loader->get_pixbuf()->save (thumb_path, "png");
      cover = loader->get_pixbuf()->scale_simple ( 128, 128, INTERP_BILINEAR );
      loader.clear();
    }

    void 
    get_cover_artist_album (Glib::ustring const& artist,
			                      Glib::ustring const& album,
                  			    Glib::RefPtr<Gdk::Pixbuf>& cover,
                  			    Glib::ustring& asin,
                  			    bool only_cached) 
    {
      using namespace Glib;
      using namespace Bmp::Library;
      using namespace std;

      MusicBrainzXML::MBReleaseV releases;
      MusicBrainzXML::mb_query_releases (artist, album, releases);

      if (releases.empty())
        {
          throw NoCoverError ("No releases found"); 
        }

      MusicBrainzXML::MBRelease& release = releases[0];
      try {
        if (release.asin)
          {
            get_cover_by_asin (release.asin.get(), cover, only_cached);
            asin = release.asin.get();
          }
          else throw 0;
        }
      catch (...) {}
    }

    void 
    get_cover (Glib::ustring const& _uri, Glib::RefPtr<Gdk::Pixbuf>& cover, bool only_cached) 
    {
      using namespace std;
      using namespace Glib;
      using namespace Bmp::Library;

      Track track;
      library->get_metadata (_uri, track); 

      if (track.asin)
        {
          debug ("amazon", "Getting cover by asin: %s", track.asin.get().c_str());
          get_cover_by_asin (track.asin.get(), cover);
          return;
        }

      if (!(track.artist && track.album))
        {
          debug ("amazon", "No artist & album");
          throw NoCoverError ("No artist and album given");
        }

      if (!Network::check_connected ())
        {
          debug ("amazon", "No Network connection");
          throw NoCoverError ("No Network connection");
        }

      MusicBrainzXML::MBReleaseV releases;
      MusicBrainzXML::mb_query_releases (track.artist.get(), track.album.get(), releases);
      if (releases.empty())
        {
          debug ("amazon", "No releases.");
          throw NoCoverError ("No Matching Releases");
        }

      MusicBrainzXML::MBRelease& release = releases[0];
      if (release.asin)
        {
          string basename = (boost::format ("%s.png") % release.asin.get()).str ();
          string thumb_path = Glib::build_filename (BMP_PATH_COVER_THUMB_DIR, basename);
          debug ("amazon", "release.asin: %s", release.asin.get().c_str());
          get_cover_by_asin (release.asin.get(), cover, only_cached);
          return;
        }
      throw NoCoverError ("Release has no ASIN");
    }
  }
}
