#ifndef BMP_DATABASE_HPP
#define BMP_DATABASE_HPP

#include <glibmm.h>
#include <sqlite3_bmp.h>
#include <queue>

#include <boost/variant.hpp>
#include <boost/tuple/tuple.hpp>

namespace Bmp
{
  Glib::ustring
  sql_uprintf (const char *format, ...);
 
  namespace DB
  {

#include "exception.hh"

    EXCEPTION(DbInitError)
    EXCEPTION(SqlError)

#define FORMAT_ERROR(sqlerr, sqlcode, sql)  ((sql_err_fmt % (sqlerr) % (sqlcode) % (sql)).str ().c_str ())

	  enum ValueType
	  {
	      VALUE_TYPE_BOOL,
	      VALUE_TYPE_INT,
	      VALUE_TYPE_REAL,	//We use "REAL" here as it's sqlite3's notation, see: http://www.sqlite.org/datatype3.html
	      VALUE_TYPE_STRING  
	  }; 

	  typedef boost::variant < bool, int, double, Glib::ustring > Variant;

	  typedef std::map    < std::string,  ValueType >	    ValueMap;
	  typedef std::pair   < std::string,  ValueType >	    PairValueType;
	  typedef std::map    < std::string,  Variant >		    Row;
	  typedef std::vector < Row >			                    RowV;
	  typedef std::map    < std::string,  ValueMap >	    TableMapMap;

    enum MatchStyle
    {
      FUZZY,
      EXACT,
      GREATER_THAN,
      LESSER_THAN,
      GREATER_THAN_OR_EQUAL,
      LESSER_THAN_OR_EQUAL,
      NOT_NULL,
      IS_NULL
    };

	  typedef boost::tuple < MatchStyle, std::string, Variant > Attribute;

	  /** Tuple for performing setting of values
      *
	    * std::string field denotes the column name
	    * Variant field holds the to-be-matched value
	    */
	  typedef boost::tuple < std::string, Variant > AttributeSet;


	  /** A vector of @link Attribute@endlink
	    *
	    */
	  typedef std::vector<Attribute> AttributeV;

	  /** A vector of @link AttributeSet@endlink
	    *
	    */
	  typedef std::vector<AttributeSet> AttributeVSet;

    void
    append_attributes (Glib::ustring &sql, AttributeV const& attributes);

	  /** Slot for notifying about changes in a table
	   *
	   */
	  typedef sigc::slot < void, sqlite_int64 > SlotSQLHook;
	  typedef std::vector < SlotSQLHook > Hooks;
	  typedef std::map < std::string, Hooks > TableHookMap;

	  enum HookType
	  {
	    HOOK_INSERT,
	    HOOK_DELETE,
	    HOOK_UPDATE,

	    N_HOOKS,
	  };

	  typedef boost::tuple  < HookType, sqlite_int64, std::string > QueueHook;
	  typedef std::queue    < QueueHook >				HookQueue;
	
	  class Database
	  {
	    public:

	      Database (std::string name, std::string path, bool truncate = true);
	      ~Database ();

	      void
	      create_table     (std::string const& name,
                          std::string const& pkey,
                          ValueMap    const& map);

	      void
	      create_attribute_view

                        (std::string const& name,
                         std::string const& view_name,
                         std::string const& attribute,
                         ValueMap    const& map);	

	      void
	      create_view_manual  

                        (std::string const& name,
                         std::string const& view_name,
                         std::string const& select,
                         ValueMap    const& map);

	      void
	      drop_table      (std::string const& name);

	      bool
	      table_exists    (std::string const& name);
    
        bool
        attr_exists     (std::string const&   name, 
                         std::string const&   pkey_name,
                         Glib::ustring const& pkey_value);

        void
        remove_by_attr  (std::string const&   name, 
                         std::string const&   pkey_name,
                         Glib::ustring const& pkey_value);

	      void
	      txn_begin	      ();
	      void
	      txn_commit      ();
	      void
	      txn_rollback    ();

	      void 
	      project	        (std::string const&   name,
                         std::string const&   project, 
                         RowV&                vector,
                         AttributeV const& attributes = AttributeV ());

	      void 
	      get	            (std::string const&     name,
                         AttributeV const&   attributes,
                         RowV&                  vector,
                         std::string const&     suffix = std::string()); 

	      void 
	      get_by_rowid    (std::string const&   name,
                         sqlite_int64	        rowid,
                         Row&                 row); 

	      void 
	      set	            (std::string const&       name,
                         std::string const&       key,
                         std::string const&       value,
                         AttributeVSet const&  attributes);

	      void 
	      set_custom      (std::string const&       name,
                         Glib::ustring const&     where_clause,
                         AttributeVSet const&  attributes);

	      void
	      add	            (std::string const& name,
                         Row const&         row);
			       

	      void
	      del	            (std::string const&   name,
			                   AttributeV const& attributes);

	      unsigned int
	      sqlite_exec_simple 
                        (Glib::ustring const& sql); 

        void
        insert_map_for_table
                        (std::string const& name,
                         ValueMap const&    map);

	    private:

	      static void sqlite_trace
                        (void         *data,
                         const char   *sqltext);

	      TableMapMap  mapmap;
	      std::string  name;
	      std::string  path;        //sqlite3 filename = path+name+".mlib";
	      sqlite3	    *sqlite;
	  };
  } 
}; // ns::Bmp

#endif // BMP_DATABASE_HPP
