//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License Version 2
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#include <glibmm.h>
#include <glibmm/i18n.h>
#include <gtkmm.h>
#include <iostream>
#include <config.h>
#include <stdio.h>
#include <dirent.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <libhal.h>
#include <libhal-storage.h>
#include <boost/format.hpp>
#include <boost/algorithm/string.hpp>

#include "x_service_core.hh"

#include "debug.hh"
#include "uri++.hh"
#include "hal.hh"

namespace
{
  inline static int
  strptrcmp (const void *strptr0, const void *strptr1)
  {
    return strcmp (*((const char **) strptr0), *((const char **) strptr1));
  }

  typedef gboolean (* DeviceAddedHandler) (const char *udi, const char *capability);

  const char* HALBus[] =
  {
    "<Unknown>",
    "IDE",
    "SCSI",
    "USB",
    "IEEE1394",
    "CCW"
  };

  void
  print_hal_volume (const Bmp::Library::HAL::Volume& volume)
  {
    using namespace Bmp;

    debug ("hal", "Volume UDI [%s]", volume.volume_udi.c_str());
    debug ("hal", "\tDevice UDI....: [%s]", volume.device_udi.c_str());
    debug ("hal", "\tMount Path....: [%s]", volume.mount_path.c_str());
    debug ("hal", "\tDevice Serial.: [%s]", volume.device_serial.c_str());
    debug ("hal", "\tVolume Name...: [%s]", volume.volume_name.c_str());
    debug ("hal", "\tDevice File...: [%s]", volume.device_file.c_str());
    debug ("hal", "\tDisc..........: [%s]", volume.is_disc ? "TRUE" : "FALSE");
    debug ("hal", "\tBus...........: [%s]", HALBus[volume.device_bus]); 
  }
}

namespace Bmp
{
  namespace Library
  {
      HAL::HAL () : dbus_connection (0), ctx (0)
      {
          message_domain_id = core->message_domain_register ("hal", _("BMP: Volume/Device Management"));
          if (!hal_init())
          {
            core->message_dispatch (message_domain_id,
                                    Gtk::MESSAGE_ERROR,
                                    _("Volume/Device Management could not be initialized; HAL monitoring of"
                                      "volumes in the library, permanent addition of items to the library"
                                      "and other parts of BMP will not be functional.\n\nPlease check "
                                      "your HAL installation and/or contact your distributor."));
            initialized = false;
          }
          else
          {
            initialized = true;
          }
      }

      void
      HAL::cdrom_policy (const char *udi)
      {
        char *device = NULL;
        char *drive_udi = NULL;
        dbus_bool_t has_audio;
        dbus_bool_t has_data;
        dbus_bool_t is_blank;
        DBusError error;

        dbus_error_init (&error);

        has_audio = libhal_device_get_property_bool (ctx, udi, "volume.disc.has_audio", NULL);
        has_data = libhal_device_get_property_bool (ctx, udi, "volume.disc.has_data", NULL);
        is_blank = libhal_device_get_property_bool (ctx, udi, "volume.disc.is_blank", NULL);
        drive_udi = libhal_device_get_property_string (ctx, udi, "info.parent", NULL);
        device = libhal_device_get_property_string (ctx, udi, "block.device", &error);

        if (!device)
        {
          debug ("hal", "cannot get block.device: %s", error.message);
          dbus_error_free (&error);
          goto out;
        }

        if (has_audio && !has_data)
        {
          signal_cdda_inserted_.emit (udi, device);
          debug ("hal", "CD is CDDA and has no data: %s at %s", udi, device);
        }
        else
        if (has_audio && has_data)
        {
          debug ("hal", "CD is CDDA/Mixed-Mode");
        }
        else
        if (has_data)
        {
          debug ("hal", "CD is CDROM");
        }
        else
        if (is_blank)
        {
          debug ("hal", "CD is blank CD-R");
        }

       out:
          libhal_free_string (device);
          libhal_free_string (drive_udi);
      }

      void
      HAL::media_changed (const char *udi, const char *storage_device)
      {
        char *media_type;
        DBusError error;

        /* Refuse to enforce policy on removable media if drive is locked */
        dbus_error_init (&error);

        if (libhal_device_property_exists (ctx, storage_device, "info.locked", NULL) &&
            libhal_device_get_property_bool (ctx, storage_device, "info.locked", NULL))
        {
          debug ("hal", "drive with udi %s is locked through hal; skipping policy", storage_device);
          /* we return TRUE here because the device is locked - we can pretend we handled it */
          return;
        }

        /*
         * Get HAL's interpretation of our media type.  Note that we must check
         * the storage device and not this UDI
         */
        if (!(media_type = libhal_device_get_property_string (ctx, storage_device, "storage.drive_type", &error)))
        {
          debug ("hal", "cannot get storage.drive_type: %s", error.message);
          dbus_error_free (&error);
          return;
        }

        if (!strcmp(media_type, "cdrom"))
        {
          cdrom_policy (udi);
        }

        libhal_free_string (media_type);
      }

      void
      HAL::block_device_added (const char *udi, const char *capability __attribute__((__unused__)))
      {
        using namespace Bmp;

        char *fsusage = 0, *device = 0, *storage_device = 0;
        bool mountable;
        DBusError error;

        dbus_error_init (&error);

        /* is this a mountable volume? */
        if (!(mountable = libhal_device_get_property_bool (ctx, udi, "block.is_volume", NULL)))
        {
          debug ("hal", "not a mountable volume: %s", udi);
          goto out;
        }

        /* if it is a volume, it must have a device node */
        if (!(device = libhal_device_get_property_string (ctx, udi, "block.device", &error)))
        {
          debug ("hal", "cannot get block.device: %s", error.message);
          goto out;
        }

        if (mountable)
        {
          fsusage = libhal_device_get_property_string (ctx, udi, "volume.fsusage", &error);
          if (!fsusage || strcmp (fsusage, "filesystem") != 0)
          {
            debug ("hal", "no mountable filesystem for %s", udi);
            mountable = false;
          }
        }

        /* get the backing storage device */
        if (!(storage_device = libhal_device_get_property_string (ctx, udi, "block.storage_device", &error)))
        {
          debug ("hal", "cannot get block.storage_device: %s", error.message);
          goto out;
        }

        /*
         * Does this device support removable media?  Note that we
         * check storage_device and not our own UDI
         */
        if (libhal_device_get_property_bool (ctx, storage_device, "storage.removable", NULL))
        {
          /* we handle media change events separately */
          debug ("hal", "changed: %s", device);
          media_changed (udi, storage_device);
        }

        out:

        if (dbus_error_is_set (&error))
          {
            dbus_error_free (&error);
          }

        libhal_free_string (device);
        libhal_free_string (fsusage);
        libhal_free_string (storage_device);
      }

      HAL::Volume
      HAL::get_volume_for_uri (Glib::ustring const& uri)
      {
        if (!initialized)
            throw NotInitializedError(); 

        unsigned int length = 0;
        std::string  target_mount_path;
        std::string  filename = Glib::filename_from_uri (uri);
        for (Volumes::const_iterator v_iter = volumes.begin (), vEnd = volumes.end (); v_iter != vEnd; ++v_iter)
        {
            Volume const& volume = v_iter->second;

            unsigned int l_vmp  = volume.mount_path.length (); 
            std::string  substr = filename.substr (0, l_vmp);
      
            debug ("hal", "vmp: '%s', substr: '%s'", volume.mount_path.c_str(), substr.c_str());

            if ((volume.mount_path == substr) && (l_vmp > length))
            {
              target_mount_path = volume.mount_path;
              length = l_vmp;
            } 
	      }

        if (target_mount_path.empty())
            throw NoVolumeForUriError ((boost::format ("No volume for Uri %s") % uri.c_str()).str());

        Volumes::iterator v_iter = volumes.find (target_mount_path); 
        if (v_iter == volumes.end())
            throw NoVolumeForUriError ((boost::format ("No volume for Uri %s") % uri.c_str()).str());

        return v_iter->second;
      }

      std::string
      HAL::get_mount_path_for_volume_udi (std::string const& volume_udi)
      {
        if (!initialized)
            throw NotInitializedError(); 
 
        for (Volumes::const_iterator v_iter = volumes.begin (), vEnd = volumes.end (); v_iter != vEnd; ++v_iter)
            {
              Volume const& volume = v_iter->second;
              if (volume.volume_udi == volume_udi) return volume.mount_path;
            }

        throw NoMountPathForUdiError ((boost::format ("No mount path for given UDI: %s") % volume_udi).str());
      }

      void
      HAL::scan_devices (const char *udi)
      {
        using boost::algorithm::split;
        using boost::algorithm::is_any_of;
        using boost::algorithm::to_lower;
        using boost::algorithm::find_last;
        using boost::iterator_range;

        debug ("hal", "Processing UDI '%s'", udi);  

        std::vector<std::string> path_elements;
        std::string omitroot = udi+1;
        split (path_elements, omitroot, is_any_of("/")); 
        to_lower (path_elements[4]);
        if (!path_elements[4].substr(0, 6).compare ("volume"))
          {
            LibHalVolume    *hal_volume;
            LibHalDrive	    *hal_drive;
            Volume	       volume;

            volume.volume_udi = udi; 
            hal_volume = libhal_volume_from_udi (ctx, udi); 

            if (hal_volume)
              {
                if (!libhal_volume_is_mounted (hal_volume))
                  {
                    libhal_volume_free (hal_volume);
                    return;
                  }

                volume.mount_path     = libhal_volume_get_mount_point (hal_volume);
                volume.is_disc	      = libhal_volume_is_disc (hal_volume);
                volume.volume_name    = (libhal_volume_get_label (hal_volume)) ? libhal_volume_get_label (hal_volume) : "";
                volume.device_file    = (libhal_volume_get_device_file (hal_volume)) ? libhal_volume_get_device_file (hal_volume) : "";
                volume.device_udi     = libhal_volume_get_storage_device_udi (hal_volume);
                hal_drive = libhal_drive_from_udi (ctx, volume.device_udi.c_str());
                volume.device_serial  = (libhal_drive_get_serial (hal_drive)) ? libhal_drive_get_serial (hal_drive) : "";
                volume.device_bus     = libhal_drive_get_bus (hal_drive);

                libhal_volume_free (hal_volume);
                libhal_drive_free  (hal_drive);
              }
            else
              {
                g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, G_STRLOC ": Unable to retrieve information from HAL for volume UDI '%s'", udi);
              }
        
            volumes.insert (std::make_pair (volume.mount_path, volume)); 
            volumes_by_udi.insert (std::make_pair (volume.volume_udi, volume)); 

            print_hal_volume (volume);

            signal_volume_added_.emit (volume); 
          }
  
          iterator_range<std::string::iterator> match = find_last (path_elements[4], "ide"); 

          if ((!path_elements[4].substr(0,7).compare ("storage")) || match)
          { 
            LibHalDrive	*hal_drive = libhal_drive_from_udi (ctx, udi); 
            if (hal_drive)
              { 
                // Check for optical device
                if (libhal_drive_get_type (hal_drive) == LIBHAL_DRIVE_TYPE_CDROM)
                {
                    Device device;

                    device.device_udi	= udi;
                    device.device_name	= libhal_drive_get_model (hal_drive);
                    device.device_file  = libhal_drive_get_device_file (hal_drive);

                    optical.push_back (device);
                }
              }
            libhal_drive_free  (hal_drive);
          }
      }

      bool
      HAL::hal_init ()
      {
        DBusError     error;
        char	    **devices;
        int	      nr;
        int	      n;

        if (!(ctx = libhal_ctx_new ()))
          {
            debug ("hal", G_STRLOC ": Failed to initialize HAL!");
            ctx = 0;
            return false;
          }

        dbus_error_init (&error);
        if (!hal_mainloop_integration (ctx, &error))
          {
            debug ("hal", G_STRLOC ": Mainloop integration failed!");
            dbus_error_free (&error);
            return false;
          }

        libhal_ctx_set_user_data (ctx, this);

#if 0	
        libhal_ctx_set_device_new_capability (ctx, Bmp::Library::device_new_capability);
        libhal_ctx_set_device_lost_capability (ctx, Bmp::Library::device_lost_capability);
        libhal_ctx_set_device_condition (ctx, Bmp::Library::device_condition);
#endif

        libhal_ctx_set_device_added (ctx, Bmp::Library::HAL::device_added);
        libhal_ctx_set_device_removed	(ctx, Bmp::Library::HAL::device_removed);
        libhal_ctx_set_device_property_modified	(ctx, Bmp::Library::HAL::property_modified);

        if (!libhal_device_property_watch_all (ctx, &error))
          {
            debug ("hal", G_STRLOC ": Failed to set up HAL properties watch: %s", error.message);
            dbus_error_free (&error);
            libhal_ctx_free (ctx);
            ctx = 0;
            return false;
          }

        if (!libhal_ctx_init (ctx, &error))
          {
            debug ("hal", G_STRLOC ": hal_ctx_init failed: %s", error.message);
            dbus_error_free (&error);
            libhal_ctx_free (ctx);
            ctx = 0;
            return false;
          }

        if (!(devices = libhal_get_all_devices (ctx, &nr, &error)))
          {
            debug ("hal", G_STRLOC ": get_all_devices failed: %s", error.message);
            dbus_error_free (&error);
            libhal_ctx_shutdown (ctx, NULL);
            libhal_ctx_free (ctx);
            ctx = 0;
            return false;
          }

          for (n = 0; n < nr; n++)
            {
              scan_devices (devices[n]);
            }

        libhal_free_string_array (devices);
        return true;
      }

      dbus_bool_t
      HAL::hal_mainloop_integration (LibHalContext *ctx,
			                               DBusError     *error)
      {
        dbus_connection = dbus_bus_get (DBUS_BUS_SYSTEM, error);

        if (dbus_error_is_set (error))
          {
            debug ("hal", G_STRLOC ": Error conntecting to DBus");
            dbus_connection = 0;
            return false;
          }

        dbus_connection_setup_with_g_main (dbus_connection, NULL);
        libhal_ctx_set_dbus_connection (ctx, dbus_connection);
        return true;
      }

      HAL::~HAL ()
      {
        DBusError error;
        dbus_error_init (&error);

        if (ctx)
          {
            libhal_ctx_shutdown (ctx, &error);
            libhal_ctx_free (ctx);
          }
        if (dbus_connection)
          {
            dbus_connection_close (dbus_connection);
          }
      }

      void
      HAL::device_added	  (LibHalContext *ctx,
			                     const char    *udi)
      {
        using namespace Bmp;

        char    **capabilities;
        size_t    i, n;

        HAL *hal = reinterpret_cast<Bmp::Library::HAL *>(libhal_ctx_get_user_data (ctx)); 

        if (!(capabilities = libhal_device_get_property_strlist (ctx, udi, "info.capabilities", NULL)))
          {
            return;
          }

        for (n = 0; capabilities[n]; n++);

        qsort (capabilities, n, sizeof (char *), strptrcmp);

        for (i = 0; capabilities[i]; ++i)
        {
          if (!strcmp (capabilities[i], "block"))
            {
              hal->block_device_added (udi, "block");
            }
        }

        libhal_free_string_array (capabilities);
      }

      void
      HAL::device_removed (LibHalContext *ctx,
                           const char	   *udi) 
      {
        using namespace Bmp;

        HAL *hal = reinterpret_cast<Bmp::Library::HAL *>(libhal_ctx_get_user_data (ctx)); 
        hal->signal_device_removed_.emit (udi);
        debug ("hal", "device removed: %s", udi);
      }

      void
      HAL::property_modified	
		      (LibHalContext    *ctx,
           const char       *udi,
           const char       *key,
           dbus_bool_t       is_removed,
           dbus_bool_t       is_added)
      {
        using namespace Bmp;
        using namespace Bmp::Library;

        HAL *hal = reinterpret_cast<Bmp::Library::HAL *>(libhal_ctx_get_user_data (ctx)); 
        bool mounted;

        if (strcmp (key, "volume.is_mounted") != 0) return;
        mounted = libhal_device_get_property_bool (ctx, udi, key, NULL);

        if (mounted)
          {
            hal->scan_devices (udi);
          }
        else
          {
            for (Volumes::iterator v_iter = hal->volumes.begin (), vEnd = hal->volumes.end (); v_iter != vEnd; ++v_iter)
              {
                Volume const& volume = v_iter->second;
                if (volume.volume_udi == udi)
                {
                  Volume vol_erased = volume;
                  hal->volumes_by_udi.erase (volume.volume_udi);
                  hal->volumes.erase (v_iter);
                  hal->signal_volume_removed_.emit (vol_erased); 
                  break;
                }
              }
          }
      }

      HAL::SignalVolume&
      HAL::signal_volume_removed ()
      {
        return signal_volume_removed_;
      }

      HAL::SignalVolume&
      HAL::signal_volume_added ()
      {
        return signal_volume_added_;
      }

      HAL::SignalBlockDeviceRemoved&
      HAL::signal_device_removed ()
      {
        return signal_device_removed_;
      }

      HAL::SignalCDDAInserted&
      HAL::signal_cdda_inserted ()
      {
        return signal_cdda_inserted_;
      }

  }
}
