//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License Version 2 
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glibmm.h>

// Musicbrainz
#include <musicbrainz/musicbrainz.h>

#include <glib/gi18n.h>
#include <glib/gstdio.h>
#include <gtkmm.h>

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <iostream>
#include <sstream>

#include <boost/shared_ptr.hpp>
#include <boost/format.hpp>
#include <boost/optional.hpp>
#include <boost/algorithm/string.hpp>

// BMP
#include "uri++.hh"
#include "database.hh"
#include "library.hh"
#include "util-file.hh"
#include <sqlite3_bmp.h>

// TagLib
#include <taglib_plugins/tplugins.h>
#ifdef HAVE_MP4V2
#  include <taglib_plugins/mp4/mp4file.h>
#  include <taglib_plugins/mp4/mp4tag.h>
#endif

#include <taglib_plugins/wma/wmafile.h>
#include <taglib_plugins/wma/wmatag.h>

#include <taglib/taglib.h>
#include <taglib/fileref.h>
#include <taglib/audioproperties.h>
#include <taglib/tfile.h>
#include <taglib/tag.h>
#include <taglib/id3v2tag.h>
#include <taglib/mpegfile.h>
#include <taglib/oggfile.h>
#include <taglib/vorbisfile.h>
#include <taglib/vorbisproperties.h>
#include <taglib/xiphcomment.h>
#include <taglib/mpcfile.h>
#include <taglib/vorbisfile.h>
#include <taglib/flacfile.h>
#include <taglib/xiphcomment.h>
#include <taglib/id3v2framefactory.h>
#include <taglib/textidentificationframe.h>
#include <taglib/uniquefileidentifierframe.h>
#include <taglib/apetag.h>
#include <taglib/apeitem.h>
#include <taglib/apefooter.h>

// GStreamer
#include <gst/gst.h>
#include <gst/gstelement.h>
#include "audio.hh"

#include "main.hh"
#include "paths.hh"
#include "debug.hh"

#include "x_library.hh"

// HAL
#ifdef HAVE_HAL
#  include "x_hal.hh"
#endif

#define TAG_READ_TUPLE(_data) ((TagReadTuple*)(_data))

#define MLIB_VERSION "32"

namespace Bmp
{
  namespace Library
  {
    namespace
    {
      static boost::format fsizefmt ("%llu");

      struct TagReadTuple
      {
          DB::Row *row;
          bool     eos;
      };

      // Attributes for the Albums meta table 
      DatumDefine album_attributes[] = 
      {
          {
              N_("Count"),
              N_("Counts"),
              N_("Track Count"),
              "count",
              DB::VALUE_TYPE_INT,
          },
      };
 
      // Attributes for the Tags meta table
      DatumDefine tags_attributes[] = 
      {
          {
              N_("ID"),
              N_("IDs"),
              N_("Tag Id"),
              "id",
              DB::VALUE_TYPE_INT,
          },

          {
              N_("Description"),
              N_("Description"),
              N_("Some descriptive text for the tag"),
              "desc",
              DB::VALUE_TYPE_STRING,
          },

          {
              N_("Table Name"),
              N_("Table Names"),
              N_("The table name of the tag"),
              "name", 
              DB::VALUE_TYPE_STRING,
          },
      };

      // Attributes for a Tag row
      DatumDefine tag_attributes[] = 
      {
          {
              ("ID"),
              ("IDs"),
              N_("Attribute Id"),
              "id",
              DB::VALUE_TYPE_INT,
          },

          {
              ("AttrName"),
              ("AttrNames"),
              N_("Attribute Type"),
              "name", //NOTE: This is identical to an attribute name from track_attributes
              DB::VALUE_TYPE_STRING,
          },

          {
              ("AttrType"),
              ("AttrType"),
              N_("Attribute Type"),
              "type",
              DB::VALUE_TYPE_INT,
          },

          {
              ("AttrPrecision"),
              ("AttrPrecision"),
              N_("Attribute Precision"),
              "precise",
              DB::VALUE_TYPE_BOOL,
          },

          {
              ("AttrVBool"),
              ("AttrVBool"),
              N_("Attribute Value Bool"),
              "v_bool",
              DB::VALUE_TYPE_BOOL,
          },

          {
              ("AttrVInteger"),
              ("AttrVInteger"),
              N_("Attribute Value Integer"),
              "v_integer",
              DB::VALUE_TYPE_INT,
          },

          {
              ("AttrVReal"),
              ("AttrVReal"),
              N_("Attribute Value Real"),
              "v_real",
              DB::VALUE_TYPE_REAL,
          },

          {
              ("AttrVString"),
              ("AttrVString"),
              N_("Attribute Value String"),
              "v_string",
              DB::VALUE_TYPE_STRING,
          },

      };

      Datum cdda_data[] =
      {
        DATUM_LOCATION,
        DATUM_ARTIST,
        DATUM_ALBUM,
        DATUM_TITLE,
        DATUM_TRACK,
        DATUM_TIME,
        DATUM_GENRE,
        DATUM_COMMENT,
        DATUM_RATING,
        DATUM_DATE,
        DATUM_COUNT,
        DATUM_PLAYDATE,
        DATUM_ASIN,
        DATUM_MB_ALBUM_ARTIST,
        DATUM_MB_ALBUM_ARTIST_ID,
        DATUM_MB_ALBUM_ARTIST_SORTNAME,
        DATUM_MB_TRACK_ID,
        DATUM_MB_ALBUM_ID,
        DATUM_MB_ARTIST_ID,
        DATUM_MB_ARTIST_SORTNAME,
        DATUM_MB_RELEASE_DATE,
      };
        
      // Attributes for a Track row
      DatumDefine track_attributes[] =
      {
          {
              N_("Location"),
              N_("Locations"),
              N_("The location of the track (file, internet stream, etc.)"),
              "location",
              DB::VALUE_TYPE_STRING,
          },

          {
              N_("Artist"),
              N_("Artists"),
              N_("The artist/peformer of the track"),
              "artist",
              DB::VALUE_TYPE_STRING,
          },

          {
              N_("Album"),
              N_("Albums"),
              N_("The album of the track"),
              "album",
              DB::VALUE_TYPE_STRING,
          },

          {
              N_("Title"),
              N_("Titles"),
              N_("The title of the current track"),
              "title",
              DB::VALUE_TYPE_STRING,
          },

          {
              N_("Track"),
              N_("Tracks"),
              N_("The track number of the track"),
              "tracknumber",
              DB::VALUE_TYPE_INT,
          },

          {
              N_("Time"),
              N_("Times"),
              N_("The length of the track"),
              "time",
              DB::VALUE_TYPE_INT,
          },

          {
              N_("Genre"),
              N_("Genres"),
              N_("The genre (kind of music) of the track"),
              "genre",
              DB::VALUE_TYPE_STRING,
          },

          {
              N_("Comment"),
              N_("Comments"),
              N_("Comments added to the track"),
              "comment",
              DB::VALUE_TYPE_STRING,
          },

          {
              N_("Rating"),
              N_("Ratings"),
              N_("The rating of the track"),
              "rating",
              DB::VALUE_TYPE_INT,
          },

          {
              N_("Date"),
              N_("Dates"),
              N_("The date (year) of the track"),
              "date",
              DB::VALUE_TYPE_INT,
          },

          {
              N_("MTIME"),
              N_("MTIMEs"),
              N_("The modification time of the track"),
              "mtime",
              DB::VALUE_TYPE_INT,
          },

          {
              N_("Bitrate"),
              N_("Bitrates"),
              N_("The bitrate of the track"),
              "bitrate",
              DB::VALUE_TYPE_INT,
          },

          {
              N_("Samplerate"),
              N_("Samplerates"),
              N_("The samplerate of the track"),
              "samplerate",
              DB::VALUE_TYPE_INT,
          },

          {
              N_("Play count"),
              N_("Play counts"),
              N_("The number of times this track was played"),
              "count",
              DB::VALUE_TYPE_INT,
          },

          {
              N_("Last played date"),
              N_("Last played dates"),
              N_("The last date of when this track was played"),
              "playdate",
              DB::VALUE_TYPE_INT,
          },

          {
              ("AMAZON_ASIN"),
              ("ASINs"),
              N_("Amazon Standard Index Number"),
              "asin",
              DB::VALUE_TYPE_STRING,
          },

#ifdef HAVE_HAL
          {
              ("Volume UDI"),
              ("Volume UDIs"),
              N_("HAL Volume UDI this track resides on"),
              "volume_udi",
              DB::VALUE_TYPE_STRING,
          },

          {
              ("Device UDI"),
              ("Device UDIs"),
              N_("HAL Device UDI this track resides on"),
              "device_udi",
              DB::VALUE_TYPE_STRING,
          },

          {
              N_("Volume Relative Path"),
              N_("Volume Relative Paths"),
              N_("Path of the track relative to it's volume's root"),
              "volume_relative_path",
              DB::VALUE_TYPE_STRING,
          },

          {
              N_("Volume Relative Base Dir"),
              N_("Volume Relative Base Dir"),
              N_("The base dir of the file relative to the volume's root"),
              "volume_relative_base_path",
              DB::VALUE_TYPE_STRING,
          },
#endif

          {
              ("MusicBrainz Album Artist"),
              ("MusicBrainz Album Artists"),
              N_("The MusicBrainz Release Artist"),
              "mb_albumartist",
              DB::VALUE_TYPE_STRING,
          },

          {
              ("MusicBrainz Album Artist Id"),
              ("MusicBrainz Album Artist Ids"),
              N_("The MusicBrainz Release Artist ID (The ID of the Artist this release is accredited to)"),
              "mb_albumartist_id",
              DB::VALUE_TYPE_STRING,
          },

          {
              ("MusicBrainz Album Artist Sortname"),
              ("MusicBrainz Album Artist Sortnames"),
              N_("The MusicBrainz Release Sort Name"),
              "mb_albumartistsortname",
              DB::VALUE_TYPE_STRING,
          },

          {
              ("MusicBrainz Track ID"),
              ("MusicBrainz Track IDs"),
              N_("The MusicBrainz Track ID"),
              "mb_track_id",
              DB::VALUE_TYPE_STRING,
          },

          {
              ("MusicBrainz Album Id"),
              ("MusicBrainz Album Ids"),
              N_("The MusicBrainz Release ID"),
              "mb_album_id",
              DB::VALUE_TYPE_STRING,
          },

          {
              ("MusicBrainz Artist Id"),
              ("MusicBrainz Artist Ids"),
              N_("The MusicBrainz Artist ID"),
              "mb_artist_id",
              DB::VALUE_TYPE_STRING,
          },

          {
              ("MusicBrainz Artist Sortname"),
              ("MusicBrainz Artist Sortnames"),
              N_("The MusicBrainz Artist Sort Name"),
              "mb_artistsortname",
              DB::VALUE_TYPE_STRING,
          },

          {
              ("MusicBrainz Release Date"),
              ("MusicBrainz Release Dates"),
              N_("The MusicBrainz Artist Sort Name"),
              "mb_release_date",
              DB::VALUE_TYPE_STRING,
          },

          {
              ("MusicIP PUID"),
              ("MusicIP PUIDs"),
              N_("PUID For the given file"),
              "puid",
              DB::VALUE_TYPE_STRING,
          },

          {
              N_("New Item"),
              N_("New Items"),
              N_("Whether this item was just addded and/or yet not approved"),
              "new_item",
              DB::VALUE_TYPE_BOOL,
          },

          {
              N_("THASH"),
              N_("THASHes"),
              N_("Binary Data Tag Hash"),
              "hash",
              DB::VALUE_TYPE_STRING,
          },

      };

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

      void
      map_insert (Bmp::DB::ValueMap& map, Bmp::Library::Datum datum)
      {
        map.insert (make_pair (string (track_attributes[datum].id),
                                       track_attributes[datum].type));
      }

      void
      row_set       (Bmp::DB::Row&          row,
                     Bmp::Library::Datum    datum,
                     Bmp::DB::Variant       variant)
      {
        row.insert (std::make_pair (track_attributes[datum].id, variant));
      }

      void
      row_tag_set   (Bmp::DB::Row&          row,
                     Bmp::Library::DatumTag datum,
                     Bmp::DB::Variant       variant)
      {
        row.insert (std::make_pair (tag_attributes[datum].id, variant));
      }


      TagLib::ID3v2::UserTextIdentificationFrame*
      find_utif (TagLib::ID3v2::Tag *tag, const TagLib::String &description)
      {
          TagLib::ID3v2::FrameList l = tag->frameList("TXXX");
          for(TagLib::ID3v2::FrameList::Iterator it = l.begin(); it != l.end(); ++it) {
            TagLib::ID3v2::UserTextIdentificationFrame *f = dynamic_cast<TagLib::ID3v2::UserTextIdentificationFrame *>(*it);
            if(f && f->description() == description)
              return f;
          }
          return 0;
      }

      void get_id3v2 (TagLib::ID3v2::Tag *tag, DB::Row &row, struct stat *tagstat)
      {
          using namespace Glib;
          using namespace TagLib;
          using namespace boost;
          using namespace std;

          using boost::algorithm::split;
          using boost::algorithm::find_nth;
          using boost::iterator_range;

          const Datum extra_tags_mpeg[] =
              {
                  DATUM_MB_ALBUM_ARTIST,
                  DATUM_MB_ALBUM_ARTIST_ID,
                  DATUM_MB_ALBUM_ID,
                  DATUM_MB_ARTIST_ID,
                  DATUM_MB_ARTIST_SORTNAME,
                  DATUM_MB_ALBUM_ARTIST_SORTNAME,
                  DATUM_ASIN,
                  DATUM_MUSICIP_PUID,
              };

          ID3v2::UserTextIdentificationFrame *frame = 0;
          for (unsigned int n = 0; n < G_N_ELEMENTS(extra_tags_mpeg); ++n)
          {
              frame = find_utif (tag, String (track_attributes[extra_tags_mpeg[n]].id, String::UTF8));
              if (frame)
              {
                  string str = frame->toString().to8Bit(true);

                  iterator_range<std::string::iterator> match = find_nth (str, "] ", 0);
                  if (match.empty()) continue;

                  string substr (match.end(), str.end());
                  if (!substr.length()) continue;

                  row_set (row, extra_tags_mpeg[n], ustring (substr));
              }
          }

          // Check for UFID
          using TagLib::ID3v2::FrameList;
          FrameList const& map = tag->frameListMap()["UFID"];
          if (!map.isEmpty())
          {
              for (FrameList::ConstIterator iter = map.begin(), end = map.end (); iter != end; ++iter)
              {
                  ID3v2::UniqueFileIdentifierFrame *ufid = reinterpret_cast<ID3v2::UniqueFileIdentifierFrame*>(*iter);
                  if (ufid->owner() == "http://musicbrainz.org")
                  {
                      ByteVector vec = ufid->identifier();
                      vec.append ('\0');
                      row_set (row, DATUM_MB_TRACK_ID, ustring (vec.data()));
                      break;
                  }
              }

          // Check for date
          using TagLib::ID3v2::FrameList;
          const char *_frames[] = { "TDRL", "TDOR" , "XDOR" };
          for (unsigned int n = 0; n < G_N_ELEMENTS (_frames); ++n)
          {
            FrameList const& map = tag->frameListMap()[_frames[n]];
            if (!map.isEmpty())
                {
                  std::string str = map.front()->toString().to8Bit();
                  row_set (row, DATUM_MB_RELEASE_DATE, ustring(str)); 
                }
              }
          }

          // Check for sortname
          if (row.find (track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].id) == row.end())
          {
            using TagLib::ID3v2::FrameList;
            const char *id3v2frame = 0;
            if (!tag->frameListMap()["XSOP"].isEmpty()) id3v2frame = "XSOP";
            else if (!tag->frameListMap()["TSOP"].isEmpty()) id3v2frame = "TSOP";

            if (id3v2frame)
            {
              FrameList const& map = tag->frameListMap()[id3v2frame];
              if (!map.isEmpty())
              {
                  std::string str = map.front()->toString().to8Bit();
                  row_set (row, DATUM_MB_ALBUM_ARTIST_SORTNAME, ustring(str)); 
              }
            }
          }

          TagLib::ByteVector tagData = tag->render();
          const char* blob = tagData.data(); 
          std::string md5_uid = Util::md5_hex ((char*)blob, tagData.size());
          std::string hash = md5_uid+(fsizefmt % (unsigned long long int)(tagstat->st_size)).str();
          row_set (row, DATUM_HASH, ustring(hash));
      }

      // Based on SoundJuicer code, (C) Ross Burton <ross@burtonini.com> 
      void
      add_txxx_tag (TagLib::ID3v2::Tag    *id3v2tag,
                    const Glib::ustring&   spec_id,
                    const Glib::ustring&   realworld_id,
                    const Glib::ustring&   id_str) 
      {
        using namespace Bmp;
        using namespace TagLib;

        ID3v2::UserTextIdentificationFrame * frame;

        if (id_str.empty()) return;

        if (!spec_id.empty())
        {
          frame = new ID3v2::UserTextIdentificationFrame (String::UTF8);
          id3v2tag->addFrame (frame);
          frame->setDescription (String (spec_id.c_str(), String::UTF8));
          frame->setText (String (id_str.c_str(), String::UTF8));
        }

        if (!realworld_id.empty())
        {
          frame = new ID3v2::UserTextIdentificationFrame (String::UTF8);
          id3v2tag->addFrame (frame);
          frame->setDescription (String (realworld_id.c_str(), String::UTF8));
          frame->setText (String (id_str.c_str(), String::UTF8));;
        }
      }

      int
      get_duration_from_sectors (int sectors)
      {
          const int bytes_per_sector   = 2352;
          const int bytes_per_seconds  = (44100 / 8) / 16 / 2;

          return (sectors * bytes_per_sector / bytes_per_seconds)/1000;
      }

      void
      tag_foreach (const GstTagList *list,
                   const gchar      *tag,
                   TagReadTuple     *tuple)
      {
          int count = gst_tag_list_get_tag_size (list, tag);

          for (int i = 0; i < count; i++)
            {
              const GValue *value = gst_tag_list_get_value_index (list, tag, i);

              if (G_VALUE_HOLDS (value, G_TYPE_INT))
                {
                  tuple->row->insert (std::make_pair (std::string (tag), int (g_value_get_int(value))));
                }
              else
              if (G_VALUE_HOLDS (value, G_TYPE_UINT))
                {
                  tuple->row->insert (std::make_pair (std::string (tag), int (g_value_get_uint(value))));
                }
              else
              if (G_VALUE_HOLDS (value, G_TYPE_STRING))
                {
                  const char *_v_C = g_value_get_string (value);
                  if (*_v_C)
                    {
                      std::string v = _v_C; 
                      boost::trim (v);
                      tuple->row->insert (std::make_pair (std::string (tag), v)); 
                    }
                }
            }
      }

      gboolean
      metadata_bus_handler (GstBus       *bus,
                            GstMessage   *message,
                            TagReadTuple *tuple)
      {
          switch (GST_MESSAGE_TYPE (message))
          {
            case GST_MESSAGE_EOS:
                {
                  tuple->eos = true;
                  break;
                }

            case GST_MESSAGE_TAG:
                {
                  GstTagList *tag_list = 0;
                  gst_message_parse_tag (message, &tag_list);
                  if (tag_list)
                    {
                      gst_tag_list_foreach (tag_list, (GstTagForeachFunc) tag_foreach, tuple);
                      gst_tag_list_free (tag_list);
                    }
                  break;
                }

            case GST_MESSAGE_ERROR:
                {
                  GError *error = 0;
                  gst_message_parse_error (message, &error, 0);
                  g_print ("%s ERROR: %s\n", G_STRLOC, error->message);
                  g_error_free (error);
                  tuple->eos = true;
                  break;
                }

            default: break;
          }
          return FALSE;
      }

      void
      metadata_event_loop (TagReadTuple *tuple,
                           GstElement   *element,
                           gboolean      block)
      {
          GstBus *bus = gst_element_get_bus (element);
          g_return_if_fail (bus != 0);

          bool done = false;
          while (!done && !tuple->eos)
          {
              GstMessage *message;

              if (block)
                  message = gst_bus_poll (bus, GST_MESSAGE_ANY, -1);
              else
                  message = gst_bus_pop  (bus);

              if (message == 0)
              {
                  gst_object_unref (bus);
                  return;
              }

              done = metadata_bus_handler (bus, message, tuple);
              gst_message_unref (message);
          }
          gst_object_unref (bus);
      }

#if 0
      void
      get_audiocd_metadata (MB_AudioCD_Status &status, DB::RowV &vector)
      {
          using namespace Glib;
          using namespace std;

          // Create the musicbrainz object, which will be needed for subsequent calls
          MusicBrainz o;
          o.UseUTF8 (true);
          o.SetDevice (mcs->key_get<string>("audio", "cdrom-device"));

          bool isMultipleArtist = false;
          int trackNum = 0;

          int ret;
          ret = o.Query (MBQ_GetCDTOC);
          if (!ret)
          {
              string error;
              o.GetQueryError (error);
              debug ("library", "%s: Query failed: %s", G_STRLOC, error.c_str());
              status = Bmp::Library::MB_FAILURE;
              return;
          }
          debug ("library", "AudioCD: Got TOC");


          string cdindex = o.Data(string (MBE_TOCGetCDIndexId));
          if (cdindex.empty())
          {
              string error;
              o.GetQueryError (error);
              debug ("library", "%s: Query failed: %s", G_STRLOC, error.c_str());
              status = Bmp::Library::MB_FAILURE;
              return;
          }
          debug ("library", "AudioCD: Got cdindexid");

          ret = o.Query(string (MBQ_GetCDInfo));
          if (!ret)
          {
              string error;
              o.GetQueryError (error);
              debug ("library", "%s: Query failed: %s", G_STRLOC, error.c_str());
              status = Bmp::Library::MB_OFFLINE;
              return;
          }
          debug ("library", "AudioCD: Got CDInfo");

          // Select the first album
          list<int> argList;
          argList.push_back (1);
          if (!o.Select (string (MBS_SelectAlbum), &argList))
          {
              status = Bmp::Library::MB_OFFLINE;
              return;
          }
          debug ("library", "AudioCD: Selected First Album");

          string album_id_url = o.Data(string (MBE_AlbumGetAlbumId));
          if (album_id_url.empty ())
            {
              debug ("library", "No AlbumIDURL");
              return;
            }

          debug ("library", "AudioCD: Got Album ID URL");

          string album_id;
          o.GetIDFromURL (album_id_url, album_id);

          // Extract the album name
          string album_name = o.Data(string (MBE_AlbumGetAlbumName));

          // Extract the number of tracks
          int n_tracks = o.DataInt(string (MBE_AlbumGetNumTracks));

          if (n_tracks < 1)
          {
            status = Bmp::Library::MB_OFFLINE;
            return;
          }
          debug ("library", "More than 1 track, proceeding to get online data");

          // Check to see if there is more than one artist for this album
          string artist_check;
          for (int i = 1; i <= n_tracks; i++)
          {
              string artist = o.Data (string (MBE_AlbumGetArtistId), i);
              if (artist.empty ())
                {
                  debug ("library", "No Artist");
                  return;
                }

              if (i == 1)
              {
                artist_check = artist;
                continue;
              }

              if (artist.compare (artist_check))
              {
                isMultipleArtist = true;
                break;
              }
          }

          string artist_name, artist_id;

          if (!isMultipleArtist)
            {
              // Extract the artist name from the album
              artist_name = o.Data (string (MBE_AlbumGetArtistName), 1);

              // Extract the artist id from the album
              artist_id   = o.Data (string (MBE_AlbumGetArtistId), 1);
            }

          for(int i = 1; i <= n_tracks; i++)
            {
              DB::Row row;

              row_set (row, DATUM_LOCATION,   (boost::format ("cdda://%s/%d") % cdindex % i).str ());
              row_set (row, DATUM_ALBUM,      ustring (album_name));
              row_set (row, DATUM_BITRATE,    int(1411));
              row_set (row, DATUM_SAMPLERATE, int(441000));

              // Extract the track name from the album.
              string title = o.Data (string (MBE_AlbumGetTrackName), i);
              if (!title.empty ())
                {
                  row_set (row, DATUM_TITLE, ustring(title));
                }

              // Extract the album id from the track. Just use the
              // first album that this track appears on
              string track_id = o.Data (string (MBE_AlbumGetTrackId), i);
              if (!track_id.empty ())
                {
                  // Extract the track number
                  trackNum = o.GetOrdinalFromList (string (MBE_AlbumGetTrackList), track_id);
                  if (trackNum > 0 && trackNum < 100)
                    {
                      row_set (row, DATUM_TRACK, int(trackNum)); 
                    }
                }

              // If its a multple artist album, print out the artist for each track
              if (isMultipleArtist)
                {
                  // Extract the artist name from this track
                  string individual_artist = o.Data(string (MBE_AlbumGetArtistName), i);
                  if (!individual_artist.empty ())
                    {
                      row_set (row, DATUM_ARTIST, ustring(individual_artist)); 
                    }
                }
              else
                {
                  row_set (row, DATUM_ARTIST, ustring(artist_name)); 
                }
              vector.push_back (row);
            }

          ret = o.Query(MBQ_GetCDTOC);
          if (!ret)
            {
              string error;
              o.GetQueryError (error);
              debug ("library", "%s: Query failed: %s", G_STRLOC, error.c_str());
              status = Bmp::Library::MB_FAILURE;
              return;
            }
          debug ("library", "AudioCD: Got TOC again");

          for(int i = 1; i <= n_tracks; i++)
            {
              DB::Row& row = vector[i-1];
              int duration = get_duration_from_sectors(o.DataInt(MBE_TOCGetTrackNumSectors, i+1));
              row_set (row, DATUM_TIME, duration); 
            } 

          status = Bmp::Library::MB_OK;
          debug ("library", "AudioCD: Everything allright, status is MB_OK");
      }
#endif

    } // !<unnamed>::

    std::string Library::metadatum_id (Datum datum) 
    {
      return track_attributes[datum].id;
    }

    void
    Library::metadata_set_common (const TrackUpdated& track,
                                         TagLib::Tag* tag)
    {
      using namespace TagLib;
      if (!tag) return;

      if (track.title)
          tag->setTitle   (String (track.title.get().c_str(), String::UTF8)); 

      if (track.artist)
          tag->setArtist  (String (track.artist.get().c_str(), String::UTF8));

      if (track.album)
          tag->setAlbum   (String (track.album.get().c_str(), String::UTF8));

      if (track.comment)
          tag->setComment (String (track.comment.get().c_str(), String::UTF8));

      if (track.genre)
          tag->setGenre   (String (track.genre.get().c_str(), String::UTF8));

      if (track.date)
          tag->setYear    (track.date.get());

      if (track.tracknumber)
          tag->setTrack   (track.tracknumber.get());
    }

    void 
    Library::metadata_set_taglib (const Glib::ustring&  location,
                                  TrackUpdated&          track)
    {
        using namespace std;
        using namespace Glib;
        using namespace TagLib;

        string filename = Glib::filename_from_uri (location); 
        string type;

        struct stat tagstat;

        if (!Audio::typefind (filename, type))
            throw MetadataWriteError ((boost::format ("Unable to create a TagLib FileRef for this file: %s") % filename).str());

        if ((type == "application/x-id3") || (type == "audio/mpeg"))
          {
            ID3v2::FrameFactory *factory = ID3v2::FrameFactory::instance(); 
            factory->setDefaultTextEncoding (String::UTF8);
            MPEG::File mpegfile (filename.c_str(), factory);

            if (!mpegfile.isOpen() && mpegfile.isValid())
                throw MetadataWriteError ((boost::format ("Unable to get an MPEG::File from filename: %s") % filename).str());

            debug ("library", "Modifying MPEG::File: %s", filename.c_str());
            ID3v2::Tag *tag = mpegfile.ID3v2Tag (true);

            if (tag)
              {
                metadata_set_common (track, tag);
                if (track.mb_album_artist)
                  {
                    add_txxx_tag (tag,
                                  track_attributes[DATUM_MB_ALBUM_ARTIST].title,
                                  track_attributes[DATUM_MB_ALBUM_ARTIST].id,
                                  track.mb_album_artist.get());
                  }

                if (track.mb_album_artist_id)
                  {
                    add_txxx_tag (tag,
                                  track_attributes[DATUM_MB_ALBUM_ARTIST_ID].title,
                                  track_attributes[DATUM_MB_ALBUM_ARTIST_ID].id,
                                  track.mb_album_artist_id.get());
                  }

                if (track.mb_album_artist_sort_name)
                  {
                    add_txxx_tag (tag,
                                  track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].title,
                                  track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].id,
                                  track.mb_album_artist_sort_name.get());
                  }

                if (track.mb_track_id)
                  {
                    ID3v2::UniqueFileIdentifierFrame *frame = new ID3v2::UniqueFileIdentifierFrame
                                                                ("http://musicbrainz.org", track.mb_track_id.get().c_str());
                    tag->addFrame (frame);
                  }

                if (track.mb_album_id)
                  {
                    add_txxx_tag (tag,
                                  track_attributes[DATUM_MB_ALBUM_ID].title,
                                  track_attributes[DATUM_MB_ALBUM_ID].id,
                                  track.mb_album_id.get());
                  }

                if (track.mb_artist_id)
                  {
                    add_txxx_tag (tag,
                                  track_attributes[DATUM_MB_ARTIST_ID].title,
                                  track_attributes[DATUM_MB_ARTIST_ID].id,
                                  track.mb_artist_id.get());
                  }

                if (track.mb_artist_sort_name)
                  {
                    add_txxx_tag (tag,
                                  track_attributes[DATUM_MB_ARTIST_SORTNAME].title,
                                  track_attributes[DATUM_MB_ARTIST_SORTNAME].id,
                                  track.mb_artist_sort_name.get());
                  }

                if (track.asin)
                  {
                    add_txxx_tag (tag,
                                track_attributes[DATUM_ASIN].title,
                                track_attributes[DATUM_ASIN].id,
                                track.asin.get());
                  }

                if (track.puid)
                  {
                    add_txxx_tag (tag,
                                track_attributes[DATUM_MUSICIP_PUID].title,
                                track_attributes[DATUM_MUSICIP_PUID].id,
                                track.puid.get());
                  }
              }
            else
              {
                throw MetadataWriteError ((boost::format ("Unable to write metadata for file %s: No Tag") % filename).str());
              }

            mpegfile.save (TagLib::MPEG::File::ID3v1 | TagLib::MPEG::File::ID3v2);
            tag = mpegfile.ID3v2Tag (false);
            stat (filename.c_str(), &tagstat);
            TagLib::ByteVector tagData = tag->render();
            const char* blob = tagData.data();
            std::string md5_uid = Util::md5_hex ((char*)blob, tagData.size());
            track.hash = md5_uid+(fsizefmt % (unsigned long long int)(tagstat.st_size)).str();

            return;
          }
        else 
        if (type == "application/ogg")
          {
            Ogg::Vorbis::File oggfile (filename.c_str());
            if (!oggfile.isOpen() && oggfile.isValid())
                throw MetadataWriteError ((boost::format ("Unable to get an Ogg::Vorbis::File from filename: %s") % filename).str());

            debug ("library", "Modifying OGG::File: %s", filename.c_str());
            metadata_set_common (track, oggfile.tag());
            TagLib::Ogg::XiphComment *comment = oggfile.tag(); 
            if (comment)
              {
                if (track.mb_album_artist)
                  comment->addField (String ("MUSICBRAINZ_ALBUMARTIST",             String::UTF8),
                                     String (track.mb_album_artist.get(),           String::UTF8));
                if (track.mb_album_artist_id)
                  comment->addField (String ("MUSICBRAINZ_ALBUMARTISTID",           String::UTF8),
                                     String (track.mb_album_artist_id.get(),        String::UTF8));
                if (track.mb_album_artist_sort_name)
                  comment->addField (String ("MUSICBRAINZ_ALBUMARTISTSORTNAME",     String::UTF8),
                                     String (track.mb_album_artist_sort_name.get(),  String::UTF8));
                if (track.mb_track_id)
                  comment->addField (String ("MUSICBRAINZ_TRACKID",                 String::UTF8),
                                     String (track.mb_track_id.get(),               String::UTF8));
                if (track.mb_album_id)
                  comment->addField (String ("MUSICBRAINZ_ALBUMID",                 String::UTF8),
                                     String(track.mb_album_id.get(),                String::UTF8));
                if (track.mb_artist_id)
                  comment->addField (String ("MUSICBRAINZ_ARTISTID",                String::UTF8),
                                     String (track.mb_artist_id.get(),              String::UTF8));
                if (track.mb_artist_sort_name)
                  comment->addField (String ("MUSICBRAINZ_SORTNAME",                String::UTF8),
                                     String (track.mb_artist_sort_name.get(),        String::UTF8));
                if (track.mb_release_date)
                  comment->addField (String ("DATE", String::UTF8),
                                     String (track.mb_release_date.get(), String::UTF8));
                if (track.asin)
                  comment->addField (String ("AMAZON_ASIN",    String::UTF8),
                                     String (track.asin.get(), String::UTF8));
                if (track.puid)
                  comment->addField (String ("MUSICIP_PUID",   String::UTF8),
                                     String (track.puid.get(), String::UTF8));
              }
            else
              {
                throw MetadataWriteError ((boost::format ("Unable to write metadata for file %s: No Tag") % filename).str());
              }

            oggfile.save ();
            comment = oggfile.tag ();
            stat (filename.c_str(), &tagstat);
            TagLib::ByteVector tagData = comment->render();
            const char* blob = tagData.data();
            std::string md5_uid = Util::md5_hex ((char*)blob, tagData.size());
            track.hash = md5_uid+(fsizefmt % (unsigned long long int)(tagstat.st_size)).str();
            return;
        }
#ifdef HAVE_MP4V2
      else
      if (type == "audio/x-m4a")
        {
          TagLib::MP4::File mp4file (filename.c_str());
          if (!mp4file.isOpen() && mp4file.isValid())
                throw MetadataWriteError ((boost::format ("Unable to get an MP4::File from filename: %s") % filename).str());

          MP4::Tag *tag = mp4file.getMP4Tag ();
          if (tag)
            {
              debug ("library", "Modifying MP4::File: %s", filename.c_str());

              metadata_set_common (track, tag);

              MP4FileHandle handle = mp4file.getMP4File ();

              if (track.mb_album_artist_sort_name)    
              MP4SetMetadataFreeForm (handle,
                                      "MusicBrainz Sortname",
                                      (uint8_t*)track.mb_album_artist_sort_name.get().c_str(),
                                      strlen(track.mb_album_artist_sort_name.get().c_str()) + 1);

              if (track.mb_track_id)    
              MP4SetMetadataFreeForm (handle,
                                      "MusicBrainz Track Id",
                                      (uint8_t*)track.mb_track_id.get().c_str(),
                                      strlen(track.mb_track_id.get().c_str()) + 1);

              if (track.mb_album_id)    
              MP4SetMetadataFreeForm (handle,
                                      "MusicBrainz Album Id",
                                      (uint8_t*)track.mb_album_id.get().c_str(),
                                      strlen(track.mb_album_id.get().c_str()) + 1);

              if (track.mb_artist_id)    
              MP4SetMetadataFreeForm (handle,
                                      "MusicBrainz Artist Id",
                                      (uint8_t*)track.mb_artist_id.get().c_str(),
                                      strlen(track.mb_artist_id.get().c_str()) + 1);

              if (track.mb_album_artist_id)    
              MP4SetMetadataFreeForm (handle,
                                      "MusicBrainz Album Artist Id",
                                      (uint8_t*)track.mb_album_artist_id.get().c_str(),
                                      strlen(track.mb_album_artist_id.get().c_str()) + 1);

              if (track.mb_album_artist_sort_name)    
              MP4SetMetadataFreeForm (handle,
                                      "MusicBrainz Album Artist Sortname",
                                      (uint8_t*)track.mb_album_artist_sort_name.get().c_str(),
                                      strlen(track.mb_album_artist_sort_name.get().c_str()) + 1);

              if (track.mb_album_artist)    
              MP4SetMetadataFreeForm (handle,
                                      "MusicBrainz Album Artist",
                                      (uint8_t*)track.mb_album_artist.get().c_str(),
                                      strlen(track.mb_album_artist.get().c_str()) + 1);

              if (track.mb_release_date)    
              MP4SetMetadataFreeForm (handle,
                                      "MusicBrainz Album Release Date",
                                      (uint8_t*)track.mb_release_date.get().c_str(),
                                      strlen(track.mb_release_date.get().c_str()) + 1);

              if (track.asin)    
              MP4SetMetadataFreeForm (handle,
                                      "Amazon ASIN",
                                      (uint8_t*)track.asin.get().c_str(),
                                      strlen(track.asin.get().c_str()) + 1);

              if (track.puid)    
              MP4SetMetadataFreeForm (handle,
                                      "MusicIP PUID",
                                      (uint8_t*)track.puid.get().c_str(),
                                      strlen(track.puid.get().c_str()) + 1);

              mp4file.save ();
              return;
            }
          else
            {
              throw MetadataWriteError ((boost::format ("Unable to write metadata for file %s: No Tag") % filename).str());
            }
        }
#endif
      else
      if (type == "video/x-ms-asf")
        {
          TagLib::WMA::File wmafile (filename.c_str());
          if (!wmafile.isOpen() && wmafile.isValid())
              throw MetadataWriteError ((boost::format ("Unable to get a WMA::File from: %s") % filename).str());

            TagLib::WMA::Tag *tag = wmafile.WMATag();
            if (tag)
              {
                debug ("library", "Modifying WMA::File: %s", filename.c_str());
                metadata_set_common (track, tag);      
                if (track.mb_track_id)
                tag->setAttribute("MusicBrainz/TrackId",
                                   TagLib::String(track.mb_track_id.get().c_str(),
                                   TagLib::String::UTF8));

                if (track.mb_artist_id)
                tag->setAttribute("MusicBrainz/ArtistId",
                                   TagLib::String(track.mb_artist_id.get().c_str(),
                                   TagLib::String::UTF8));

                if (track.mb_artist_sort_name)
                tag->setAttribute("MusicBrainz/SortName",
                                   TagLib::String(track.mb_artist_sort_name.get().c_str(),
                                   TagLib::String::UTF8));

                if (track.mb_album_id)
                tag->setAttribute("MusicBrainz/AlbumId",
                                   TagLib::String(track.mb_album_id.get().c_str(),
                                   TagLib::String::UTF8));

                if (track.mb_album_artist_id)
                tag->setAttribute("MusicBrainz/AlbumArtistId",
                                   TagLib::String(track.mb_album_artist_id.get().c_str(),
                                   TagLib::String::UTF8));

                if (track.mb_album_artist_sort_name)
                tag->setAttribute("MusicBrainz/AlbumArtistSortName",
                                   TagLib::String(track.mb_album_artist_sort_name.get().c_str(),
                                   TagLib::String::UTF8));

                if (track.mb_release_date)
                tag->setAttribute("MusicBrainz/AlbumReleaseDate",
                                   TagLib::String(track.mb_release_date.get().c_str(),
                                   TagLib::String::UTF8));

                if (track.asin)
                tag->setAttribute("Amazon/ASIN",
                                   TagLib::String(track.asin.get().c_str(),
                                   TagLib::String::UTF8));

                if (track.puid)
                tag->setAttribute("MusicIP/PUID",
                                   TagLib::String(track.puid.get().c_str(),
                                   TagLib::String::UTF8));

                wmafile.save();
                return;
              }
            else
              {
                throw MetadataWriteError ((boost::format ("Unable to write metadata for file %s: No Tag") % filename).str());
              }
        }
      else
      if (type == "audio/x-flac")
        {
          TagLib::FLAC::File file (filename.c_str());
          if (file.isValid() && file.tag() && !file.readOnly())
            {
              debug ("library", "Modifying FLAC::File: %s", filename.c_str());
              metadata_set_common (track, file.tag());
              TagLib::Ogg::XiphComment *comment = file.xiphComment(true);
              if (comment)
                {
                  if (track.mb_album_artist)
                      comment->addField (String ("MUSICBRAINZ_ALBUMARTIST", String::UTF8),
                                         String (track.mb_album_artist.get(), String::UTF8));
                  if (track.mb_album_artist_id)
                      comment->addField (String ("MUSICBRAINZ_ALBUMARTISTID", String::UTF8),
                                         String (track.mb_album_artist_id.get(), String::UTF8));
                  if (track.mb_album_artist_sort_name)
                      comment->addField (String ("MUSICBRAINZ_ALBUMARTISTSORTNAME", String::UTF8),
                                         String (track.mb_album_artist_sort_name.get(), String::UTF8));
                  if (track.mb_track_id)
                      comment->addField (String ("MUSICBRAINZ_TRACKID", String::UTF8),
                                         String (track.mb_track_id.get(), String::UTF8));
                  if (track.mb_album_id)
                      comment->addField (String ("MUSICBRAINZ_ALBUMID", String::UTF8),
                                         String(track.mb_album_id.get(), String::UTF8));
                  if (track.mb_artist_id)
                      comment->addField (String ("MUSICBRAINZ_ARTISTID", String::UTF8),
                                         String (track.mb_artist_id.get(), String::UTF8));
                  if (track.mb_artist_sort_name)
                      comment->addField (String ("MUSICBRAINZ_SORTNAME", String::UTF8),
                                         String (track.mb_artist_sort_name.get(), String::UTF8));
                  if (track.mb_release_date)
                      comment->addField (String ("DATE", String::UTF8),
                                         String (track.mb_release_date.get(), String::UTF8));
                  if (track.asin)
                      comment->addField (String ("AMAZON_ASIN", String::UTF8),
                                         String (track.asin.get(), String::UTF8));
                  if (track.puid)
                      comment->addField (String ("MUSICIP_PUID", String::UTF8),
                                         String (track.puid.get(), String::UTF8));
                }
              else
                {
                  throw MetadataWriteError ((boost::format ("Unable to write metadata for file %s: No Tag") % filename).str());
                }

              file.save();
              comment = file.xiphComment ();
              stat (filename.c_str(), &tagstat);
              TagLib::ByteVector tagData = comment->render();
              const char* blob = tagData.data();
              std::string md5_uid = Util::md5_hex ((char*)blob, tagData.size());
              track.hash = md5_uid+(fsizefmt % (unsigned long long int)(tagstat.st_size)).str();
              return;
            }
        }
      else
      if (type == "application/x-apetag")
        {
          TagLib::MPC::File file (filename.c_str());
          if (file.isValid() && file.tag() && !file.readOnly())
            {
              debug ("library", "Modifying MPC::File: %s", filename.c_str());
              metadata_set_common (track, file.tag());
              APE::Tag *tag = file.APETag (true);
              if (tag)
                {
                  if (track.mb_album_artist)
                    tag->addValue ("MUSICBRAINZ_ALBUMARTIST",
                                    String (track.mb_album_artist.get(),            String::UTF8));

                  if (track.mb_album_artist_id)
                    tag->addValue ("MUSICBRAINZ_ALBUMARTISTID",
                                    String (track.mb_album_artist_id.get(),         String::UTF8));

                  if (track.mb_album_artist_sort_name)
                    tag->addValue ("MUSICBRAINZ_ALBUMARTISTSORTNAME",
                                    String (track.mb_album_artist_sort_name.get(),  String::UTF8));

                  if (track.mb_track_id)
                    tag->addValue ("MUSICBRAINZ_TRACKID",
                                    String (track.mb_track_id.get(),                String::UTF8));

                  if (track.mb_album_id)
                    tag->addValue ("MUSICBRAINZ_ALBUMID",
                                    String(track.mb_album_id.get(),                 String::UTF8));

                  if (track.mb_artist_id)
                    tag->addValue ("MUSICBRAINZ_ARTISTID",
                                    String (track.mb_artist_id.get(),               String::UTF8));

                  if (track.mb_artist_sort_name)
                    tag->addValue ("MUSICBRAINZ_SORTNAME",
                                    String (track.mb_artist_sort_name.get(),        String::UTF8));

                  if (track.mb_release_date)
                    tag->addValue ("DATE",
                                    String (track.mb_release_date.get(),            String::UTF8));

                  if (track.asin)
                    tag->addValue ("AMAZON_ASIN",
                                    String (track.asin.get(),                       String::UTF8));

                  if (track.asin)
                    tag->addValue ("MUSICIP_PUID",
                                    String (track.asin.get(),                       String::UTF8));
                } 
              else
                {
                  throw MetadataWriteError ((boost::format ("Unable to write metadata for file %s: No Tag") % filename).str());
                }

              file.save();
              tag = file.APETag ();
              stat (filename.c_str(), &tagstat);
              TagLib::ByteVector tagData = tag->render();
              const char* blob = tagData.data();
              std::string md5_uid = Util::md5_hex ((char*)blob, tagData.size());
              track.hash = md5_uid+(fsizefmt % (unsigned long long int)(tagstat.st_size)).str();
              return;
            }
        }

      debug ("library", "Unable to modify file %s", filename.c_str());
      throw MetadataWriteError ((boost::format ("Unable to write metadata for file %s (Reason Unknown, Presumably unsupported format)") % filename).str());
    }

    void 
    Library::metadata_get_taglib (Glib::ustring const&  uri,
                                  DB::Row&              row) const
    {
      using namespace TagLib;
      using namespace boost;
      using namespace std;
      using namespace Glib;

      string filename;
      try {
          filename = filename_from_uri (uri);
        }
      catch (ConvertError& cxe)
        {
          throw ConversionError ((boost::format ("Unable to convert URI to filename for uri: %s") % uri.c_str()).str());
        } 

      if (!File::isReadable (filename.c_str()))
          throw NonReadableFile ((boost::format ("File is not readable: %s") % filename).str());

      File *track = 0;
      const Datum extra_tags_vorbis[] =
      {
        DATUM_MB_ALBUM_ARTIST,
        DATUM_MB_ALBUM_ARTIST_ID,
        DATUM_MB_ALBUM_ARTIST_SORTNAME,
        DATUM_MB_TRACK_ID,
        DATUM_MB_ALBUM_ID,
        DATUM_MB_ARTIST_ID,
        DATUM_MB_ARTIST_SORTNAME,
        DATUM_MB_RELEASE_DATE,
        DATUM_ASIN,
        DATUM_MUSICIP_PUID,
      };

      const char* extra_tags_vorbis_ids[] =
      {
        "MUSICBRAINZ_ALBUMARTIST",
        "MUSICBRAINZ_ALBUMARTISTID",
        "MUSICBRAINZ_ALBUMARTISTSORTNAME",
        "MUSICBRAINZ_TRACKID",
        "MUSICBRAINZ_ALBUMID",
        "MUSICBRAINZ_ARTISTID",
        "MUSICBRAINZ_SORTNAME",
        "DATE",
        "AMAZON_ASIN",
        "PUID",
      };

      TagLib::FileRef fileref;
      fileref = TagLib::FileRef (filename.c_str());

      if (fileref.isNull())
          throw NoMetadataTaglib ((boost::format ("Unable to create a TagLib FileRef for this file: %s") % filename).str());

      if (!fileref.file()->isValid()) 
          throw NoMetadataTaglib ((boost::format ("TagLib::FileRef is not valid: %s") % filename).str());

      boost::optional<ustring> type;
      string type_found;
      struct stat tagstat;
      stat (filename.c_str(), &tagstat);
      if (Audio::typefind (filename, type_found)) type = type_found;
      if (type && ((type.get() == "application/x-id3") || (type.get() == "audio/mpeg")))
        {
          using namespace TagLib;
          MPEG::File *file;
          if (!(file = dynamic_cast<MPEG::File*> (fileref.file())))
            throw NoMetadataTaglib ((boost::format ("Unable to cast FileRef to MPEG::File: %s") % filename).str());

          ID3v2::Tag *tag = file->ID3v2Tag (false);
          if (tag) get_id3v2 (tag, row, &tagstat);
        }
      else
      if (type && (type.get() == "application/ogg"))
        {
          using namespace TagLib;
          Ogg::Vorbis::File *file = 0;
          if (!(file = dynamic_cast<Ogg::Vorbis::File*> (fileref.file())))
            throw NoMetadataTaglib ((boost::format ("Unable to cast FileRef to Ogg::Vorbis::File: %s") % filename).str());

          Ogg::XiphComment *comment = dynamic_cast<Ogg::Vorbis::File*>(file)->tag (); 
          if (comment)
            {
              TagLib::Map<TagLib::String, TagLib::StringList> const& map = comment->fieldListMap();
              TagLib::Map<TagLib::String, TagLib::StringList>::ConstIterator iter;

              for (unsigned int n = 0; n < G_N_ELEMENTS (extra_tags_vorbis); ++n)
              {
                iter = map.find (extra_tags_vorbis_ids[n]);
                if (iter == map.end()) continue;
                TagLib::StringList const& list = iter->second;
                row_set (row, extra_tags_vorbis[n], ustring (list[0].to8Bit(true)));
              }
            }

          TagLib::ByteVector tagData = comment->render();
          const char* blob = tagData.data(); 
          std::string md5_uid = Util::md5_hex ((char*)blob, tagData.size());
          std::string hash = md5_uid+(fsizefmt % (unsigned long long int)(tagstat.st_size)).str();
          row_set (row, DATUM_HASH, ustring(hash));
        }
      else
      if (type && (type.get() == "audio/x-flac"))
        {
          using namespace TagLib;
          FLAC::File *file;
          if (!(file = dynamic_cast<FLAC::File*> (fileref.file())))
            throw NoMetadataTaglib ((boost::format ("Unable to cast FileRef to FLAC::File: %s") % filename).str());

          Ogg::XiphComment *comment = file->xiphComment ();
          if (comment)
            {
              TagLib::Map<TagLib::String, TagLib::StringList> const& map = comment->fieldListMap();
              TagLib::Map<TagLib::String, TagLib::StringList>::ConstIterator iter;

              for (unsigned int n = 0; n < G_N_ELEMENTS (extra_tags_vorbis); ++n)
              {
                iter = map.find (extra_tags_vorbis_ids[n]);
                if (iter == map.end()) continue;
                TagLib::StringList const& list = iter->second;
                row_set (row, extra_tags_vorbis[n], ustring (list[0].to8Bit(true)));

                TagLib::ByteVector tagData = comment->render();
                const char* blob = tagData.data(); 
                std::string md5_uid = Util::md5_hex ((char*)blob, tagData.size());
                std::string hash = md5_uid+(fsizefmt % (unsigned long long int)(tagstat.st_size)).str();
                row_set (row, DATUM_HASH, ustring(hash));
              }
            }
          else
            {
              ID3v2::Tag *tag = file->ID3v2Tag (false);
              if (tag) get_id3v2 (tag, row, &tagstat);
            }
        }
      else
      if (type && (type.get() == "video/x-ms-asf"))
        {
          WMA::File *file;
          if (!(file = dynamic_cast<WMA::File*> (fileref.file())))
            throw NoMetadataTaglib ((boost::format ("Unable to cast FileRef to WMA::File: %s") % filename).str());

          WMA::Tag *tag = file->WMATag ();
          const TagLib::WMA::AttributeMap &attrMap = tag->attributeMap();
          struct { 
              Bmp::Library::Datum   datum;
              char                 *id;
          } extra_tags_wma[] = {
              { DATUM_MB_TRACK_ID,                "MusicBrainz/TrackId"             },
              { DATUM_MB_ARTIST_ID,               "MusicBrainz/ArtistId"            },
              { DATUM_MB_ALBUM_ID,                "MusicBrainz/AlbumId"             },
              { DATUM_MB_ALBUM_ARTIST,            "MusicBrainz/AlbumArtist"         },
              { DATUM_MB_ALBUM_ARTIST_ID,         "MusicBrainz/AlbumArtistId"       },
              { DATUM_MB_ALBUM_ARTIST_SORTNAME,   "MusicBrainz/AlbumArtistSortName" },
              { DATUM_MB_RELEASE_DATE,            "MusicBrainz/AlbumReleaseDate"    },
              { DATUM_MB_ARTIST_SORTNAME,         "MusicBrainz/SortName"            },
              { DATUM_ASIN,                       "Amazon/ASIN"                     },
              { DATUM_MUSICIP_PUID,               "MusicIP/PUID"                    },
          };

          for (unsigned int n = 0; n < G_N_ELEMENTS (extra_tags_wma); ++n)
            {
              if (attrMap.contains(extra_tags_wma[n].id))
                {
                  row_set (row, extra_tags_wma[n].datum, ustring (attrMap[extra_tags_wma[n].id]->toString().toCString(true)));
                }
            }
        }
#ifdef HAVE_MP4V2
      else
      if (type && (type.get() == "audio/x-m4a"))
        {
          MP4::File *file;
          if (!(file = dynamic_cast<MP4::File*> (fileref.file())))
            throw NoMetadataTaglib ((boost::format ("Unable to cast FileRef to MP4::File: %s") % filename).str());

          MP4FileHandle mp4file = file->getMP4File();
          struct { 
              Bmp::Library::Datum   datum;
              char                 *id;
          } extra_tags_mp4[] = {
              { DATUM_MB_ARTIST_SORTNAME,         "MusicBrainz Sortname"              },
              { DATUM_MB_TRACK_ID,                "MusicBrainz Track Id"              },
              { DATUM_MB_ALBUM_ID,                "MusicBrainz Album Id"              },
              { DATUM_MB_ARTIST_ID,               "MusicBrainz Artist Id"             },
              { DATUM_MB_ALBUM_ARTIST_ID,         "MusicBrainz Album Artist Id"       },
              { DATUM_MB_ALBUM_ARTIST_SORTNAME,   "MusicBrainz Album Artist Sortname" },
              { DATUM_MB_ALBUM_ARTIST,            "MusicBrainz Album Artist"          },
              { DATUM_MB_RELEASE_DATE,            "MusicBrainz Album Release Date"    },
              { DATUM_ASIN,                       "Amazon ASIN"                       },
              { DATUM_MUSICIP_PUID,               "MusicIP PUID"                      },
          };

          for (unsigned int n = 0; n < G_N_ELEMENTS (extra_tags_mp4); ++n) 
            {
              char *value = 0;
              uint32_t size = 0;
              if (MP4GetMetadataFreeForm (mp4file, extra_tags_mp4[n].id, 
                                                (uint8_t**)&value, &size) && value != NULL)
                {
                  value = (char*)realloc (value, size+1);
                  value[size] = '\0';
                  row_set (row, extra_tags_mp4[n].datum, ustring (value));
                  free (value);
                }
            }
        }
#endif
      else
      if (type && (type.get() == "application/x-apetag"))
        {
          MPC::File *file;
          if (!(file = dynamic_cast<MPC::File*> (fileref.file())))
            throw NoMetadataTaglib ((boost::format ("Unable to cast FileRef to MPC::File: %s") % filename).str());

          APE::Tag *tag = file->APETag();
          if (tag)
            {
              const TagLib::APE::ItemListMap &items = tag->itemListMap();
              for (unsigned int n = 0; n < G_N_ELEMENTS (extra_tags_vorbis); ++n)
              {
                if (!items[extra_tags_vorbis_ids[n]].isEmpty())
                  {
                    row_set (row, extra_tags_vorbis[n], ustring (items[extra_tags_vorbis_ids[n]].toString().toCString(true)));
                  }
              }

              TagLib::ByteVector tagData = tag->render();
              const char* blob = tagData.data(); 
              std::string md5_uid = Util::md5_hex ((char*)blob, tagData.size());
              std::string hash = md5_uid+(fsizefmt % (unsigned long long int)(tagstat.st_size)).str();
              row_set (row, DATUM_HASH, ustring(hash));
            }
        }

      track = fileref.file ();
      if (!track) 
            throw NoMetadataTaglib ((boost::format ("Unable to read metadata with TagLib: %s") % filename).str());

      struct stat file_stat;
      stat (filename.c_str(), &file_stat);
      string value;

      row_set (row, DATUM_TRACK, int(track->tag()->track()));
      row_set (row, DATUM_DATE, int(track->tag()->year()));

#ifndef HAVE_HAL
      row_set (row, DATUM_LOCATION, ustring (uri));
#endif

      value = track->tag()->artist().to8Bit (true);
      if (!value.empty())
          row_set (row, DATUM_ARTIST, ustring (value));

      value = track->tag()->album().to8Bit (true);
      if (!value.empty())
          row_set (row, DATUM_ALBUM, ustring (value));

      value = track->tag()->genre().to8Bit (true);
      if (!value.empty())
          row_set (row, DATUM_GENRE, ustring (value));

      value = track->tag()->comment().to8Bit (true);
      if (!value.empty())
          row_set (row, DATUM_COMMENT, ustring (value));

      row_set (row, DATUM_RATING, int(0)); 
      row_set (row, DATUM_COUNT, int(0)); 

      if (track->audioProperties())
        {
          row_set (row, DATUM_BITRATE, int(track->audioProperties()->bitrate()));
          row_set (row, DATUM_SAMPLERATE, int(track->audioProperties()->sampleRate()));
          row_set (row, DATUM_TIME, int(track->audioProperties()->length()));
        }
      else
        {
          row_set (row, DATUM_BITRATE, int(0));
          row_set (row, DATUM_SAMPLERATE, int(0));
          row_set (row, DATUM_TIME, int(0));
        }

      if  ((!track->tag()->title().toCString(true)) || (!strlen(track->tag()->title().toCString(true))))
        {
          row_set (row, DATUM_TITLE, ustring(path_get_basename(filename)));
        }
      else
        {
          value = track->tag()->title().to8Bit(true);
          if (!value.empty())
            {
              trim (value);
              row_set (row, DATUM_TITLE, ustring (value));
            }
        }
    }

    void 
    Library::metadata_get_gst  (const Glib::ustring&  uri,
                                DB::Row&              row) const
    {
      using namespace std;
      using namespace Glib;

      string filename;

      try {
          filename = filename_from_uri (uri);
        }
      catch (ConvertError& cxe)
        {
          throw ConversionError ((boost::format ("Unable to convert URI to filename for uri: %s") % uri.c_str()).str());
        } 

      boost::shared_ptr<TagReadTuple> ttuple;

      ttuple = boost::shared_ptr<TagReadTuple> (new TagReadTuple);
      ttuple->eos = false;
      ttuple->row = &row;

#ifndef HAVE_HAL
      ttuple->row->insert (std::make_pair (track_attributes[DATUM_LOCATION].id, std::string (uri)));
#endif

      GstElement *element   = gst_element_factory_make ("playbin",  "playbin0");
      GstElement *fakesink  = gst_element_factory_make ("fakesink", "fakesink0");
      GstElement *videosink = gst_element_factory_make ("fakesink", "fakesink1");

      g_object_set (G_OBJECT (element), "uri", uri.c_str(), "audio-sink", fakesink, "video-sink", videosink, NULL);

      GstStateChangeReturn state_ret = gst_element_set_state (element, GST_STATE_PAUSED);
      int change_timeout = 5;

      while (state_ret == GST_STATE_CHANGE_ASYNC && change_timeout > 0)
        {
          GstState state;
          state_ret = gst_element_get_state (GST_ELEMENT (element), &state, 0, 1 * GST_SECOND);
          change_timeout--;
        }

      if (state_ret != GST_STATE_CHANGE_FAILURE && state_ret != GST_STATE_CHANGE_ASYNC)
        {
          // Post application specific message so we'll know when to stop the message loop
          GstBus *bus;
          bus = gst_element_get_bus (GST_ELEMENT (element));
          if (bus)
            {
              gst_bus_post (bus, gst_message_new_application (GST_OBJECT (element), 0));
              gst_object_unref (bus);
            }
          // Poll the bus for messages
          metadata_event_loop (ttuple.get(), GST_ELEMENT (element), FALSE);
        }

      GstFormat  format = GST_FORMAT_TIME;
      GstQuery  *query  = gst_query_new_duration (format);

      if (gst_element_query (GST_ELEMENT(element), query))
        {
          gst_element_get_state (GST_ELEMENT(element), 0, 0, 100 * GST_MSECOND);
          gint64 length_in_nanoseconds;
          gst_query_parse_duration (query, &format, &length_in_nanoseconds);
          int length = length_in_nanoseconds / GST_SECOND;
          ttuple->row->insert (std::make_pair (track_attributes[DATUM_TIME].id, DB::Variant (length)));
        }
      gst_query_unref (query);
      gst_element_set_state (GST_ELEMENT (element), GST_STATE_NULL);
      gst_object_unref (GST_ELEMENT (element));
    }

    bool 
    Library::add (const Glib::ustring &uri)
    {
      using namespace Bmp::DB;
      using namespace Glib; 

      Row row;

      try {
          metadata_get_taglib (uri, row);
        }
      catch (Bmp::Library::InvalidFile& cxe)
        {
          g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Metadata Reading Error: %s", cxe.what());
          throw NoMetadata();
        }
      catch (Bmp::Library::NoMetadataTaglib& cxe)
        {
          g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Taglib Metadata Reading Error: %s", cxe.what());
          metadata_get_gst (uri, row);
        }
      catch (Bmp::Library::NoMetadataGst& cxe)
        {}


      if (!row.empty ())
        {

#ifdef HAVE_HAL
          if (hal->is_initialized()) try {
                HAL::Volume volume = hal->get_volume_for_uri (uri);
                row_set (row, DATUM_HAL_VOLUME_UDI, ustring (volume.volume_udi));
                row_set (row, DATUM_HAL_DEVICE_UDI, ustring (volume.device_udi));
                ustring volume_relative_path = filename_from_uri (uri).substr (volume.mount_path.length());
                ustring volume_relative_base_dir = Glib::path_get_dirname (volume_relative_path); 
                row_set (row, DATUM_VOLUME_RELATIVE_PATH, ustring (volume_relative_path));
                row_set (row, DATUM_VOLUME_RELATIVE_BASE_DIR, ustring (volume_relative_base_dir));
              }
          catch (HAL::Exception& cxe) 
              {
                g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "HAL Error: %s", cxe.what());
                return false;
              }
#endif
          Track track (row);
          if (!(track.artist && track.album && track.title))
            {
              debug ("library", "Track '%s' does NOT qualify for insertion", uri.c_str());
              return false;
            }


          /////////////////////////////////////// Hash Based Replacing 
          debug ("library", "Track '%s' qualifies for insertion", uri.c_str());

#ifdef HAVE_HAL
          // Check Hash
          if (track.hash)
            {
              if (db_main->attr_exists ("main", track_attributes[DATUM_HASH].id, track.hash.get()))
                {
                  DB::AttributeV a;
                  a.push_back (Attribute (DB::EXACT, track_attributes[DATUM_HASH].id, track.hash.get())); 
                  DB::RowV rowv;
                  query (a, rowv); 
                  if ((!rowv.empty()))
                    {
                      DB::Row& row = rowv[0];
                      Track l_track (row);
                      if (l_track.location.get() != uri)
                        {
                            db_main->sqlite_exec_simple (sql_uprintf ("UPDATE main SET %q='%q', %q='%q' WHERE %q='%q';",
                                                                       track_attributes[DATUM_VOLUME_RELATIVE_PATH].id, 
                                                                       track.volume_relative_path.get().c_str(),
                                                                       track_attributes[DATUM_HAL_VOLUME_UDI].id, 
                                                                       track.volume_udi.get().c_str(),
                                                                       track_attributes[DATUM_HASH].id, 
                                                                       track.hash.get().c_str()));
                          TrackUpdated update_track;
                          update_track = track;
                          signal_row_updated_.emit (track.location.get(), update_track);   
                        }
                      return false;
                    }
                }
            }
#else
          // Check Hash
          if (track.hash)
            {
              if (db_main->attr_exists ("main", track_attributes[DATUM_HASH].id, track.hash.get()))
                {
                  DB::AttributeV a;
                  a.push_back (Attribute (DB::EXACT, track_attributes[DATUM_HASH].id, track.hash.get())); 
                  DB::RowV rowv;
                  query (a, rowv); 
                  if ((!rowv.empty()))
                    {
                      DB::Row& row = rowv[0];
                      Track l_track (row);
                      if (l_track.location.get() != uri)
                        {
                            db_main->sqlite_exec_simple (sql_uprintf ("UPDATE main SET %q='%q' WHERE %q='%q';",
                                                                       track_attributes[DATUM_LOCATION].id, 
                                                                       track.location.get().c_str(),
                                                                       track_attributes[DATUM_HASH].id, 
                                                                       track.hash.get().c_str()));

                            TrackUpdated update_track;
                            update_track = track;
                            signal_row_updated_.emit (track.location.get(), update_track);   
                        }
                      return false;
                    }
                }
            }
#endif

          /////////////////////////////////////// Static adding

          // Add the Row anew 
          try {
              row_set (row, DATUM_NEW_ITEM, bool(true)); 
              db_main->add ("main", row);
              return true;
            }
          catch (Bmp::DB::SqlError& cxe)
            {
              AttributeV a;
              RowV vector;

#ifndef HAVE_HAL
              a.push_back (Attribute (EXACT, track_attributes[DATUM_LOCATION].id, uri)); 
#else
              a.push_back (Attribute (EXACT, track_attributes[DATUM_VOLUME_RELATIVE_PATH].id, track.volume_relative_path.get()));
              a.push_back (Attribute (EXACT, track_attributes[DATUM_HAL_VOLUME_UDI].id, track.volume_udi.get()));
#endif

              query (a, vector);

              if (!vector.empty())
                {
                  try {
                      db_main->del ("main", a);
                      row_set (row, DATUM_NEW_ITEM, bool(false)); 
                      db_main->add ("main", row);
                      TrackUpdated update_track;
                      update_track = track;
                      signal_row_updated_.emit (track.location.get(), update_track);   
                      return true;
                    }
                  catch (Bmp::DB::SqlError& cxe)
                    {
                      return false;
                    }
                }
            }
        }
      return false;
    }

    void
    Library::del (const DB::AttributeV &a)
    {
      signal_processing_start_.emit ();
      db_main->del("main", a);
      signal_processing_end_.emit ();
    }

    void
    Library::query (const DB::AttributeV & a,
                    DB::RowV & vector,
                    DB::Database * db,
                    std::string name) const
    {
      Glib::ustring suffix;

      suffix.append (" ORDER BY ");
      suffix.append (track_attributes[DATUM_TRACK].id);

      if (!db)
        {
          db_main->get (name, a, vector, suffix);
        }
      else
        {
          db->get (name, a, vector, suffix);
        }
    }

    bool
    Library::is_library_track (Track const& track)
    {
      using namespace Bmp::DB;
      using namespace Glib;

      Bmp::URI u (track.location.get());
      Bmp::URI::Protocol p = u.get_protocol();
      if (p != Bmp::URI::PROTOCOL_FILE) return false;

      AttributeV a;
      RowV vector;

#ifndef HAVE_HAL
      a.push_back (Attribute (EXACT, track_attributes[DATUM_LOCATION].id, track.location.get())); 
#else
      a.push_back (Attribute (EXACT, track_attributes[DATUM_VOLUME_RELATIVE_PATH].id, track.volume_relative_path.get()));
      a.push_back (Attribute (EXACT, track_attributes[DATUM_HAL_VOLUME_UDI].id, track.volume_udi.get()));
#endif

      query (a, vector);
      return !vector.empty();
    }

    bool Library::album_exists (Album const& album_in)
    {
      using namespace Glib; 
      using namespace Bmp::DB;

      AttributeV a;
      RowV v;
      album_to_attrs (album_in, a);
      db_main->get ("albums", a, v);
      return (!v.empty());
    }

    bool Library::library_has_new_items ()
    {
      using namespace Glib; 

      ustring query ("SELECT DISTINCT new_item FROM main WHERE new_item>0;"); 
      return (db_main->sqlite_exec_simple (query) > 0);
    }

    bool Library::album_has_new_items (Album const& album_in)
    {
      using namespace Glib; 
      using namespace Bmp::DB;

      ustring query ("SELECT count(*) FROM main WHERE "); 
      AttributeV a;
      album_to_attrs (album_in, a);
      append_attributes (query, a); 
      query.append (" AND new_item > 0;");

      return (db_main->sqlite_exec_simple (query) > 0);
    }

    Album Library::get_album (Album const& album)
    {
      using namespace Glib; 
      using namespace Bmp::DB;

      AttributeV a;
      RowV v;
      album_to_attrs (album, a);
      db_main->get ("albums", a, v);
      if (v.empty()) return Album(); 
      return Bmp::Library::Album(v[0]);
    }

    void Library::get_albums (DB::RowV & vector)
    {
      using namespace Bmp::DB;

      AttributeV a;
      db_main->get ("albums", a, vector);
    }

    GHashTable* Library::get_hashtable (Glib::ustring const& uri) const
    {
      using namespace Glib;
      using namespace std;

      GHashTable *table;
      GValue     *value;
      DB::Row     row;

      get_metadata_row (uri, row);

      table = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
      for (unsigned int n = 0; n < G_N_ELEMENTS (track_attributes); n++)
      {
        DB::Row::const_iterator i = row.find (track_attributes[n].id);
        if (i == row.end())
          {
            g_warning (G_STRLOC ": DB::Attribute '%s' not found in Row (this is currently normal for cdda:/// URIs)", track_attributes[n].id);
            continue;
          }

        value = g_new0 (GValue,1);
        switch (Bmp::DB::ValueType(i->second.which()))
          {
          case Bmp::DB::VALUE_TYPE_STRING:
              g_value_init (value, G_TYPE_STRING);
              g_value_set_string (value, (boost::get<ustring>(i->second)).c_str());
              break;

          case Bmp::DB::VALUE_TYPE_INT:
              g_value_init (value, G_TYPE_INT);
              g_value_set_int (value, (boost::get<int>(i->second)));
              break;

          case Bmp::DB::VALUE_TYPE_BOOL:
              g_value_init (value, G_TYPE_BOOLEAN);
              g_value_set_boolean (value, (boost::get<bool>(i->second)));
              break;

          case Bmp::DB::VALUE_TYPE_REAL:
              g_value_init (value, G_TYPE_DOUBLE);
              g_value_set_double (value, (boost::get<double>(i->second)));
              break;
        }
        g_hash_table_insert (table, g_strdup(track_attributes[n].id), value);
      }
      return table;
    }

    void Library::get_metadata_row (Glib::ustring const& uri, Bmp::DB::Row& row) const
    {
      using namespace Bmp::DB;
      using namespace std;
      using namespace Glib;

      AttributeV a;
      RowV vector;
#ifndef HAVE_HAL
      a.push_back (Attribute (EXACT, track_attributes[DATUM_LOCATION].id, uri)); 
      query (a, vector);
#else
      HAL::Volume volume = hal->get_volume_for_uri (uri);
      ustring volume_relative_path = filename_from_uri (uri).substr (volume.mount_path.length());
      a.push_back (Attribute (EXACT, track_attributes[DATUM_HAL_VOLUME_UDI].id, volume.volume_udi)); 
      a.push_back (Attribute (EXACT, track_attributes[DATUM_VOLUME_RELATIVE_PATH].id, volume_relative_path)); 
      query (a, vector);
#endif
      if (vector.empty())
        {
          try {
              metadata_get_taglib (uri, row);
            }
          catch (Bmp::Library::InvalidFile& cxe)
            {
              g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Metadata Reading Error: %s", cxe.what());
              throw NoMetadata();
            }
          catch (Bmp::Library::NoMetadataTaglib& cxe)
            {
              g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Taglib Metadata Reading Error: %s", cxe.what());
              metadata_get_gst (uri, row);
            }
          catch (Bmp::Library::NoMetadataGst& cxe)
            {}

#ifdef HAVE_HAL
          if (hal->is_initialized()) try {
                HAL::Volume volume = hal->get_volume_for_uri (uri);
                row_set (row, DATUM_HAL_VOLUME_UDI, ustring (volume.volume_udi));
                row_set (row, DATUM_HAL_DEVICE_UDI, ustring (volume.device_udi));
                ustring volume_relative_path = filename_from_uri (uri).substr (volume.mount_path.length());
                ustring volume_relative_base_dir = path_get_dirname (volume_relative_path); 
                row_set (row, DATUM_VOLUME_RELATIVE_PATH, ustring (volume_relative_path));
                row_set (row, DATUM_VOLUME_RELATIVE_BASE_DIR, ustring (volume_relative_base_dir));
              }
          catch (HAL::Exception& cxe) //FIXME: Throw something
              {
                g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "HAL Error: %s", cxe.what());
                return;
              }
#endif
        }
      else
        {
          row = vector[0]; 
        } 
    }

    void
    Library::get_metadata (Glib::ustring const& uri,
                           Track& track)
    {
      Bmp::DB::Row row;
      Bmp::URI u (uri);
      Bmp::URI::Protocol p = u.get_protocol ();

      get_metadata_row (uri, row);

      if (p == Bmp::URI::PROTOCOL_HTTP || 
          p == Bmp::URI::PROTOCOL_MMS || 
          p == Bmp::URI::PROTOCOL_MMSU || 
          p == Bmp::URI::PROTOCOL_MMST ||
          p == Bmp::URI::PROTOCOL_CDDA)
        {
#ifdef HAVE_HAL
          track = Track (row, false);
#else
          track = Track (row);
#endif
          return;
        }
        track = Track (row);
    }

    DatumDefine Library::get_metadatum_define (Bmp::Library::Datum datum) const
    {
        return track_attributes[datum];
    }

    void
    Library::drop_items ()
    {
      db_main->sqlite_exec_simple (sql_uprintf("DELETE FROM main WHERE new_item > 0;"));  
    }


    void
    Library::approve_items ()
    {
      db_main->sqlite_exec_simple (sql_uprintf("UPDATE main SET new_item=0;")); 
    }


    void
    Library::approve_items (Album const& a)
    {
      using namespace Bmp::DB;
      using namespace Glib;

      ustring query ("UPDATE main SET new_item=0 WHERE "); 
      AttributeV attrs;
      album_to_attrs (a, attrs);
      append_attributes (query, attrs); 
      query.append (";");

      db_main->sqlite_exec_simple (query);
    }


    void
    Library::increase_play_count (Glib::ustring const& uri)
    {
      using namespace Bmp::DB;
      using namespace Glib;
#ifndef HAVE_HAL
      db_main->sqlite_exec_simple (sql_uprintf ("UPDATE main SET %s='%d', %s = %s +1 WHERE %s='%q';",
                                   track_attributes[DATUM_PLAYDATE].id,
                                   time (NULL),
                                   track_attributes[DATUM_COUNT].id,
                                   track_attributes[DATUM_COUNT].id,
                                   track_attributes[DATUM_LOCATION].id,
                                   uri.c_str()));
      DB::AttributeV a; 
      RowV v;
      a.push_back (DB::Attribute (DB::EXACT, track_attributes[DATUM_LOCATION].id, uri)); 
      db_main->get ("main", a, v);
      if (!v.empty())
        {
          Track track (v[0]);
          TrackUpdated update_track;
          update_track = track; 
          signal_row_updated_.emit (update_track.location_cur.get(), update_track);   
        }
#else
      HAL::Volume volume = hal->get_volume_for_uri (uri);
      ustring volume_relative_path = filename_from_uri (uri).substr (volume.mount_path.length());
      db_main->sqlite_exec_simple (sql_uprintf ("UPDATE main SET %s='%d', %s = %s +1 WHERE %s='%q' AND %s='%q';",
                                   track_attributes[DATUM_PLAYDATE].id,
                                   time (NULL),
                                   track_attributes[DATUM_COUNT].id,
                                   track_attributes[DATUM_COUNT].id,
                                   track_attributes[DATUM_HAL_VOLUME_UDI].id,
                                   volume.volume_udi.c_str(),
                                   track_attributes[DATUM_VOLUME_RELATIVE_PATH].id,
                                   volume_relative_path.c_str()));
      DB::AttributeV a; 
      RowV v;

      a.push_back (DB::Attribute (DB::EXACT, track_attributes[DATUM_HAL_VOLUME_UDI].id, volume.volume_udi)); 
      a.push_back (DB::Attribute (DB::EXACT, track_attributes[DATUM_VOLUME_RELATIVE_PATH].id, volume_relative_path)); 

      db_main->get ("main", a, v);

      if (!v.empty())
        {
          Track track (v[0]);
          TrackUpdated update_track;
          update_track = track; 
          signal_row_updated_.emit (update_track.location_cur.get(), update_track);   
        }
#endif
    }

#if 0
    MB_AudioCD_Status 
    Library::add_audiocd (std::string const& cdindex)
    {
        using namespace Bmp::DB;
        using namespace std; 

        MB_AudioCD_Status status = MB_NONE;
        RowV rows;

        get_audiocd_metadata (status, rows);

        if (status == Bmp::Library::MB_FAILURE)
          {
            debug ("library", "get_audiocd_metadata() status is MB_FAILURE");
            throw Exception ("Unable to acquire metadata for AudioCD", Exception::NO_METADATA, cdindex);
          }

        if (status == Bmp::Library::MB_OFFLINE)
          {
            return status;
          }

        AttributeV a;
        RowV vector;
        a.push_back (Attribute (EXACT, track_attributes[DATUM_LOCATION].id,
              Glib::ustring((boost::format ("cdda://%s/1") % cdindex).str())));
        query (a, vector, db_cdda, "main");

        if (vector.empty())
        {
          debug ("library", "CD Seems to be not cached yet");
          for (RowV::const_iterator iter = rows.begin (); iter != rows.end (); ++iter)
          {
            db_cdda->add ("main", (*iter));
          }
        }
        return status;
    }
#endif

    void
    Library::recreate_db ()
    {
        using namespace Bmp::DB;
        using namespace std;
        using namespace Glib;

        delete db_main;
        db_main = new Database ("library", BMP_PATH_DATA_DIR, true);

        ValueMap meta;
        meta.insert (make_pair ("version", VALUE_TYPE_STRING));
        db_main->create_table ("meta", "version", meta);

        ustring sql;

        sql .append ("INSERT INTO meta (version) VALUES ('")
            .append (string (MLIB_VERSION))
            .append ("');");

        db_main->sqlite_exec_simple (sql.c_str());
    }

    bool
    Library::modify_tracks (TrackUpdatedV& tracks, bool retag, bool db_update, bool relocate)
    {
      signal_processing_start_.emit ();
      signal_modify_start_.emit (tracks.size ());

      for (TrackUpdatedV::iterator i_tracks  = tracks.begin (),
                                   i_end     = tracks.end ();
                                   i_tracks != i_end; ++i_tracks)
        {
          using namespace Bmp::DB;
          using namespace Glib; 
          using namespace std; 

          TrackUpdated& track = (*i_tracks);
          DB::AttributeVSet a;

          if (db_update)
            {
              if (track.title)
                a.push_back (DB::AttributeSet (track_attributes[DATUM_TITLE].id,
                                                   track.title.get().raw())); 
              if (track.artist)
                a.push_back (DB::AttributeSet (track_attributes[DATUM_ARTIST].id,
                                                   track.artist.get().raw())); 
              if (track.album)
                a.push_back (DB::AttributeSet (track_attributes[DATUM_ALBUM].id,
                                                   track.album.get().raw())); 
              if (track.genre)
                a.push_back (DB::AttributeSet (track_attributes[DATUM_GENRE].id,
                                                   track.genre.get().raw())); 
              if (track.comment)
                a.push_back (DB::AttributeSet (track_attributes[DATUM_COMMENT].id,
                                                   track.comment.get().raw())); 
              if (track.tracknumber)
                a.push_back (DB::AttributeSet (track_attributes[DATUM_TRACK].id,
                                                   int(track.tracknumber.get()))); 
              if (track.rating)
                 a.push_back (DB::AttributeSet (track_attributes[DATUM_RATING].id,
                                                    int(track.rating.get()))); 
              if (track.count)
                a.push_back (DB::AttributeSet (track_attributes[DATUM_COUNT].id,
                                                   int(track.count.get()))); 
              if (track.asin)
                a.push_back (DB::AttributeSet (track_attributes[DATUM_ASIN].id,
                                                   track.asin.get().raw())); 
              if (track.puid)
                a.push_back (DB::AttributeSet (track_attributes[DATUM_MUSICIP_PUID].id,
                                                   track.puid.get().raw())); 
              if (track.mb_album_artist)
                a.push_back (DB::AttributeSet (track_attributes[DATUM_MB_ALBUM_ARTIST].id,
                                                   track.mb_album_artist.get().raw())); 
              if (track.mb_album_artist_id)
                a.push_back (DB::AttributeSet (track_attributes[DATUM_MB_ALBUM_ARTIST_ID].id,
                                                   track.mb_album_artist_id.get().raw())); 
              if (track.mb_album_artist_sort_name)
                a.push_back (DB::AttributeSet (track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].id,
                                                   track.mb_album_artist_sort_name.get().raw())); 
              if (track.mb_track_id)
                a.push_back (DB::AttributeSet (track_attributes[DATUM_MB_TRACK_ID].id,
                                                   track.mb_track_id.get().raw())); 
              if (track.mb_album_id)
                a.push_back (DB::AttributeSet (track_attributes[DATUM_MB_ALBUM_ID].id,
                                                   track.mb_album_id.get().raw())); 
              if (track.mb_artist_id)
                a.push_back (DB::AttributeSet (track_attributes[DATUM_MB_ARTIST_ID].id,
                                                   track.mb_artist_id.get().raw())); 
              if (track.mb_artist_sort_name)
                a.push_back (DB::AttributeSet (track_attributes[DATUM_MB_ARTIST_SORTNAME].id,
                                                   track.mb_artist_sort_name.get().raw())); 
              if (track.mb_release_date)
                {
                    a.push_back (DB::AttributeSet (track_attributes[DATUM_MB_RELEASE_DATE].id,
                                                       track.mb_release_date.get().raw())); 
                    int y = 0, m = 0, d = 0;
                    sscanf(track.mb_release_date.get().c_str(), "%04d-%02d-%02d", &y, &m, &d);
                    a.push_back (DB::AttributeSet (track_attributes[DATUM_DATE].id,
                                                       int(y))); 
                }
              else
                {
                  if (track.date)
                    a.push_back (DB::AttributeSet (track_attributes[DATUM_DATE].id,
                                                       int(track.date.get()))); 
                }

              if (track.new_item)
                a.push_back (DB::AttributeSet (track_attributes[DATUM_NEW_ITEM].id,
                                                   bool(track.new_item.get())));
            }
      
          boost::optional<ustring> modify_location;

#ifdef HAVE_HAL
          if ((track.location_new) && hal->is_initialized() && relocate)
          {
            HAL::Volume volume = hal->get_volume_for_uri (track.location_new.get());
            string filename1 = filename_from_uri (track.location_cur.get());
            string filename2 = filename_from_uri (track.location_new.get());
            try
              {
                Util::copy_file (filename1, filename2);
                if (Util::compare_files (filename1, filename2))
                {
                  ustring volume_relative_path = filename2.substr (volume.mount_path.length());
                  g_unlink (filename1.c_str());
                  a.push_back (DB::AttributeSet (track_attributes[DATUM_VOLUME_RELATIVE_PATH].id, volume_relative_path));
                  a.push_back (DB::AttributeSet (track_attributes[DATUM_HAL_VOLUME_UDI].id, volume.volume_udi));
                  a.push_back (DB::AttributeSet (track_attributes[DATUM_HAL_DEVICE_UDI].id, volume.device_udi));
                  modify_location = filename_to_uri (filename2);
                }
              }
            catch (std::exception& cxe)
              {
                g_critical ("%s: Unable to copy file to destination. Reason: '%s'. Source: '%s'. Destination: '%s'",
                             G_STRLOC, cxe.what(), filename1.c_str(), filename2.c_str());
                continue;
              }
          }
#else
          if (track.location_new && relocate) 
          {
            string filename1 = filename_from_uri (track.location_cur.get());
            string filename2 = filename_from_uri (track.location_new.get());

            try {
              Util::copy_file (filename1, filename2);
              if (Util::compare_files (filename1, filename2))
                {
                  g_unlink (filename1.c_str());
                  a.push_back (DB::AttributeSet (track_attributes[DATUM_LOCATION].id, Glib::filename_to_uri(filename2))); 
                  modify_location = Glib::filename_to_uri(filename2);
                }
              else
                {
                  g_unlink (filename2.c_str());
                }

              }
            catch (std::exception& cxe)
              {
                g_critical ("%s: Unable to copy file to destination. Reason: '%s'. Source: '%s'. Destination: '%s'",
                            G_STRLOC, cxe.what(), filename1.c_str(), filename2.c_str());
                continue;
              }

          }
#endif
          try {

              if (retag)
                {
                  metadata_set_taglib (modify_location ? modify_location.get() : track.location_cur.get() , track);
                  if (track.hash)
                    {
                      a.push_back (DB::AttributeSet (track_attributes[DATUM_HASH].id,
                                                         track.hash.get().raw())); 
                    }
                }

#ifdef HAVE_HAL
              ustring where; 
              where.append (sql_uprintf (" WHERE %q='%q' AND %q='%q' ",
                                           track_attributes[DATUM_HAL_VOLUME_UDI].id,
                                           track.volume_udi.get().c_str(),
                                           track_attributes[DATUM_VOLUME_RELATIVE_PATH].id,
                                           track.volume_relative_path.get().c_str()));
              db_main->set_custom ("main",
                                    where,
                                    a);
#else
              db_main->set        ("main",
                                    track_attributes[DATUM_LOCATION].id,
                                    track.location_cur.get(),
                                    a);
#endif
              signal_row_updated_.emit (track.location_cur.get(), track);   
            }
          catch (Bmp::Library::Exception& cxe) {}
          catch (Bmp::DB::SqlError& cxe) {}
          
          signal_modify_step_.emit (); 

          while (g_main_context_pending (NULL)) g_main_context_iteration (NULL, TRUE);
      }   

      signal_modify_end_.emit ();
      signal_processing_end_.emit ();

      return true;
    }

    bool Library::collection_exists (std::string const& name) const
    {
      return db_main->sqlite_exec_simple (sql_uprintf ("SELECT name FROM tags WHERE name='%q';", name.c_str())); 
    }

    void Library::collection_create (std::string const& name, Glib::ustring const& desc, CollectionAttributes const& attrs) 
    {
      using namespace std;
      using namespace Bmp::DB;
      using namespace Glib;

      db_main->sqlite_exec_simple (sql_uprintf ("INSERT INTO tags (name, desc) VALUES ('%q', '%q');", name.c_str(),
                                                                                                      desc.c_str())); 
      db_tags->create_table  (name.c_str(), string (tag_attributes[TAG_DATUM_ID].id), map_tag);

      for (CollectionAttributes::const_iterator a = attrs.begin (), e = attrs.end (); a != e; ++a)
        {
          Row row;
          CollectionAttribute const& attr = (*a);
        
          row_tag_set (row, TAG_DATUM_ATTR_NAME, ustring (track_attributes[attr.datum].id));
          row_tag_set (row, TAG_DATUM_ATTR_TYPE, int(attr.value.which())); 
          row_tag_set (row, TAG_DATUM_ATTR_PRECISE, bool(attr.precise)); 

          switch (ValueType (attr.value.which ()))
            {
              case VALUE_TYPE_BOOL:
                  row_tag_set (row,
                               TAG_DATUM_ATTR_V_BOOL,
                               attr.value);
                  break;

              case VALUE_TYPE_INT:
                  row_tag_set (row,
                               TAG_DATUM_ATTR_V_INT,
                               attr.value);
                  break;

              case VALUE_TYPE_REAL:
                  row_tag_set (row,
                               TAG_DATUM_ATTR_V_REAL,
                               attr.value);
                  break;

              case VALUE_TYPE_STRING:
                  row_tag_set (row,
                               TAG_DATUM_ATTR_V_STRING,
                               attr.value);
                  break;

            }
          db_tags->add (name.c_str(), row);
        }
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    Library::Library ()
    {
      using namespace Bmp::DB;
      using namespace std;
      using namespace Glib;
  
      register_taglib_plugins (); 

      //-- Create 'main' database
#ifndef HAVE_HAL
      for (unsigned int n = 0; n < G_N_ELEMENTS(track_attributes); n++)
        {
          map_main.insert (std::make_pair (std::string (track_attributes[n].id), track_attributes[n].type));
        }
#else
      for (unsigned int n = 1; n < G_N_ELEMENTS(track_attributes); n++)
        {
          map_main.insert (std::make_pair (std::string (track_attributes[n].id), track_attributes[n].type));
        }
#endif

      try {
        db_main = new Database ("library", BMP_PATH_DATA_DIR, false);
      } catch (Bmp::DB::DbInitError& cxe) { abort (); }

      ValueMap meta;
      meta.insert (make_pair ("version", VALUE_TYPE_STRING));
      db_main->create_table  ("meta", "version", meta);

      AttributeV a;
      a.push_back (DB::Attribute (DB::EXACT, std::string("version"), std::string(MLIB_VERSION))); 

      RowV v;
      try {
          db_main->get ("meta", a, v);
          if (v.empty())
            { 
              recreate_db ();
            }
          else
            {
              if (boost::get<ustring>((*(v[0].find ("version"))).second) != ustring (MLIB_VERSION)) recreate_db ();
            }
        }
      catch (Bmp::DB::SqlError& cxe) { recreate_db (); }

#ifndef HAVE_HAL
      db_main->create_table  ("main", string (track_attributes[DATUM_LOCATION].id), map_main);
#else

      try {
          ustring sql;
          sql.append ("CREATE TABLE IF NOT EXISTS main (");
          for (unsigned int n = 1; n < G_N_ELEMENTS(track_attributes); n++)
          {
            sql.append ("'");
            sql.append (track_attributes[n].id);
            sql.append ("' ");
            
            if (track_attributes[n].type == VALUE_TYPE_STRING)
              {
                sql.append (" TEXT ");
              }
            else
            if (track_attributes[n].type == VALUE_TYPE_INT)
              {
                sql.append (" INTEGER DEFAULT 0 ");
              }
            else
            if (track_attributes[n].type == VALUE_TYPE_BOOL)
              {
                sql.append (" BOOL DEFAULT 0 ");
              }
            sql.append (", ");
          }

          sql.append (" PRIMARY KEY ('");
          sql.append (track_attributes[DATUM_HAL_VOLUME_UDI].id);
          sql.append ("', '"); 
          sql.append (track_attributes[DATUM_VOLUME_RELATIVE_PATH].id);
          sql.append ("') ON CONFLICT REPLACE);"); 
          db_main->sqlite_exec_simple (sql.c_str());
        }
      catch (Bmp::DB::SqlError& cxe)
        {
          abort (); 
        }

      db_main->insert_map_for_table ("main", map_main);
#endif

      //Create albums view
      map_insert (map_albums, DATUM_ARTIST);
      map_insert (map_albums, DATUM_ALBUM);
      map_insert (map_albums, DATUM_ASIN);
      map_insert (map_albums, DATUM_NEW_ITEM);
      map_insert (map_albums, DATUM_MB_ALBUM_ID);
      map_insert (map_albums, DATUM_MB_ALBUM_ARTIST_ID);
      map_insert (map_albums, DATUM_MB_ALBUM_ARTIST_SORTNAME);
      map_albums.insert (make_pair (string (album_attributes[0].id),
                                            album_attributes[0].type));

#if 0
      db_main->create_view_manual ("main", "albums",
                                    sql_uprintf ("SELECT DISTINCT %q, %q, %q, max(%q) as %s, %q, %q, %q, count(%q) AS count "
                                                 "FROM main GROUP BY %q, %q, %q, %q, ifnull(%q, %q) ORDER BY %q;",
#endif
      db_main->create_view_manual ("main", "albums",
                                    sql_uprintf ("SELECT DISTINCT %q, %q, %q, %q, %q, %q, %q, count(%q) AS count "
                                                 "FROM main GROUP BY %q, %q, %q, %q, ifnull(%q, %q) ORDER BY %q;",


                                                  track_attributes[DATUM_ARTIST].id, // SELECT
                                                  track_attributes[DATUM_ALBUM].id,
                                                  track_attributes[DATUM_ASIN].id,
                                                  track_attributes[DATUM_NEW_ITEM].id,
                                                  track_attributes[DATUM_MB_ALBUM_ID].id,
                                                  track_attributes[DATUM_MB_ALBUM_ARTIST_ID].id,
                                                  track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].id,
#ifdef HAVE_HAL
                                                  track_attributes[DATUM_VOLUME_RELATIVE_PATH].id, // count (%q)
#else
                                                  track_attributes[DATUM_LOCATION].id, // count (%q)
#endif

                                                  track_attributes[DATUM_ALBUM].id,
                                                  track_attributes[DATUM_ASIN].id,
                                                  track_attributes[DATUM_MB_ALBUM_ID].id,
                                                  track_attributes[DATUM_MB_ALBUM_ARTIST_ID].id,
                                                    
                                                  track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].id, // ifnull (%q, %q) 
                                                  track_attributes[DATUM_ARTIST].id,

                                                  track_attributes[DATUM_ALBUM].id), // ORDER BY

                                                  map_albums);

      // Create 'tags' map 
      for (unsigned int n = 0; n < G_N_ELEMENTS(tags_attributes); n++)
        {
          map_tags.insert (std::make_pair (std::string (tags_attributes[n].id), tags_attributes[n].type));
        }

      // Create 'tag' map 
      for (unsigned int n = 0; n < G_N_ELEMENTS(tag_attributes); n++)
        {
          map_tag.insert (std::make_pair (std::string (tag_attributes[n].id), tag_attributes[n].type));
        }

      db_main->create_table  ("tags", string (tags_attributes[TS_DATUM_ID].id), map_tags);
      db_tags = new Database ("tags", BMP_PATH_DATA_DIR, false);

      //-- Create 'cdda' database
      for (unsigned int n = 0; n < G_N_ELEMENTS(cdda_data); n++)
        {
          map_cdda.insert (std::make_pair (std::string (track_attributes[cdda_data[n]].id), track_attributes[cdda_data[n]].type));
        }
      db_cdda = new Database ("cdda", BMP_PATH_DATA_DIR, false);
      db_cdda->create_table ("main", string (track_attributes[DATUM_LOCATION].id), map_cdda);

      db_main->sqlite_exec_simple ("ANALYZE;");
    }

    Library::~Library ()
    {
        delete db_main;
        delete db_tags;
    }

    // Generic "processing" progress #FIXME: What is this good for?  --deadchip
    Library::SignalProcessing&
    Library::signal_processing_start ()
    { 
      return signal_processing_start_;
    }

    Library::SignalProcessing&
    Library::signal_processing_end ()
    { 
      return signal_processing_end_;
    }

    Library::SignalRowUpdated&
    Library::signal_row_updated ()
    {
      return signal_row_updated_;
    }

    // Modification progress
    Library::ModifyStart&
    Library::signal_modify_start()
    { 
      return signal_modify_start_;
    }

    Library::ModifyStep&
    Library::signal_modify_step()
    { 
      return signal_modify_step_;
    }

    Library::ModifyEnd&
    Library::signal_modify_end()
    {
      return signal_modify_end_;
    }

    Album::Album (const Bmp::DB::Row& row)
    {
      using namespace Bmp::DB;
      using namespace Glib; 

      Row::const_iterator i;
      Row::const_iterator end = row.end();

      i = row.find (track_attributes[DATUM_ARTIST].id);
      if (i != end && !(boost::get<ustring>(i->second).empty()))
          artist = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_ALBUM].id);
      if (i != end && !(boost::get<ustring>(i->second).empty()))
          album = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_ASIN].id);
      if (i != end && !(boost::get<ustring>(i->second).empty()))
          asin = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_MB_ALBUM_ID].id);
      if (i != end && !(boost::get<ustring>(i->second).empty()))
          release_id = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_MB_ALBUM_ARTIST_ID].id);
      if (i != end && !(boost::get<ustring>(i->second).empty()))
          artist_id = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].id);
      if (i != end && !(boost::get<ustring>(i->second).empty()))
          sortname = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_NEW_ITEM].id);
      new_item = boost::get<bool> (i->second);

      i = row.find (album_attributes[0].id);
      tracks = boost::get<int> (i->second);
    }

    // Track 
#ifdef HAVE_HAL
    Track::Track (const Bmp::DB::Row& row, bool use_hal)
#else
    Track::Track (const Bmp::DB::Row& row)
#endif
    {
      using namespace Glib; 
      using namespace Bmp::DB;

      Row::const_iterator i, end = row.end ();

      i = row.find (track_attributes[DATUM_ARTIST].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          artist = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_ALBUM].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          album = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_TITLE].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          title = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_GENRE].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          genre = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_COMMENT].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          comment = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_TRACK].id);
      if (i != end)
            tracknumber = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_TIME].id);
      if (i != end)
            duration = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_RATING].id);
      if (i != end)
            rating = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_COUNT].id);
      if (i != end)
            count = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_DATE].id);
      if (i != end)
            date = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_MTIME].id);
      if (i != end)
            mtime = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_BITRATE].id);
      if (i != end)
            bitrate = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_SAMPLERATE].id);
      if (i != end)
            samplerate = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_COUNT].id);
      if (i != end)
            samplerate = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_ASIN].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          asin = boost::get<ustring> (i->second);

#ifdef HAVE_HAL
      i = row.find (track_attributes[DATUM_HAL_VOLUME_UDI].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          volume_udi = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_HAL_DEVICE_UDI].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          device_udi = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_VOLUME_RELATIVE_PATH].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          volume_relative_path = boost::get<ustring> (i->second);
      
      if (use_hal)
        {
          try {
            std::string mount_path = hal->get_mount_path_for_volume_udi (volume_udi.get().c_str());
            location = Glib::filename_to_uri (Glib::build_filename (mount_path.c_str(), volume_relative_path.get().c_str()));
            }
          catch (...) {}
        }
      else
        {
          i = row.find (track_attributes[DATUM_LOCATION].id);
          if ((i != end) && !(boost::get<ustring>(i->second).empty()))
              location = boost::get<ustring> (i->second);
        }
#else
      i = row.find (track_attributes[DATUM_LOCATION].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          location = boost::get<ustring> (i->second);
#endif
      i = row.find (track_attributes[DATUM_MB_ARTIST_SORTNAME].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          mb_album_artist_sort_name = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_MB_ALBUM_ARTIST_ID].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          mb_album_artist_id = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          mb_album_artist_sort_name = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_MB_TRACK_ID].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          mb_track_id = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_MB_ALBUM_ID].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          mb_album_id = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_MB_ARTIST_ID].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          mb_artist_id = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_MUSICIP_PUID].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          puid = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_HASH].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          hash = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_NEW_ITEM].id);
      if (i != end)
        new_item = boost::get<bool> (i->second);

    }

#if 0
    void print_album (Album const& a)
    {
      std::cerr << "Artist: " << (a.artist ? a.artist.get() : "");
      std::cerr << "\nAlbum: " << (a.album ? a.album.get() : "");
      std::cerr << "\nReleaseID: " << (a.release_id ? a.release_id.get() : "");
      std::cerr << "\nArtistID: " << (a.artist_id ? a.artist_id.get() : "");
      std::cerr << "\nASIN: " << (a.asin ? a.asin.get() : "");
      std::cerr << "\nNewItem: " << (a.new_item ? a.new_item.get() : false);
      std::cerr << "\nSortname: " << (a.sortname ? a.sortname.get() : "");
    }
#endif

    bool operator== (Album const& a, Album const &b)
    {
      int truth = 0;
      int mask  = 0;

      enum ComparisonMask
      {
        NONE          = 0,
        M_ARTIST      = 1 << 0,
        M_ALBUM       = 1 << 1,
        M_RELEASE_ID  = 1 << 2,
        M_ARTIST_ID   = 1 << 3,
        M_ASIN        = 1 << 4,
        M_SORTNAME    = 1 << 5,
      };

      if (a.artist && b.artist && !(a.sortname || b.sortname))
        {
          mask |= M_ARTIST;
          if (a.artist == b.artist) truth |= M_ARTIST;
        }

      if (a.album && b.album)
        {
          mask |= M_ALBUM;
          if (a.album == b.album) truth |= M_ALBUM;
        }

      if (a.release_id && b.release_id)
        {
          mask |= M_RELEASE_ID;
          if (a.release_id == b.release_id) truth |= M_RELEASE_ID;
        }

      if (a.artist_id && b.artist_id)
        {
          mask |= M_ARTIST_ID;
          if (a.artist_id == b.artist_id) truth |= M_ARTIST_ID;
        }

      if (a.asin && b.asin)
        {
          mask |= M_ASIN;
          if (a.asin == b.asin) truth |= M_ASIN;
        }

      if (a.sortname && b.sortname)
        {
          mask |= M_SORTNAME;
          if (a.sortname == b.sortname) truth |= M_SORTNAME;
        }

#if 0
      std::cerr << "Album A:\n";
      print_album (a);
      std::cerr << "\n\nAlbum B:\n";
      print_album (b);
      std::cerr << "\nmask: '" << std::hex << mask << "', truth: '" << std::hex << truth << "'\n\n\n";
#endif

      return (mask == truth); 
    } 

    void album_to_attrs (Bmp::Library::Album const& album, Bmp::DB::AttributeV& attrs)
    {
      using namespace Bmp::DB;

      if (album.sortname)
        {
          attrs.push_back (Attribute (album.sortname ? EXACT : IS_NULL,
                           Bmp::MetadatumId (DATUM_MB_ALBUM_ARTIST_SORTNAME),
                           album.sortname ? album.sortname.get() : ""));
        }
      else
        {
          attrs.push_back (Attribute (album.artist ? EXACT : IS_NULL,
                           Bmp::MetadatumId (DATUM_ARTIST),
                           album.artist ? album.artist.get() : ""));
        }

      attrs.push_back (Attribute (album.album ? EXACT : IS_NULL,
                        Bmp::MetadatumId (DATUM_ALBUM),
                        album.album ? album.album.get() : ""));

      attrs.push_back (Attribute (album.asin ? EXACT : IS_NULL,
                        Bmp::MetadatumId (DATUM_ASIN),
                        album.asin ? album.asin.get() : ""));

      attrs.push_back (Attribute (album.release_id ? EXACT : IS_NULL,
                        Bmp::MetadatumId (DATUM_MB_ALBUM_ID),
                        album.release_id ? album.release_id.get() : ""));

      attrs.push_back (Attribute (album.artist_id ? EXACT : IS_NULL,
                        Bmp::MetadatumId (DATUM_MB_ALBUM_ARTIST_ID),
                        album.artist_id ? album.artist_id.get() : ""));
    }
  } //!Library::

  std::string MetadatumId (Bmp::Library::Datum datum)
  {
      return Bmp::Library::Library::metadatum_id (datum);
  }

} //!Bmp::
