//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License Version 2
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>
#include <glibmm.h>
#include <glib/gi18n.h>

#include <gst/gst.h>
#include <gst/gstelement.h>
#include <gst/interfaces/mixer.h>
#include <gst/interfaces/mixertrack.h>
#include <gst/interfaces/mixeroptions.h>

#include <boost/shared_ptr.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/format.hpp>
#include <boost/optional.hpp>

#include "main.hh"
#include "uri++.hh"
#include "util.hh"
#include "play.hh"
#include "debug.hh"

#include "dialog-gsterror.hh"

#include "x_service_core.hh"

#define BMP_GST_BUFFER_TIME   ((gint64) 50000)
#define BMP_GST_PLAY_TIMEOUT  ((gint64) 3000000000)

namespace
{
  const char *NAME_DECODER  = _("Stream Decoder");
  const char *NAME_CONVERT  = _("Raw Audio Format Converter");
  const char *NAME_QUEUE    = _("Data Buffer (Queue)");
  const char *NAME_VOLUME   = _("Volume Adjustment");
  const char *NAME_RESAMPLE = _("Resampler");
}

namespace Bmp
{
    Play::Play () 
        : Glib::ObjectBase (typeid(Play)),
          property_stream_        (*this, "stream", ""),
          property_lastfm_mode_   (*this, "lastfm-mode", false),
          property_volume_        (*this, "volume", 50),
          property_position_      (*this, "position", 0),
          property_status_        (*this, "playstatus", PLAYSTATUS_STOPPED),
          property_sane_          (*this, "sane", false),
          property_length_        (*this, "length", 0), 
          property_has_cdda_      (*this, "has_cdda", false), 
          property_has_http_      (*this, "has_http", false), 
          property_has_mms_       (*this, "has_mms", false), 
          current_pipe            (P_FILE),
          source                  (sigc::connection())
    {
      property_volume() = mcs->key_get<int>("bmp", "volume"); 

      message_domain_id = core->message_domain_register ("play", _("BMP: Audio"));

      property_volume().signal_changed().connect(sigc::mem_fun(*this, &Bmp::Play::on_volume_changed));
      property_stream().signal_changed().connect(sigc::mem_fun(*this, &Bmp::Play::on_stream_changed));
      property_lastfm_mode().signal_changed().connect(sigc::mem_fun(*this, &Bmp::Play::on_lastfm_mode_changed));
      property_status().signal_changed().connect(sigc::mem_fun(*this, &Bmp::Play::on_playstatus_changed));

      for (int n = 0; n < N_ELEMENTS; n++)
        {
          elements[P_FILE][n] = 0;
          elements[P_HTTP][n] = 0;
          elements[P_MMS][n]  = 0;
          elements[P_CDDA][n] = 0;
        }

      reset ();
    }

    //dtor
    Play::~Play ()
    {
      property_status () = PLAYSTATUS_STOPPED;
      pipelines_destroy ();
    }

    gboolean
    Play::foreach_structure (GQuark     field_id,
                 const GValue  *value,
                  gpointer      data)
    {
      Bmp::Play *play;
      play = reinterpret_cast<Bmp::Play*>(data);

      if (field_id == g_quark_from_string ("rate"))
        {
          int samplerate = g_value_get_int (value);
          if (samplerate) play->signal_samplerate_.emit (samplerate);
          return FALSE;
        }
      return TRUE;
    }

    bool
    Play::timeout_handler ()
    {
      GstQuery       *query;
      GstFormat       format = GST_FORMAT_TIME;
      gint64          time_in_nanoseconds;
      gint            time_in_seconds;

      if (!(GST_STATE (elements[current_pipe][PIPELINE]) == GST_STATE_PLAYING))
        {
          return false;
        }

      query = gst_query_new_position (format);
      gst_element_query (elements[current_pipe][SINK], query);
      gst_query_parse_position (query, &format, &time_in_nanoseconds);
      time_in_seconds = time_in_nanoseconds / GST_SECOND;
      gst_query_unref (query);

      if ((time_in_seconds >= 0) && (GST_STATE(elements[current_pipe][PIPELINE]) >= GST_STATE_PLAYING))
        {
          signal_position_.emit( time_in_seconds );
          return true;
        }
      else
        {
          signal_position_.emit( 0 );
          return false;
        }

      return false;
    }

    bool
    Play::is_audio_file (std::string const& str)
    {
      char* const extensions[] =
      { "mp3",
        "aac",
        "mp4",
        "m4a",
        "m4b",
        "m4p",
        "m4v",
        "mp4v",
        "flac",
        "wma",
        "asf",
        "sid",
        "psid",
        "mod",
        "oct",
        "xm",
        "669",
        "sht", 
        "mpc",
        "ogg",
        0 };

      return Bmp::Util::str_has_suffixes_nocase (str.c_str(), extensions); 
    }

    void 
    Play::stop_stream ()
    {
      gst_element_set_state (elements[current_pipe][PIPELINE], GST_STATE_NULL);
      source.disconnect(); 
      signal_position_.emit( 0 );
      signal_bitrate_.emit( 0 );
      signal_samplerate_.emit( 0 );
    }
    
    void 
    Play::play_stream ()
    {
      Bmp::URI uri (property_stream().get_value());

        if (((uri.get_protocol() == Bmp::URI::PROTOCOL_MMS)  ||
             (uri.get_protocol() == Bmp::URI::PROTOCOL_MMSU) ||
             (uri.get_protocol() == Bmp::URI::PROTOCOL_MMST)) && (!elements[P_MMS][PIPELINE]))
          {
            property_status() = PLAYSTATUS_STOPPED;
            return;
          }
        else
        if ((uri.get_protocol() == Bmp::URI::PROTOCOL_HTTP) && (!elements[P_HTTP][PIPELINE]))
          {
            property_status() = PLAYSTATUS_STOPPED;
            return;
          }
        else 
        if ((uri.get_protocol() == Bmp::URI::PROTOCOL_CDDA) && (!elements[P_CDDA][PIPELINE]))
          {
            property_status() = PLAYSTATUS_STOPPED;
            return;
          }
        
        g_object_set (elements[current_pipe][VOLUME], "volume", property_volume().get_value()/100. , NULL);

        if (gst_element_set_state (elements[current_pipe][PIPELINE], GST_STATE_PLAYING) == GST_STATE_CHANGE_FAILURE)
          {
            gst_element_set_state (elements[current_pipe][PIPELINE], GST_STATE_READY);
            return;
          }
    }
    
    void 
    Play::pause_stream ()
    {
        if (GST_STATE (elements[current_pipe][PIPELINE]) == GST_STATE_PAUSED)
          {
            property_status() = PLAYSTATUS_PLAYING;
          }
        
        gst_element_set_state (elements[current_pipe][PIPELINE], GST_STATE_PAUSED);
    }

    void
    Play::on_volume_changed () 
    {
      gdouble _volume;

      _volume = property_volume().get_value()/100.; 
      g_object_set (elements[current_pipe][VOLUME], "volume", _volume, NULL);
    }

    void
    Play::on_lastfm_mode_changed () 
    {
      if (GST_IS_ELEMENT (elements[P_HTTP][SRC]))
        {
          g_object_set (G_OBJECT (elements[P_HTTP][SRC]), "lastfm-mode", property_lastfm_mode_.get_value() ? TRUE : FALSE, NULL);
        }
    }

    void
    Play::on_stream_changed () 
    {
      if (property_sane().get_value() == false) return;

      Bmp::URI uri (property_stream().get_value());

      switch (uri.get_protocol())
      {
          case Bmp::URI::PROTOCOL_FILE:
          { 
            current_pipe = P_FILE;
            try
              {
                g_object_set (elements[current_pipe][SRC], 
                              "location",
                              Glib::filename_from_uri (property_stream().get_value()).c_str(), 
                              NULL);
              }
            catch (Glib::ConvertError& cxe)
               {
                boost::format format ("Unable to play stream '%s'");
                core->message_dispatch (message_domain_id,
                                        Gtk::MESSAGE_ERROR,
                                        (format % property_stream().get_value()).str()); 
               }
            break;
          }
  
          case Bmp::URI::PROTOCOL_MMS:
          case Bmp::URI::PROTOCOL_MMSU:
          case Bmp::URI::PROTOCOL_MMST:
          {
            current_pipe = P_MMS;
            g_object_set (elements[current_pipe][SRC], "location", property_stream().get_value().c_str(), NULL);
            break;
          }
 
          case Bmp::URI::PROTOCOL_HTTP:
          {
            gst_element_set_state (elements[P_HTTP][PIPELINE], GST_STATE_NULL);
            current_pipe = P_HTTP;
            pipeline_create (P_HTTP);
            g_object_set (elements[current_pipe][SRC], "lastfm-mode", gboolean(property_lastfm_mode().get_value()), NULL);
            g_object_set (elements[current_pipe][SRC], "location", property_stream().get_value().c_str(), NULL);
            break;
          }
           
          case Bmp::URI::PROTOCOL_CDDA:
          {
            unsigned int track;
            current_pipe = P_CDDA;
            track = (unsigned int)atoi((uri.path.c_str())+1);
            g_object_set (elements[current_pipe][SRC], "track", track, NULL);
            break;
          }

          default: return;
      }
    }

    void
    Play::on_playstatus_changed () 
    {
      if (property_status().get_value() & PLAYSTATUS_SEEKING)
        { 
          pause_stream (); 
          return;
        }

      if (property_status().get_value() & PLAYSTATUS_STOPPED)
        {
          stop_stream ();
          return;
        }

      if (property_status().get_value() & PLAYSTATUS_WAITING)
        {
          stop_stream ();
          return;
        }

      if (property_status().get_value() & PLAYSTATUS_PAUSED)
        {
          pause_stream (); 
          return;
        }

      if (property_status().get_value() & PLAYSTATUS_PLAYING)
        {
          play_stream ();
          return;
        }
    }

    void
    Play::seek (int position) 
    {
      GstEvent   *seek_event;
      guint64     seek_offset;

      seek_offset = (position * GST_SECOND);
      seek_event = gst_event_new_seek (1.0, GST_FORMAT_TIME, GST_SEEK_FLAG_FLUSH,
            GST_SEEK_TYPE_SET, seek_offset, GST_SEEK_TYPE_NONE, 0);

      if (gst_element_send_event (elements[current_pipe][PIPELINE], seek_event))
        {
          gst_pipeline_set_new_stream_time (GST_PIPELINE(elements[current_pipe][PIPELINE]), 0);
          gst_element_get_state (GST_ELEMENT(elements[current_pipe][PIPELINE]), 0, 0, 150 * GST_MSECOND);
          signal_seek_.emit(position);
        }
    }

    void
    Play::pipelines_destroy ()    
    {
      if (elements[P_FILE][PIPELINE])
        {
          gst_element_set_state (elements[P_FILE][PIPELINE], GST_STATE_NULL);
          gst_object_unref (elements[P_FILE][PIPELINE]);
        }

      if (elements[P_HTTP][PIPELINE])
        {
          gst_element_set_state (elements[P_HTTP][PIPELINE], GST_STATE_NULL);
          gst_object_unref (elements[P_HTTP][PIPELINE]);
        }

      if (elements[P_MMS][PIPELINE])
        {
          gst_element_set_state (elements[P_MMS][PIPELINE], GST_STATE_NULL);
          gst_object_unref (elements[P_MMS][PIPELINE]);
        }

      if (elements[P_CDDA][PIPELINE])
        {
          gst_element_set_state (elements[P_CDDA][PIPELINE], GST_STATE_NULL);
          gst_object_unref (elements[P_CDDA][PIPELINE]);
        }

      for (int n = 0; n < N_ELEMENTS; n++)
        {
          elements[P_FILE][n] = 0;
          elements[P_HTTP][n] = 0;
          elements[P_CDDA][n] = 0;
          elements[P_MMS][n] = 0;
        }
    }

    void
    Play::link_pad (GstElement *element,
                    GstPad     *pad,
                    gboolean    last,
                    gpointer    data) 
    {
      Bmp::Play   *play;
      GstPad      *sinkpad;

      play = reinterpret_cast<Bmp::Play*>(data);
      sinkpad = gst_element_get_pad (play->elements[play->current_pipe][CONVERT], "sink");
      gst_pad_link (pad, sinkpad);
    }

    bool
    Play::adjust_queue ()
    {
      g_object_set (G_OBJECT (elements[P_HTTP][QUEUE]),
                        "min-threshold-bytes", (unsigned int)(16 * 1024), NULL);
      return false;
    }

    gboolean
    Play::bus_watch (GstBus     *bus,
                     GstMessage *message,
                     gpointer    data)
    {
        static GstState old_state=GstState (0), new_state=GstState(0), pending_state=GstState(0);
        Bmp::Play *play = reinterpret_cast<Bmp::Play*>(data);
        
        switch (GST_MESSAGE_TYPE (message))
          {
            case GST_MESSAGE_APPLICATION:
              {
                const GstStructure *s = gst_message_get_structure (message);

                debug ("play", "Got GST_MESSAGE_APPLICATION: %s", gst_structure_get_name (s));
                  
                if (std::string(gst_structure_get_name (s)) == "lastfm-sync")
                  {
                    play->signal_lastfm_sync_.emit ();
                  }
                else
                if (std::string(gst_structure_get_name (s)) == "lastfm-status")
                  {
                    int status;
                    gst_structure_get_int (s, "status", &status);
                    play->signal_http_status_.emit (status);
                  }
                else
                if (std::string(gst_structure_get_name (s)) == "buffering")
                  {
                    double size;
                    gst_structure_get_double (s, "size", &size);
                    play->signal_buffering_.emit (size);
                  }
                else
                if (std::string(gst_structure_get_name (s)) == "buffering-done")
                  {
                    Glib::signal_idle().connect(sigc::mem_fun (play, &Bmp::Play::adjust_queue));
                  }
                break;
              }

            case GST_MESSAGE_TAG:
              {
                if (play->property_lastfm_mode_.get_value()) break;

                debug ("play", "Got GST_MESSAGE_TAG");

                if ((old_state == GST_STATE_PLAYING) && (new_state < GST_STATE_PLAYING)) break; 

                GstTagList *tag_list;
                gst_message_parse_tag (message, &tag_list);
                if (tag_list)
                  {
                    char *title = 0;
                    unsigned int bitrate = 0;

                    if (gst_tag_list_get_string (tag_list, GST_TAG_TITLE, &title));
                    {
                        if ((title) && (play->current_pipe == P_HTTP))
                        {
                          play->signal_title_.emit ((const char*)title);
                          g_free (title);
                        }
                    }

                    if (gst_tag_list_get_uint (tag_list, GST_TAG_BITRATE, &bitrate));
                    {
                      if (bitrate)
                      {
                        play->signal_bitrate_.emit (bitrate/1000);
                      }
                    }

                  }
                break;
          }

          case GST_MESSAGE_STATE_CHANGED:
            {
              debug ("play", "Got GST_MESSAGE_STATE_CHANGED");

              gst_message_parse_state_changed  (message, &old_state, &new_state, &pending_state);

              if (new_state == GST_STATE_PLAYING)
                {
                  if (message->src == GST_OBJECT(play->elements[play->current_pipe][PIPELINE]))
                    {
                      if (!play->source.connected())
                        {
                          play->source = Glib::signal_timeout().connect
                                                (sigc::mem_fun(play, &Bmp::Play::timeout_handler), 500);
                        }

                      GstPad *pad = gst_element_get_pad (play->elements[play->current_pipe][SINK], "sink");
                      GstCaps *caps = gst_pad_get_negotiated_caps (pad);
                      gst_object_unref (pad);

                      for (unsigned int n = 0; n < gst_caps_get_size (caps); ++n)
                        {
                          gst_structure_foreach (gst_caps_get_structure (caps, n),
                                                 GstStructureForeachFunc (Bmp::Play::foreach_structure),
                                                 play);
                        }
                    }
              }
              break;
            }

            case GST_MESSAGE_ERROR:
            {
                GstElement    *element;
                GError        *error = 0;
                Glib::ustring  main, name, location;
                char          *debug_string;

                element = GST_ELEMENT(GST_MESSAGE_SRC (message));
                name = gst_object_get_name (GST_OBJECT(element));
                location = play->property_stream().get_value(); 

                gst_message_parse_error (message,
                                        &error,
                                        &debug_string);
        
                if (error->domain == GST_CORE_ERROR)
                  {
                            switch (error->code)
                                {
                                  case GST_CORE_ERROR_MISSING_PLUGIN:
                                  {
                                    main += "1.01  [No plugin available to play stream]";
                                    break;
                                  }
        
                                  case GST_CORE_ERROR_SEEK:
                                  {
                                    main += "1.02  [Error during Seek (core)]";
                                    break;
                                  }
        
                                  case GST_CORE_ERROR_STATE_CHANGE:
                                  {
                                        struct StringPairs
                                        {
                                            Glib::ustring   state_pre;
                                            Glib::ustring   state_post;
                                        };
        
                                        static StringPairs string_pairs[] =
                                        {
                                            {"0",         "READY"},
                                            {"READY",     "PAUSED"},
                                            {"PAUSED",    "PLAYING"},
                                            {"PLAYING",   "PAUSED"},
                                            {"PAUSED",    "READY"},
                                            {"READY",     "0"},
                                        };
       
                                    main += "1.03  [Error occured during state change from "+string_pairs[GST_STATE_PENDING(message->src)].state_pre+" to "
                                                        +string_pairs[GST_STATE_PENDING(message->src)].state_post+"]"; 
                                    break;
                                  }
        
                                  case GST_CORE_ERROR_PAD:
                                  {
                                    main += "1.04  [Error while performing pad linking]";
                                    break;
                                  }
        
                                  case GST_CORE_ERROR_NEGOTIATION:
                                  {
                                    main =+ "1.05  [Error occured during element(s) negotiation]";
                                    break;
                                  }
        
                                    default: break;
                                }
                  }
                else if (error->domain == GST_RESOURCE_ERROR)
                  {
                            switch (error->code)
                            {
        
                                  case GST_RESOURCE_ERROR_SEEK:
                                  {
                                    main += "2.01  [Error during Seek (resource; end of stream?)]";
                                    break;
                                  }
        
                                  case GST_RESOURCE_ERROR_NOT_FOUND:
                                  case GST_RESOURCE_ERROR_OPEN_READ:
                                  case GST_RESOURCE_ERROR_OPEN_READ_WRITE:
                                  case GST_RESOURCE_ERROR_READ:
                                  {
                                    main += "2.02  [Unable to Open Resource (stream)]";
                                    break;
                                  }

                                  case GST_RESOURCE_ERROR_SYNC:
                                  {
                                    main += "2.03  [Synchronization Error]";
                                    break;
                                  }

                                  case GST_RESOURCE_ERROR_BUSY:
                                  {
                                    main += "2.04  [Device is Busy]";
                                    break;
                                  }
      
                                  default: break;
                                }
                    }
                else if (error->domain == GST_STREAM_ERROR)
                    {
                            switch (error->code)
                                {
                                  case GST_STREAM_ERROR_FAILED:
                                  {
                                    main += "3.01  [Unable Play Stream (Perhaps missing plugin?)]";
                                    break;
                                  }

                                  case GST_STREAM_ERROR_TYPE_NOT_FOUND:
                                  {
                                    main += "3.02  [Unable to Determine Stream Type (invalid stream?)\n";
                                    break;
                                  }

                                  case GST_STREAM_ERROR_WRONG_TYPE:
                                  {
                                    main += "3.03  [Element Unable to Handle this Stream Type]";
                                    break;
                                  }

                                  case GST_STREAM_ERROR_CODEC_NOT_FOUND:
                                  {
                                    main += "3.04  [No Codec installed to Handle this Stream]";
                                    break;
                                  }
      
                                  case GST_STREAM_ERROR_DECODE:
                                  {
                                    main += "3.05  [Error Decoding the Stream]";
                                    break;
                                  }
      
                                  case GST_STREAM_ERROR_DEMUX:
                                  {
                                    main += "3.06  [Error Demultiplexing the Stream]";
                                    break;
                                  }
      
                                  case GST_STREAM_ERROR_FORMAT:
                                  {
                                    main += "3.07  [Error Parsing the Stream Format]";
                                    break;
                                  }
    
                                default: break;
                                }
                    }
        
                g_error_free (error);

                Glib::signal_idle().connect (sigc::mem_fun (play, &Bmp::Play::reset_idle));

                Bmp::DialogGSTError *dialog = Bmp::DialogGSTError::create ();
                dialog->run (main, name, location, debug_string);
                g_free (debug_string);
                delete dialog;

                break;
            }
        
            case GST_MESSAGE_WARNING:
              {
                GError  *error = 0;
                gchar   *debug;
        
                gst_message_parse_warning (message, &error, &debug);
                g_print ("%s WARNING: %s (%s)\n", G_STRLOC, error->message, debug);
                g_error_free (error);
                break;
              }
        
            case GST_MESSAGE_EOS:
              {
                play->signal_eos_.emit();
                break;
              }
        
            default: break;
          }
        
          return TRUE;
    }


    void
    Play::pipeline_create (Pipeline pipeline)
    {
      boost::optional<std::string> device, propname;
      std::string                  sink;
                                  
      guint64 buffer_time_value;
      bool    buffer_time_set;

      sink = mcs->key_get<std::string>( "audio", "sink");

      buffer_time_set = false;
      buffer_time_value = 200000;

      if (sink == "osssink")
        {
          buffer_time_set = true;
          buffer_time_value = mcs->key_get<int>("audio", "oss-buffer-time");
          propname = "device";
          device = mcs->key_get<std::string>( "audio", "device-oss");
        }
      else
      if (sink == "esdsink")
        {
          buffer_time_set = true;
          buffer_time_value = mcs->key_get<int>("audio", "esd-buffer-time");
          propname = "host";
          device = mcs->key_get<std::string>( "audio", "device-esd");
        }

#ifdef HAVE_ALSA
      else
      if (sink == "alsasink")
        {
          buffer_time_set = true;
          buffer_time_value = mcs->key_get<int>("audio", "alsa-buffer-time");
          propname = "device";
          device = mcs->key_get<std::string>( "audio", "device-alsa");
        }
#endif

#ifdef HAVE_SUN
      else
      if (sink == "sunaudiosink")
        {
          buffer_time_set = true;
          buffer_time_value = mcs->key_get<int>("audio", "sun-buffer-time");
          propname = "device";
          device = mcs->key_get<std::string>( "audio", "device-sun");
        }
#endif

#ifdef HAVE_HAL
      else
      if (sink == "halaudiosink")
        {
          buffer_time_set = false;
        }
#endif

      switch (pipeline)
      {
        case P_FILE:
        {
          elements[P_FILE][SRC]       = gst_element_factory_make ("filesrc", "src"); 
          elements[P_FILE][DECODER]   = gst_element_factory_make ("decodebin", NAME_DECODER); 
          elements[P_FILE][CONVERT]   = gst_element_factory_make ("audioconvert", NAME_CONVERT); 
          elements[P_FILE][VOLUME]    = gst_element_factory_make ("volume", NAME_VOLUME); 
          elements[P_FILE][RESAMPLE]  = gst_element_factory_make ("audioresample", NAME_RESAMPLE); 
          elements[P_FILE][SINK]      = gst_element_factory_make (sink.c_str(), "sink"); 

          if (propname && device)          
            {
              g_object_set (G_OBJECT(elements[P_FILE][SINK]), propname.get().c_str(), device.get().c_str(), NULL);
            }
          
          if (buffer_time_set)
            {
              g_object_set (G_OBJECT(elements[P_FILE][SINK]), "buffer-time", buffer_time_value, NULL);
            }
          
          elements[P_FILE][PIPELINE] = gst_pipeline_new ("pipeline-file");
          
          gst_bin_add_many (GST_BIN(elements[P_FILE][PIPELINE]),
                                    elements[P_FILE][SRC],
                                    elements[P_FILE][DECODER],
                                    elements[P_FILE][CONVERT],
                                    elements[P_FILE][VOLUME],
                                    elements[P_FILE][RESAMPLE],
                                    elements[P_FILE][SINK],
                                    NULL);
          
          gst_element_link_many (elements[P_FILE][SRC],
                                 elements[P_FILE][DECODER],
                                 NULL);
          
          gst_element_link_many (elements[P_FILE][CONVERT],
                                 elements[P_FILE][VOLUME],
                                 elements[P_FILE][RESAMPLE],
                                 elements[P_FILE][SINK],
                                 NULL);
      
          g_signal_connect (G_OBJECT(elements[P_FILE][DECODER]), "new-decoded-pad", G_CALLBACK (Bmp::Play::link_pad), gpointer(this)); 
          gst_bus_add_watch (gst_pipeline_get_bus (GST_PIPELINE(elements[P_FILE][PIPELINE])), GstBusFunc(Bmp::Play::bus_watch), gpointer(this));
          break;
        }

      case P_HTTP: 
        {
          elements[P_HTTP][SRC] = gst_element_factory_make ("bmpx-neonhttpsrc", "src"); 

          if (!elements[P_HTTP][SRC])
            {
              g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%s: No HTTP support available!", G_STRLOC);
              elements[P_HTTP][PIPELINE] = 0;
              return; 
            }

          property_has_http_ = true;

          g_object_set (G_OBJECT (elements[P_HTTP][SRC]),
                        "iradio-mode",
                         property_lastfm_mode_.get_value() ? FALSE : TRUE,
                        "lastfm-mode",
                         property_lastfm_mode_.get_value() ? TRUE : FALSE,
                         NULL);

          elements[P_HTTP][DECODER]  = gst_element_factory_make ("decodebin", NAME_DECODER); 
          elements[P_HTTP][CONVERT]  = gst_element_factory_make ("audioconvert", NAME_CONVERT); 
          elements[P_HTTP][QUEUE]    = gst_element_factory_make ("queue", NAME_QUEUE); 
          elements[P_HTTP][VOLUME]   = gst_element_factory_make ("volume", NAME_VOLUME); 
          elements[P_HTTP][RESAMPLE] = gst_element_factory_make ("audioresample", NAME_RESAMPLE); 
          elements[P_HTTP][SINK]     = gst_element_factory_make (sink.c_str(), "sink"); 
             
          if (propname && device)
            {
              g_object_set (G_OBJECT(elements[P_HTTP][SINK]), propname.get().c_str(), device.get().c_str(), NULL);
            }
          
          if (buffer_time_set)
            {
              g_object_set (G_OBJECT(elements[P_HTTP][SINK]), "buffer-time", buffer_time_value, NULL);
            }

          g_object_set (G_OBJECT (elements[P_HTTP][QUEUE]),
                        "max-size-bytes", (unsigned int)(256 * 1024),
                        "min-threshold-bytes", (unsigned int)(32 * 1024),
                        "leaky", int(0), NULL);

          elements[P_HTTP][PIPELINE] = gst_pipeline_new ("pipeline-http");

          gst_bin_add_many (GST_BIN (elements[P_HTTP][PIPELINE]),
                                     elements[P_HTTP][SRC],
                                     elements[P_HTTP][DECODER],
                                     elements[P_HTTP][CONVERT],
                                     elements[P_HTTP][QUEUE],
                                     elements[P_HTTP][VOLUME],
                                     elements[P_HTTP][RESAMPLE],
                                     elements[P_HTTP][SINK],
                                     NULL);

          gst_element_link_many (elements[P_HTTP][SRC],
                                 elements[P_HTTP][DECODER],
                                 NULL);
          
          gst_element_link_many (elements[P_HTTP][CONVERT],
                                 elements[P_HTTP][QUEUE],
                                 elements[P_HTTP][VOLUME],
                                 elements[P_HTTP][RESAMPLE],
                                 elements[P_HTTP][SINK],
                                 NULL);
          g_signal_connect (G_OBJECT(elements[P_HTTP][DECODER]), "new-decoded-pad", G_CALLBACK (Bmp::Play::link_pad), gpointer(this)); 
          gst_bus_add_watch (gst_pipeline_get_bus (GST_PIPELINE(elements[P_HTTP][PIPELINE])), GstBusFunc(Bmp::Play::bus_watch), gpointer(this));
          break;
        }

      case P_MMS:
        {
          elements[P_MMS][SRC] = gst_element_factory_make ("mmssrc", "src"); 

          if (!elements[P_MMS][SRC])
            {
              g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%s: No MMS support available!", G_STRLOC);
              elements[P_MMS][PIPELINE] = 0;
              return; 
            }

          property_has_mms_ = true;

          elements[P_MMS][DECODER]  = gst_element_factory_make ("decodebin", NAME_DECODER); 
          elements[P_MMS][CONVERT]  = gst_element_factory_make ("audioconvert", NAME_CONVERT); 
          elements[P_MMS][VOLUME]   = gst_element_factory_make ("volume", NAME_VOLUME);
          elements[P_MMS][RESAMPLE] = gst_element_factory_make ("audioresample", NAME_RESAMPLE); 
          elements[P_MMS][SINK]     = gst_element_factory_make (sink.c_str(), "sink"); 
          
          if (propname && device)
            {
              g_object_set (G_OBJECT(elements[P_MMS][SINK]), propname.get().c_str(), device.get().c_str(), NULL);
            }
          
          if (buffer_time_set)
            {
              g_object_set (G_OBJECT(elements[P_MMS][SINK]), "buffer-time", buffer_time_value, NULL);
            }
          
          elements[P_MMS][PIPELINE] = gst_pipeline_new ("pipeline-mms");
          
          gst_bin_add_many (GST_BIN (elements[P_MMS][PIPELINE]),
                                     elements[P_MMS][SRC],
                                     elements[P_MMS][DECODER],
                                     elements[P_MMS][CONVERT],
                                     elements[P_MMS][VOLUME],
                                     elements[P_MMS][RESAMPLE],
                                     elements[P_MMS][SINK],
                                     NULL);
      
          gst_element_link_many (elements[P_MMS][SRC],
                                 elements[P_MMS][DECODER],
                                 NULL);
          
          gst_element_link_many (elements[P_MMS][CONVERT],
                                 elements[P_MMS][VOLUME],
                                 elements[P_MMS][RESAMPLE],
                                 elements[P_MMS][SINK],
                                 NULL);

          g_signal_connect (G_OBJECT(elements[P_MMS][DECODER]), "new-decoded-pad", G_CALLBACK (Bmp::Play::link_pad), gpointer(this)); 
          gst_bus_add_watch (gst_pipeline_get_bus (GST_PIPELINE(elements[P_MMS][PIPELINE])), GstBusFunc(Bmp::Play::bus_watch), gpointer(this));
          break;
        }

      case P_CDDA:
        {
          elements[P_CDDA][SRC] = gst_element_factory_make ("cdparanoiasrc", "src"); 
          
          if (!elements[P_CDDA][SRC])
            {
              g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%s: No CDDA support available!", G_STRLOC);
              elements[P_CDDA][PIPELINE] = 0;
              return;
            }

          property_has_cdda_ = true;
          
          elements[P_CDDA][CONVERT]   = gst_element_factory_make ("audioconvert", NAME_CONVERT); 
          elements[P_CDDA][VOLUME]    = gst_element_factory_make ("volume", NAME_VOLUME); 
          elements[P_CDDA][RESAMPLE]  = gst_element_factory_make ("audioresample", NAME_RESAMPLE); 
          elements[P_CDDA][SINK]      = gst_element_factory_make (sink.c_str(), "sink"); 
      
          if (propname && device) 
            {
              g_object_set (G_OBJECT(elements[P_CDDA][SINK]), propname.get().c_str(), device.get().c_str(), NULL);
            }
      
          if (buffer_time_set)
            {
              g_object_set (G_OBJECT(elements[P_CDDA][SINK]), "buffer-time", buffer_time_value, NULL);
            }
      
          elements[P_CDDA][PIPELINE] = gst_pipeline_new ("pipeline-cdda");
      
          gst_bin_add_many (GST_BIN (elements[P_CDDA][PIPELINE]),
                                     elements[P_CDDA][SRC],
                                     elements[P_CDDA][CONVERT],
                                     elements[P_CDDA][VOLUME],
                                     elements[P_CDDA][RESAMPLE],
                                     elements[P_CDDA][SINK],
                                     NULL);
      
          gst_element_link_many (elements[P_CDDA][SRC],
                                 elements[P_CDDA][CONVERT],
                                 elements[P_CDDA][VOLUME],
                                 elements[P_CDDA][RESAMPLE],
                                 elements[P_CDDA][SINK],
                                 NULL);

          gst_bus_add_watch (gst_pipeline_get_bus (GST_PIPELINE(elements[P_CDDA][PIPELINE])), GstBusFunc(Bmp::Play::bus_watch), gpointer(this));
          break;
        }
        default: g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "No such pipeline: %d", pipeline); break;
      }
    }

    bool
    Play::reset_idle ()
    {
      property_status() = PLAYSTATUS_STOPPED;
      reset ();
      return false;
    }

    void
    Play::reset ()
    {
      GstElement      *element;
      Glib::ustring    sink;

      pipelines_destroy ();

      // Check if the sink can be created
      sink = mcs->key_get<std::string>( "audio", "sink");
      element = gst_element_factory_make (sink.c_str(), "sinktest0");

      if (!element)
        {
          core->message_dispatch (message_domain_id,
                                  Gtk::MESSAGE_ERROR,
                                  _("BMP was not able to initialize the playback system. Please check the audio "
                                    "settings in the Preferences panel."));
          property_sane_ = false;
          return;
        }

      //FIXME: Possibly even more sanity checks
      property_sane_ = true; 

      pipeline_create (P_FILE);
      pipeline_create (P_HTTP);
      pipeline_create (P_MMS);
      pipeline_create (P_CDDA);
    }

    ///////////////////////////////////////////////
    /// Glib::Object Properties
    ///////////////////////////////////////////////

    Glib::PropertyProxy<bool>
    Play::property_lastfm_mode()
      {
        return property_lastfm_mode_.get_proxy();
      }

    Glib::PropertyProxy<Glib::ustring>
    Play::property_stream()
      {
        return property_stream_.get_proxy();
      }

    Glib::PropertyProxy<int>
    Play::property_volume()
      {
        return property_volume_.get_proxy();
      }

    Glib::PropertyProxy<int>
    Play::property_status()
      {
        return property_status_.get_proxy();
      }

    /// RO Proxies

    const Glib::PropertyProxy_ReadOnly<int>
    Play::property_position() const
      {
        GstQuery       *query;
        GstFormat       format = GST_FORMAT_TIME;
        gint64          length_in_nanoseconds;

        query = gst_query_new_position (format);
        gst_element_query (elements[current_pipe][PIPELINE], query);
        gst_query_parse_position (query, &format, &length_in_nanoseconds);
        const int value = length_in_nanoseconds / GST_SECOND;
        gst_query_unref (query);

        g_object_set (G_OBJECT (gobj()), "position", int(value), NULL);
        return Glib::PropertyProxy_ReadOnly<int>(const_cast<const Bmp::Play *>(this), "position");
      }

    const Glib::PropertyProxy_ReadOnly<bool>
    Play::property_sane() const
      {
        return Glib::PropertyProxy_ReadOnly<bool>(const_cast<const Bmp::Play *>(this), "sane");
      }

    const Glib::PropertyProxy_ReadOnly<int>
    Play::property_length() const
      {
        GstQuery       *query;
        GstFormat       format = GST_FORMAT_TIME;
        gint64          length_in_nanoseconds;

        query = gst_query_new_duration (format);
        gst_element_query (elements[current_pipe][PIPELINE], query);
        gst_query_parse_duration (query, &format, &length_in_nanoseconds);
        const int value = length_in_nanoseconds / GST_SECOND;
        gst_query_unref (query);

        g_object_set (G_OBJECT (gobj()), "length", int(value), NULL);
        return Glib::PropertyProxy_ReadOnly<int>(const_cast<const Bmp::Play *>(this), "length");
      }

    const Glib::PropertyProxy_ReadOnly<bool>
    Play::property_has_cdda() const
    {
        return Glib::PropertyProxy_ReadOnly<bool>(const_cast<const Bmp::Play *>(this), "has_cdda");
    }

    const Glib::PropertyProxy_ReadOnly<bool>
    Play::property_has_http() const
    {
        return Glib::PropertyProxy_ReadOnly<bool>(const_cast<const Bmp::Play *>(this), "has_http");
    }

    const Glib::PropertyProxy_ReadOnly<bool>
    Play::property_has_mms() const
    {
        return Glib::PropertyProxy_ReadOnly<bool>(const_cast<const Bmp::Play *>(this), "has_mms");
    }

    // Signals
    Play::SignalSamplerate&
    Play::signal_samplerate ()
    {
      return signal_samplerate_;
    }

    Play::SignalBitrate&
    Play::signal_bitrate ()
    {
      return signal_bitrate_;
    }

    Play::SignalTitle&
    Play::signal_title ()
    {
      return signal_title_;
    }

    Play::SignalEos&
    Play::signal_eos ()
    { 
      return signal_eos_;
    }

    Play::SignalSeek&
    Play::signal_seek ()
    {
      return signal_seek_;
    }

    Play::SignalPosition&
    Play::signal_position ()
    {
      return signal_position_;
    }

    Play::SignalHttpStatus&
    Play::signal_http_status ()
    {
      return signal_http_status_;
    }

    Play::SignalLastFMSync&
    Play::signal_lastfm_sync ()
    {
      return signal_lastfm_sync_;
    }

    Play::SignalBuffering&
    Play::signal_buffering ()
    {
      return signal_buffering_;
    }
};
