//  BMPx - The Dumb Music Player
//  Copyright (C) 2005-2006 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non GPL-compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <revision.h>
#include <build.h>

#include <utility>

#include <glibmm.h>
#include <glib/gi18n.h>
#include <gtkmm.h>
#include <libglademm.h>

#include <boost/format.hpp>
#include <boost/algorithm/string.hpp>

#ifdef HAVE_ALSA
#  define ALSA_PCM_NEW_HW_PARAMS_API
#  define ALSA_PCM_NEW_SW_PARAMS_API

#  include <alsa/global.h>
#  include <alsa/asoundlib.h>
#  include <alsa/pcm_plugin.h>
#  include <alsa/control.h>
#endif

#include <ne_basic.h>
#include <ne_compress.h>
#include <ne_socket.h>
#include <ne_utils.h>
#include <ne_props.h>
#include <ne_session.h>
#include <ne_request.h>
#include <ne_uri.h>

#include <mcs/mcs.h>

#include "main.hh"
#include "paths.hh"
#include "network.hh"

#include "stock.hh"
#include "actions.hh"

#ifdef HAVE_HAL
#  include <libhal.h>
#  include <libhal-storage.h>
#  include "x_hal.hh"
#endif

#include "x_library.hh"
#include "x_play.hh"
#include "x_mcsbind.hh"

#include "ui-toolbox.hh"
#include "preferences-ui.hh"

// Find some place to put these

namespace Bmp
{
  namespace
  {
    bool
    test_element (const char *name)
    {
      GstElement *element = 0;
      element = gst_element_factory_make (name, "test0");
      bool exists = GST_IS_ELEMENT (element);
      if (GST_IS_ELEMENT (element))
        {
          gst_object_unref (element);
          element = 0;
        }
      return exists;
    }

    struct Category
    {
        const char *icon_path;
        const char *name;
        int         page;
    };

    Category categories[] =
    {
        {BMP_IMAGE_DIR_PREFS G_DIR_SEPARATOR_S "playback.png",    N_("Audio"),            0},
        {BMP_IMAGE_DIR_PREFS G_DIR_SEPARATOR_S "library.png",     N_("Library"),          1},
        {BMP_IMAGE_DIR_PREFS G_DIR_SEPARATOR_S "misc.png",        N_("Miscellaneous"),    2},
#ifdef HAVE_OFA
        {BMP_IMAGE_DIR_PREFS G_DIR_SEPARATOR_S "musicbrainz.png", N_("MusicBrainz"),      3},
#endif
    };

    struct AudioSystem
    {
        const char                   *description;
        const char                   *sink_name;
        gint                          tab_id;
        Bmp::Preferences::AudioSink   sink_id;
    };

    AudioSystem audio_systems[] =
    {
#ifdef HAVE_ALSA
        {"ALSA (Advanced Linux Sound Architecture)",
         "alsasink",
         1,
         Preferences::SINK_ALSA},
#endif
        {"OSS (Open Sound System)",
         "osssink",
         2,
         Preferences::SINK_OSS},
#ifdef HAVE_SUN
        {"Sun/Solaris Audio",
         "sunaudiosink",
         3,
         Preferences::SINK_SUNAUDIO},
#endif
        {"ESD (Enlightenment Sound Daemon)",
         "esdsink",
         4,
         Preferences::SINK_ESD},
#ifdef HAVE_HAL
        {"HAL-Based Device Detection",
         "halaudiosink",
         5,
         Preferences::SINK_HAL},
#endif
    };
  } // <anonymous> namespace

  class Preferences::CategoryView
      : public Gtk::TreeView
  {
    public:

        CategoryView (BaseObjectType                        *cobject,
                      const Glib::RefPtr<Gnome::Glade::Xml> &xml);

        int
        get_selected_page ();

    private:

        // Category columns record
        class ColumnRecord
            : public Gtk::TreeModel::ColumnRecord
        {
        public:

            Gtk::TreeModelColumn<Glib::RefPtr<Gdk::Pixbuf> > icon;
            Gtk::TreeModelColumn<Glib::ustring>              name;
            Gtk::TreeModelColumn<int>                        page;

            ColumnRecord ()
            {
                add (icon);
                add (name);
                add (page);
            }
        };

        Glib::RefPtr<Gtk::ListStore> list_store;
        ColumnRecord                 columns;
  };

#ifdef HAVE_HAL
  // HAL Devices View 
  class Preferences::HALView
      : public Gtk::TreeView
  {
      public:

        HALView (BaseObjectType                        *cobject,
                 const Glib::RefPtr<Gnome::Glade::Xml> &xml);

      private:

        class ColumnRecord
          : public Gtk::TreeModel::ColumnRecord
        {
          public:

            Gtk::TreeModelColumn<LibHalDriveBus>      device_bus;
            Gtk::TreeModelColumn<Glib::ustring>       mount_path;
            Gtk::TreeModelColumn<Glib::ustring>       device_file;
            Gtk::TreeModelColumn<Glib::ustring>       volume_udi;
            Gtk::TreeModelColumn<bool>                is_disc;
            Gtk::TreeModelColumn<Library::HAL::Volume>  volume;

            ColumnRecord ()
            {
              add (device_bus);
              add (mount_path);
              add (device_file);
              add (volume_udi);
              add (is_disc);
              add (volume);
            }
        };

      Glib::RefPtr<Gtk::ListStore> list_store;
      ColumnRecord                 columns;

      enum VolumeType
      {
        VOLUME_DISC,
        VOLUME_1394,
        VOLUME_USB,
        VOLUME_HARDDISK,
    
        N_VOLUMES
      };

      Glib::RefPtr<Gdk::Pixbuf> volume_images[N_VOLUMES];

      void
      cell_data_func_device_icon (Gtk::CellRenderer* cell, Gtk::TreeModel::iterator const& iter);

      void
      hal_volume_added    (Library::HAL::Volume volume);

      void
      hal_volume_removed  (Library::HAL::Volume volume);

      int
      store_sort_func (Gtk::TreeModel::iterator const& a, Gtk::TreeModel::iterator const&b);

      typedef std::map < std::string, Gtk::TreeModel::iterator > VolumeUDIIterMap;
      VolumeUDIIterMap mapping;
  };
#endif

  //-- Preferences::HALView implementation -------------------------------------

#ifdef HAVE_HAL
  int
  Preferences::HALView::store_sort_func (Gtk::TreeModel::iterator const& a, Gtk::TreeModel::iterator const&b)
  {
    using namespace Glib;
    
    ustring device_a = (*a)[columns.mount_path];
    ustring device_b = (*b)[columns.mount_path];

    return device_a.compare (device_b); 
  }

  void
  Preferences::HALView::hal_volume_added    (Library::HAL::Volume volume)
  {
    VolumeUDIIterMap::iterator v_iter = mapping.find (volume.volume_udi);
    if (v_iter == mapping.end())
    {
        Gtk::TreeModel::iterator m_iter = list_store->append ();
        (*m_iter)[columns.device_bus]   = volume.device_bus; 
        (*m_iter)[columns.device_file]  = volume.device_file;
        (*m_iter)[columns.mount_path]   = volume.mount_path;
        (*m_iter)[columns.volume_udi]   = volume.volume_udi;
        (*m_iter)[columns.is_disc]      = volume.is_disc;
        (*m_iter)[columns.volume]       = volume; 
        mapping.insert (std::make_pair (volume.volume_udi, m_iter));
    }
  }

  void
  Preferences::HALView::hal_volume_removed  (Library::HAL::Volume volume)
  {
    VolumeUDIIterMap::iterator v_iter = mapping.find (volume.volume_udi);
    if (v_iter != mapping.end())
    {
      list_store->erase (v_iter->second);
      mapping.erase (volume.volume_udi);
    }
  }

  void
  Preferences::HALView::cell_data_func_device_icon (Gtk::CellRenderer *cell, Gtk::TreeModel::iterator const& m_iter)
  {
    Gtk::CellRendererPixbuf *cell_pixbuf = dynamic_cast<Gtk::CellRendererPixbuf *>(cell);
    LibHalDriveBus  device_bus  = (*m_iter)[columns.device_bus];
    bool            is_disc     = (*m_iter)[columns.is_disc];

    if (is_disc)
    {
      cell_pixbuf->property_pixbuf() = volume_images[VOLUME_DISC];
      return;
    }

    switch (device_bus)   
    {
        case LIBHAL_DRIVE_BUS_IDE:
        case LIBHAL_DRIVE_BUS_SCSI:
            cell_pixbuf->property_pixbuf() = volume_images[VOLUME_HARDDISK];
            break;

        case LIBHAL_DRIVE_BUS_USB:
            cell_pixbuf->property_pixbuf() = volume_images[VOLUME_USB];
            break;

        case LIBHAL_DRIVE_BUS_IEEE1394:
            cell_pixbuf->property_pixbuf() = volume_images[VOLUME_1394];
            break;

        default: break;
    }
  }

  Preferences::HALView::HALView (BaseObjectType                        *cobject,
                                 const Glib::RefPtr<Gnome::Glade::Xml> &xml)
      : Gtk::TreeView (cobject)
  {
      if (hal->is_initialized())
      {
        set_headers_clickable (false);
        set_headers_visible (true);

        Gtk::CellRendererPixbuf *cell_pixbuf = Gtk::manage (new Gtk::CellRendererPixbuf ());
        cell_pixbuf->property_ypad () = 4;
        cell_pixbuf->property_xpad () = 2;
        append_column ("", *cell_pixbuf);
        get_column (0)->set_resizable (false);
        get_column (0)->set_cell_data_func (*cell_pixbuf, sigc::mem_fun (this, &Bmp::Preferences::HALView::cell_data_func_device_icon));

        Gtk::CellRendererText *cell_text;
        cell_text = Gtk::manage (new Gtk::CellRendererText ());
        append_column ("Mount Path", *cell_text);
        get_column (1)->add_attribute (*cell_text, "text", 1);
        get_column (1)->set_resizable (false);
      
        cell_text = Gtk::manage (new Gtk::CellRendererText ());
        append_column ("Device", *cell_text);
        get_column (2)->add_attribute (*cell_text, "text", 2);
        get_column (2)->set_resizable (false);
      
        cell_text = Gtk::manage (new Gtk::CellRendererText ());
        append_column ("Volume UDI", *cell_text);
        get_column (3)->add_attribute (*cell_text, "text", 3);
        get_column (3)->set_resizable (false);

        const char* icon_names[]=
        {
          "dev-cdrom.png",
          "dev-volume-1394.png",
          "dev-volume-usb.png",
          "dev-volume-ide.png"
        };

        for (unsigned int n = 0; n < G_N_ELEMENTS (icon_names); ++n)
        {  
          volume_images[n] = Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_IMAGE_DIR_HAL, icon_names[n])); 
        }

        list_store = Gtk::ListStore::create (columns);
        list_store->set_default_sort_func (sigc::mem_fun (this, &Bmp::Preferences::HALView::store_sort_func));
        list_store->set_sort_column (Gtk::TreeSortable::DEFAULT_SORT_COLUMN_ID, Gtk::SORT_ASCENDING);
        Library::HAL::Volumes const& volumes = hal->get_volumes ();

        for (Library::HAL::Volumes::const_iterator v_iter = volumes.begin (),
                                                   v_end  = volumes.end (); v_iter != v_end; ++v_iter)
        {
          Library::HAL::Volume const& volume = v_iter->second;

          Gtk::TreeModel::iterator m_iter = list_store->append ();
          (*m_iter)[columns.device_bus]   = volume.device_bus; 
          (*m_iter)[columns.device_file]  = volume.device_file;
          (*m_iter)[columns.mount_path]   = volume.mount_path;
          (*m_iter)[columns.volume_udi]   = volume.volume_udi;
          (*m_iter)[columns.is_disc]      = volume.is_disc;
          (*m_iter)[columns.volume]       = volume; 
          mapping.insert (std::make_pair (volume.volume_udi, m_iter));
        }
        set_model (list_store);
        hal->signal_volume_added().connect (sigc::mem_fun (this, &Bmp::Preferences::HALView::hal_volume_added));
        hal->signal_volume_removed().connect (sigc::mem_fun (this, &Bmp::Preferences::HALView::hal_volume_removed));
      }
      else
      {
        set_headers_visible (false);
        set_sensitive (false);
      }
  }
#endif

  //-- Preferences::CategoryView implementation --------------------------------

  Preferences::CategoryView::CategoryView (BaseObjectType                        *cobject,
                                           const Glib::RefPtr<Gnome::Glade::Xml> &xml)
      : Gtk::TreeView (cobject)
  {
      set_headers_clickable (false);
      set_headers_visible (false);

      Gtk::CellRendererPixbuf *cell_pixbuf = Gtk::manage (new Gtk::CellRendererPixbuf ());
      cell_pixbuf->property_ypad () = 4;
      cell_pixbuf->property_xpad () = 2;

      append_column ("Icon", *cell_pixbuf);
      get_column (0)->add_attribute (*cell_pixbuf, "pixbuf", 0);
      get_column (0)->set_resizable (false);

      Gtk::CellRendererText *cell_text = Gtk::manage (new Gtk::CellRendererText ());

      append_column ("Description", *cell_text);
      get_column (1)->add_attribute (*cell_text, "text", 1);
      get_column (1)->set_resizable (false);
      cell_text->property_xalign () = 0.0;

      list_store = Gtk::ListStore::create (columns);

      for (unsigned int i = 0; i < G_N_ELEMENTS (categories); i++)
        {
          Gtk::TreeModel::iterator iter = list_store->append ();
          (*iter)[columns.icon] = Gdk::Pixbuf::create_from_file (categories[i].icon_path);
          (*iter)[columns.name] = _(categories[i].name);
          (*iter)[columns.page] = categories[i].page;
        }
      set_model (list_store);
  }

  int
  Preferences::CategoryView::get_selected_page ()
  {
      Gtk::TreeModel::iterator iter = get_selection ()->get_selected ();
      return (*iter)[columns.page];
  }

  //-- Preferences implementation ----------------------------------------------
  Preferences *
  Preferences::create ()
  {
      const std::string path = BMP_GLADE_DIR G_DIR_SEPARATOR_S "preferences-ui.glade";
      Glib::RefPtr<Gnome::Glade::Xml> glade_xml = Gnome::Glade::Xml::create (path);

      Preferences *preferences = 0;
      glade_xml->get_widget_derived ("preferences", preferences);
      return preferences;
  }

  // ctor
  Preferences::Preferences (BaseObjectType                        *cobject,
                            const Glib::RefPtr<Gnome::Glade::Xml> &xml)
      : Gtk::Window (cobject),
        m_ref_xml (xml)
  {
      // acquire consistently needed widgets
      m_ref_xml->get_widget ("category_notebook", notebook_categories);
      m_ref_xml->get_widget_derived ("category_view", category_view);

#ifdef HAVE_HAL
      m_ref_xml->get_widget_derived ("hal_view", hal_view);
#else
      m_ref_xml->get_widget ("scrolledwindow_hal")->hide ();
#endif

      m_ref_xml->get_widget ("l_version", l_version);
      m_ref_xml->get_widget ("close", close);

      Glib::RefPtr<Gtk::TreeSelection> selection = category_view->get_selection ();
      selection->signal_changed ().connect (sigc::mem_fun (this, &Preferences::on_category_changed));

      close->signal_clicked().connect (sigc::mem_fun (this, &Preferences::hide));

      // Version String 
      boost::format version_format ("<span size='small' color='#606060'>BMPx %s:%s-R%s (%s / %s)</span>");
      std::string version = (version_format % PACKAGE_VERSION % RV_LAST_CHANGED_DATE % RV_REVISION % BUILD_DATE % BUILD_BUILDUSER).str ();
      l_version->set_markup (version);

      // Misc Images 
      Gtk::Image *image;
      m_ref_xml->get_widget ("i-audio-caps", image);
      image->set (Gtk::StockID (BMP_STOCK_AUDIO), Gtk::ICON_SIZE_SMALL_TOOLBAR);
      m_ref_xml->get_widget ("i-audio-settings", image);
      image->set (Gtk::StockID (BMP_STOCK_AUDIO), Gtk::ICON_SIZE_SMALL_TOOLBAR);

      // Other Misc Widgets 
      m_ref_xml->get_widget ("cbox_audio_system", cbox_audio_system);

#ifdef HAVE_ALSA
      m_ref_xml->get_widget ("cbox_alsa_card", cbox_alsa_card);
      m_ref_xml->get_widget ("cbox_alsa_device", cbox_alsa_device);
      m_ref_xml->get_widget ("alsa_buffer_time", alsa_buffer_time);
#endif

      m_ref_xml->get_widget ("cbe_oss_device", cbe_oss_device);
      m_ref_xml->get_widget ("oss_buffer_time", oss_buffer_time);

      m_ref_xml->get_widget ("esd_host", esd_host);
      m_ref_xml->get_widget ("esd_buffer_time", esd_buffer_time);

#ifdef HAVE_SUNDAUDIO
      m_ref_xml->get_widget ("cbe_sun_device", cbe_sun_device);
      m_ref_xml->get_widget ("sun_buffer_time", sun_buffer_time);
#endif

      m_ref_xml->get_widget ("notebook_audio_system", notebook_audio_system);

      m_ref_xml->get_widget ("b_audio_system_apply", b_audio_system_apply);
      m_ref_xml->get_widget ("hbox_warning", hbox_warning);

#ifndef HAVE_HAL
      Gtk::ComboBoxEntry *cbe_cdrom;
      m_ref_xml->get_widget ("cbe_cdrom", cbe_cdrom);
      cbe_cdrom->show_all ();
      mcs_bind->bind_cbox_entry (cbe_cdrom, "audio", "cdrom-device");
      m_ref_xml->get_widget ("alignment189")->show_all ();
      m_ref_xml->get_widget ("alignment202")->show_all ();
#endif

      // Setup Audio 
      setup_audio ();

#ifdef HAVE_OFA
      // MusicBrainz 
      mcs_bind->bind_entry (dynamic_cast<Gtk::Entry *>(m_ref_xml->get_widget ("musicbrainz-username")) , "musicbrainz", "username");
      mcs_bind->bind_entry (dynamic_cast<Gtk::Entry *>(m_ref_xml->get_widget ("musicbrainz-password")) , "musicbrainz", "password");
#endif

      // Togglebuttons
      const char* toggle_buttons[] =
      {
          "ui-esc-trayconify",
          "display-tooltips",
          "display-notifications",
          "send-statistics",
          "physically-delete-files",
          "force-rgba-disable",
      };

      for (unsigned int n = 0; n < G_N_ELEMENTS (toggle_buttons); n++)
      {
          Gtk::ToggleButton *button = 0;
          m_ref_xml->get_widget (toggle_buttons[n], button);
          if (button) 
            mcs_bind->bind_toggle_button (button, "bmp", toggle_buttons[n]);
          else
            g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Widget '%s' not found in preferences glade xml UI", toggle_buttons[n]);
      }

      // Tooltips
      mcs->subscribe ("Preferences", "bmp", "display-tooltips", sigc::mem_fun (this, &Preferences::mcs_display_tooltips_changed));
  }

  void
  Preferences::on_category_changed ()
  {
      if (category_view->get_selection ()->count_selected_rows () == 0)
          return;

      notebook_categories->set_current_page (category_view->get_selected_page ());
  }

  void
  Preferences::mcs_display_tooltips_changed (MCS_CB_DEFAULT_SIGNATURE)
  {
  }

  // Audio preferences
  void
  Preferences::on_cbox_audio_system_changed     ()
  {
      Gtk::TreeModel::iterator iter = cbox_audio_system->get_active ();
      notebook_audio_system->set_current_page ((*iter)[audio_system_columns.tab_id]);
  }

  void
  Preferences::set_apply_sensitive ()
  {
      b_audio_system_apply->set_sensitive (true);
      hbox_warning->set_sensitive (true);
  }

#ifdef HAVE_ALSA
  void
  Preferences::on_cbox_alsa_card_changed ()
  {
      Gtk::TreeModel::iterator iter_model = cbox_alsa_card->get_active ();

      list_store_alsa_device = Gtk::ListStore::create (alsa_device_columns);
      cbox_alsa_device->set_model (list_store_alsa_device);

      if (cbox_alsa_card->get_active_row_number () == 0)
        {
          cbox_alsa_device->set_sensitive (false);
        }
      else
        {
          AlsaCard                   card = (*iter_model)[alsa_card_columns.card];
          AlsaDevices               &devices = card.devices;
          Gtk::TreeModel::iterator  iter_model_devices;

          for (AlsaDevices::iterator iter_devices = devices.begin (); iter_devices != devices.end (); iter_devices++ )
            {
              AlsaDevice &device = *iter_devices;
              iter_model_devices = list_store_alsa_device->append ();
              (*iter_model_devices)[alsa_device_columns.name]   = device.description;
              (*iter_model_devices)[alsa_device_columns.device] = device;
            }
          cbox_alsa_device->set_sensitive (true);
          cbox_alsa_device->set_active (0);
      }
  }
#endif

  void
  Preferences::on_b_audio_system_apply ()
  {
      b_audio_system_apply->set_sensitive (false);
      hbox_warning->set_sensitive (false);

      Gtk::TreeModel::iterator  iter      = cbox_audio_system->get_active ();
      AudioSink                 sink_id   = (*iter)[audio_system_columns.sink_id];
      std::string               sink_name = (*iter)[audio_system_columns.sink_name];
      std::string               device;

      play->property_status() = PLAYSTATUS_STOPPED;

      mcs->key_set<std::string>("audio", "sink", sink_name);

      switch (sink_id)
      {

#ifdef HAVE_ALSA

        case SINK_ALSA:
        {
          AlsaDevice alsa_device;
          int active;

          active = cbox_alsa_card->get_active_row_number ();

          if (active > 0)
            {
              Gtk::TreeModel::iterator iter = cbox_alsa_device->get_active ();
              alsa_device = (*iter)[alsa_device_columns.device];
              device = alsa_device.dev;

              mcs->key_set<std::string>("audio", "device-alsa", device);
            }
          else
            {
              mcs->key_set<std::string>("audio", "device-alsa", "default");
            }
          break;
        }
#endif

        case SINK_OSS:
        {
          device = cbe_oss_device->get_entry()->get_text();
          mcs->key_set<std::string>("audio", "device-oss", device);
          break;
        }

        case SINK_ESD:
        {
          device = esd_host->get_text();
          mcs->key_set<std::string>("audio", "device-esd", device);
          break;
        }

#ifdef HAVE_SUN

        case SINK_SUNAUDIO:
        {
          device = cbe_sun_device->get_entry()->get_text();
          mcs->key_set<std::string>("audio", "device-sun", device);
          break;
        }

#endif
        default: return;

      }
      play->reset();
  }

#ifdef HAVE_ALSA

  Bmp::Preferences::AlsaCards
  Preferences::get_alsa_cards ()
  {
      Bmp::Preferences::AlsaCards cards;
      int card = -1;
      int snd_err = snd_card_next (&card);

      if (snd_err)
        {
          g_critical ("%s: %s", G_STRLOC, snd_strerror (snd_err));
          return cards;
        }

      while (card > -1)
        {
          snd_pcm_info_t *pcm_info;
          snd_ctl_t  *ctl;
          int             pcm_device;

          Bmp::Preferences::AlsaCard        alsa_card;

          char *name;
          snd_card_get_name (card, &name);

          std::string dev = (boost::format ("hw:%d") % card).str ();

          snd_err = snd_ctl_open (&ctl, dev.c_str (), 0);
          if (snd_err)
          {
              g_critical ("%s: %s", G_STRLOC, snd_strerror (snd_err));
              break;
          }

          alsa_card.dev = dev;
          alsa_card.description = name;
          alsa_card.card_id = card;

          pcm_device = -1;

          while (42)
            {
              Bmp::Preferences::AlsaDevice alsa_device;

              if (!ctl) break;

              snd_err = snd_ctl_pcm_next_device (ctl, &pcm_device);
              if (snd_err)
                {
                  g_warning ("%s: %s", G_STRLOC, snd_strerror (snd_err));
                  break;
                }

              // FIXME: Do we really need this here still?
              if (pcm_device < 0)
                  break;

              snd_pcm_info_malloc (&pcm_info);
              snd_pcm_info_set_device (pcm_info, pcm_device);
              snd_pcm_info_set_subdevice (pcm_info, 0);
              snd_pcm_info_set_stream (pcm_info, SND_PCM_STREAM_PLAYBACK);

              snd_err = snd_ctl_pcm_info (ctl, pcm_info);

              if (snd_err)
                {
                  g_warning ("%s: %s", G_STRLOC, snd_strerror (snd_err));
                  break;
                }

              alsa_device.dev         = (boost::format ("hw:%d,%d") % card % pcm_device).str ();
              alsa_device.description = (boost::format ("%d: %s") % pcm_device % snd_pcm_info_get_name (pcm_info)).str ();
              alsa_device.card_id     = card;
              alsa_device.device_id   = pcm_device;

              alsa_card.devices.push_back (alsa_device);

              snd_pcm_info_free (pcm_info);
            }

          cards.push_back (alsa_card);

          if (ctl)
            {
              snd_ctl_close (ctl);
            }

          snd_card_next (&card);
          if (card < 0 ) break;
      }
      return cards;
  }
#endif

  void
  Preferences::setup_audio ()
  {
      Gtk::CellRendererText *cell_text;
      std::string            sink;
      int                    current = -1;

      list_store_audio_systems = Gtk::ListStore::create (audio_system_columns);

      sink = mcs->key_get<std::string>("audio", "sink");

      for (unsigned int n = 0; n < G_N_ELEMENTS (audio_systems); n++)
        {
          GstElement  *element;
          Gtk::TreeModel::iterator iter;
          element = gst_element_factory_make (audio_systems[n].sink_name, "sinktest0");
          if (!element) continue;
          gst_object_unref (element);
          available_sinks.insert (std::make_pair (audio_systems[n].sink_name, true));
          iter = list_store_audio_systems->append ();
          (*iter)[audio_system_columns.description] = audio_systems[n].description;
          (*iter)[audio_system_columns.sink_name] = audio_systems[n].sink_name;
          (*iter)[audio_system_columns.tab_id] = audio_systems[n].tab_id;
          (*iter)[audio_system_columns.sink_id] = audio_systems[n].sink_id;
          if (!sink. compare(audio_systems[n].sink_name))
            {
              current = n;
            }
        }

      cell_text = Gtk::manage (new Gtk::CellRendererText() );
      cbox_audio_system->clear ();
      cbox_audio_system->pack_start (*cell_text);
      cbox_audio_system->add_attribute (*cell_text, "text", 0);
      cbox_audio_system->set_model (list_store_audio_systems);
      cbox_audio_system->set_active (current);

      cbox_audio_system->signal_changed().connect (sigc::mem_fun (this, &Preferences::on_cbox_audio_system_changed));

      if (current == -1)
          notebook_audio_system->set_current_page (0);
      else
          notebook_audio_system->set_current_page (audio_systems[current].tab_id);

#ifdef HAVE_ALSA

      if (available_sinks.find ("alsasink") != available_sinks.end ())
        {
          list_store_alsa_cards = Gtk::ListStore::create (alsa_card_columns);
          Gtk::TreeModel::iterator iter_model;
          AlsaCards cards = Bmp::Preferences::get_alsa_cards ();

          iter_model = list_store_alsa_cards->append ();
          (*iter_model)[alsa_card_columns.name] = "System Default";

          for (AlsaCards::iterator iter = cards.begin (); iter != cards.end (); iter++)
            {
              AlsaCard &card = (*iter);

              iter_model = list_store_alsa_cards->append ();
              (*iter_model)[alsa_card_columns.name] = card.description;
              (*iter_model)[alsa_card_columns.card] = card;
            }

          cell_text = Gtk::manage (new Gtk::CellRendererText () );
          cbox_alsa_card->clear ();
          cbox_alsa_card->pack_start (*cell_text);
          cbox_alsa_card->add_attribute (*cell_text, "text", 0);
          cbox_alsa_card->set_model (list_store_alsa_cards);

          cbox_alsa_card->signal_changed ().connect (sigc::mem_fun (this, &Preferences::on_cbox_alsa_card_changed));

          cell_text = Gtk::manage (new Gtk::CellRendererText () );
          cbox_alsa_device->clear ();
          cbox_alsa_device->pack_start (*cell_text);
          cbox_alsa_device->add_attribute (*cell_text, "text", 0);

          // We have to parse the current device here in case in case sink is alsasink
          if (!sink.compare("alsasink"))
            {
              std::string device;
              device = mcs->key_get<std::string>("audio", "device-alsa");

              if (!device. compare("default"))
                {
                  cbox_alsa_card->set_active (0);
                }
              else
                {
                  using namespace boost::algorithm;
                  std::vector<std::string> subs;

                  split (subs, device, is_any_of (":,"));

                  if (subs.size() && subs[0].compare("hw"))
                    {
                      g_message ("Invalid identifier '%s", subs[0].c_str ());
                    }
                  else if (subs.size() && !subs[0].compare("hw"))
                    {
                      int card,
                      device;

                      card    = atoi (subs[1].c_str ()) + 1;
                      device  = atoi (subs[2].c_str ());

                      cbox_alsa_card->set_active (card);
                      cbox_alsa_device->set_active (device);
                    }
                }
            }

          mcs_bind->bind_spin_button (alsa_buffer_time, "audio", "alsa-buffer-time");

          alsa_buffer_time->signal_value_changed().connect(sigc::mem_fun (this, &Preferences::set_apply_sensitive));
          cbox_alsa_device->signal_changed().connect(sigc::mem_fun (this, &Preferences::set_apply_sensitive));
          cbox_alsa_card->signal_changed().connect(sigc::mem_fun (this, &Preferences::set_apply_sensitive));
      }

#endif

      if (available_sinks.find ("osssink") != available_sinks.end ())
        {
          mcs_bind->bind_cbox_entry (cbe_oss_device, "audio", "device-oss");
          mcs_bind->bind_spin_button (oss_buffer_time, "audio", "oss-buffer-time");

          if (!sink.compare("osssink"))
            {
              std::string device;
              device = mcs->key_get<std::string>("audio", "device-oss");
              cbe_oss_device->get_entry()->set_text (device );
            }

          cbe_oss_device->signal_changed().connect(sigc::mem_fun (this, &Preferences::set_apply_sensitive));
          oss_buffer_time->signal_value_changed().connect(sigc::mem_fun (this, &Preferences::set_apply_sensitive));
        }

#ifdef HAVE_SUN

      if (available_sinks.find ("sunaudiosink") != available_sinks.end ())
        {
          mcs_bind->bind_cbox_entry (cbe_sun_device, "audio", "device-sun");
          mcs_bind->bind_spin_button (sun_buffer_time, "audio", "sun-buffer-time");
          sun_buffer_time->signal_value_changed ().connect(sigc::mem_fun (this, &Preferences::set_apply_sensitive));
          cbe_sun_device->signal_changed ().connect(sigc::mem_fun (this, &Preferences::set_apply_sensitive));
        }

#endif

      if (available_sinks.find ("esdsink") != available_sinks.end ())
        {
          mcs_bind->bind_entry (esd_host, "audio", "device-esd");
          mcs_bind->bind_spin_button (esd_buffer_time, "audio", "esd-buffer-time");
          esd_buffer_time->signal_value_changed ().connect(sigc::mem_fun (this, &Preferences::set_apply_sensitive));
          esd_host->signal_changed ().connect(sigc::mem_fun (this, &Preferences::set_apply_sensitive));
        }

      cbox_audio_system->signal_changed().connect(sigc::mem_fun (this, &Preferences::set_apply_sensitive));
      b_audio_system_apply->signal_clicked().connect(sigc::mem_fun (this, &Preferences::on_b_audio_system_apply));

      // Setup support status indicators
      if (play->property_has_cdda())
        {
          m_ref_xml->get_widget ("expander-cdda")->show_all ();
        }

      dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget ("img_status_cdda"))->set
         (play->property_has_cdda ()     ?
          Gtk::StockID(BMP_STOCK_PLUGIN) :
          Gtk::StockID(BMP_STOCK_PLUGIN_DISABLED),
          Gtk::ICON_SIZE_SMALL_TOOLBAR);

      dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget ("img_status_http"))->set
         (play->property_has_http ()     ?
          Gtk::StockID(BMP_STOCK_PLUGIN) :
          Gtk::StockID(BMP_STOCK_PLUGIN_DISABLED),
          Gtk::ICON_SIZE_SMALL_TOOLBAR);

      dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget ("img_status_mms"))->set
         (play->property_has_mms ()      ?
          Gtk::StockID(BMP_STOCK_PLUGIN) :
          Gtk::StockID(BMP_STOCK_PLUGIN_DISABLED),
          Gtk::ICON_SIZE_SMALL_TOOLBAR);

      struct SupportCheckTuple
      {
        char *widget;
        char *element;
      };

      const SupportCheckTuple support_check[] = 
      {
        {"img_status_mpc",    "musepackdec"},
        {"img_status_flac",   "flacdec"},
        {"img_status_m4a",    "faad"},
        {"img_status_wma",    "ffdec_wmav2"},
        {"img_status_sid",    "siddec"},
        {"img_status_wav",    "wavparse"},
        {"img_status_mod",    "modplug"},
        {"img_status_spc",    "spcdec"},
      };

      // For MP3 we need both id3demux and mad
      dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget ("img_status_mp3"))->set
          ((test_element ("mad") && test_element ("id3demux"))
            ? Gtk::StockID(BMP_STOCK_PLUGIN) : Gtk::StockID(BMP_STOCK_PLUGIN_DISABLED), Gtk::ICON_SIZE_SMALL_TOOLBAR);
           
      dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget ("img_status_ogg"))->set
        ((test_element("oggdemux") && test_element("vorbisdec")) ? 
          Gtk::StockID(BMP_STOCK_PLUGIN) : 
          Gtk::StockID(BMP_STOCK_PLUGIN_DISABLED), Gtk::ICON_SIZE_SMALL_TOOLBAR);

      dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget ("img_status_vorbisenc"))->set
        ((test_element("oggmux") && test_element("vorbisenc") && play->property_has_cdda()) ? 
          Gtk::StockID(BMP_STOCK_PLUGIN) : 
          Gtk::StockID(BMP_STOCK_PLUGIN_DISABLED), Gtk::ICON_SIZE_SMALL_TOOLBAR);

      dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget ("img_status_mp3enc"))->set
        ((test_element("lame") && play->property_has_cdda()) ? 
          Gtk::StockID(BMP_STOCK_PLUGIN) : 
          Gtk::StockID(BMP_STOCK_PLUGIN_DISABLED), Gtk::ICON_SIZE_SMALL_TOOLBAR);

      dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget ("img_status_flacenc"))->set
        ((test_element("flacenc") && play->property_has_cdda()) ? 
          Gtk::StockID(BMP_STOCK_PLUGIN) : 
          Gtk::StockID(BMP_STOCK_PLUGIN_DISABLED), Gtk::ICON_SIZE_SMALL_TOOLBAR);

      for (unsigned int n = 0; n < G_N_ELEMENTS (support_check); ++n)
      {
          dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget (support_check[n].widget))->set
                (test_element (support_check[n].element) ?
                 Gtk::StockID(BMP_STOCK_PLUGIN) : 
                 Gtk::StockID(BMP_STOCK_PLUGIN_DISABLED), Gtk::ICON_SIZE_SMALL_TOOLBAR);
      }

      b_audio_system_apply->set_sensitive (false);
      hbox_warning->set_sensitive (false);
  }
} // namespace Bmp

