/*  BMP - The Dumb Music Player
 *  Copyright (C) 2003-2006 BMP Project
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */

#include <iostream>
#include <sstream>

#include <glibmm.h>
#include <glibmm/markup.h>
#include <glibmm/i18n.h>
#include <gtkmm.h>

#include <boost/format.hpp>

#include "xml.hh"
#include "util.hh"
#include "uri++.hh"
#include "stock.hh"
#include "paths.hh"
#include "main.hh"

#include "x_bookmarks.hh"
#include "x_vfs.hh"

#include "ui-toolbox.hh"
#include "streams-icecast.hh"

namespace
{
    static char* icecast_headers[] =
    {
      N_("Stream Name"),
      N_("Bitrate"),
      N_("Stream Genre"),
    };
}

namespace Bmp
{
  namespace Streams
  {
     void
      Icecast::refresh ()
      {
        using namespace Gtk;

        struct IcecastEntry
        {
            Glib::ustring   Name;
            unsigned int    Bitrate;
            Glib::ustring   Genre;
            Glib::ustring   URI;

            IcecastEntry () : Name (""), Bitrate (0), Genre (""), URI ("") {};
        };

        xmlDocPtr             doc;
        xmlXPathObjectPtr     xpathobj;
        xmlNodeSetPtr         nv;
        xmlNodePtr            node;

#define ICECAST_LIST_URI "http://dir.xiph.org/yp.xml"

#define IN_URI      "listen_url"
#define IN_NAME     "server_name"
#define IN_GENRE    "genre"
#define IN_BITRATE  "bitrate"

        Bmp::VFS::Handle handle (Glib::ustring(ICECAST_LIST_URI));
        try {
            vfs->read (handle, Bmp::VFS::TRANSPORT);
          }
        catch (Bmp::VFS::Exception& cxe) 
          {
            return;
          }
        doc = xmlParseMemory ((const char*)handle.get_buffer(), handle.get_buffer_size());
        if (!doc) return;

        xpathobj = xml_execute_xpath_expression (doc, BAD_CAST "//entry", NULL);
        nv = xpathobj->nodesetval;
        m_streams->clear ();
        for (int n = 0; n < nv->nodeNr; n++)
          {
            IcecastEntry entry;
            xmlNodePtr node_sibling;

            node = nv->nodeTab[n];
            node_sibling = node->children;

            while (node_sibling)
              {
                if (node_sibling->type != XML_TEXT_NODE)
                {
                    std::string name = reinterpret_cast<const char*>(node_sibling->name);

                    if (!name.compare(IN_NAME) )
                    {
                        if (node_sibling->children)
                        {
                            char *_data = (char*)XML_GET_CONTENT (node_sibling->children);
                            entry. Name = _data;
                            g_free (_data);
                        }
                    }
                    else if (!name.compare(IN_URI))
                    {
                        if (node_sibling->children)
                        {
                            char *_data = (char*)XML_GET_CONTENT (node_sibling->children);
                            entry. URI = _data;
                            g_free (_data);
                        }
                    }
                    else if (!name.compare(IN_GENRE))
                    {
                        if (node_sibling->children)
                        {
                            char *_data = (gchar*)XML_GET_CONTENT (node_sibling->children);
                            entry. Genre = _data;
                            g_free (_data);
                        }
                    }
                    else if (!name.compare(IN_BITRATE))
                    {
                        char *_data =  (char*)XML_GET_CONTENT (node_sibling->children);
                        entry. Bitrate = strtol (_data, NULL, 10);
                        g_free (_data);
                    }
                }

                node_sibling = node_sibling->next;
                while (gtk_events_pending()) gtk_main_iteration ();
            }

            TreeModel::iterator m_iter = m_streams->append ();
            (*m_iter)[columns.genre]       = entry.Genre;
            (*m_iter)[columns.bitrate]     = entry.Bitrate;
            (*m_iter)[columns.name]        = entry.Name;
            (*m_iter)[columns.uri]         = entry.URI;
          }

          xmlXPathFreeObject (xpathobj);
          //xmlFreeDoc (doc);
          while (gtk_events_pending ()) gtk_main_iteration();
      }

      bool
      Icecast::visible_func (Gtk::TreeModel::iterator const& m_iter)
      {
        Glib::ustring  name,
                       genre,
                       needle;

        if (!m_filter.length ()) return true;

        needle  = m_filter.casefold ();
        name = (*m_iter)[columns.name];
        genre = (*m_iter)[columns.genre];
        name.casefold();
        genre.casefold();

        return ((Util::match_keys (name, needle)) || (Util::match_keys (genre, needle)));
      }

      void
      Icecast::filter (Glib::ustring const& filter)
      {
        m_filter = filter;
        m_filtered->refilter ();
      }

      void
      Icecast::column_clicked (int column)
      {
        Gtk::SortType sort_type, sort_type_new;
        int           sort_id;

        m_streams->get_sort_column_id (sort_id, sort_type);

        if ((sort_id >= 0) && (sort_id != column))
          {
            get_column (sort_id)->set_sort_indicator (false);
          }

        if (sort_id >= 0)
          sort_type_new = (sort_type == Gtk::SORT_ASCENDING) ? Gtk::SORT_DESCENDING : Gtk::SORT_ASCENDING;
        else
          sort_type_new = Gtk::SORT_ASCENDING;

        m_streams->set_sort_column_id (column, sort_type_new);
        get_column (column)->set_sort_indicator (true);
        get_column (column)->set_sort_order (sort_type_new);
      }

      void 
      Icecast::get_stream (Glib::ustring& title, Glib::ustring& uri)
      {
        Gtk::TreeModel::iterator m_iter = get_selection()->get_selected ();
        uri = (*m_iter)[columns.uri];
        title = (*m_iter)[columns.name];
      }

      Icecast::Icecast (BaseObjectType                       * cobject,
                        Glib::RefPtr<Gnome::Glade::Xml> const& xml)
          : Gtk::TreeView (cobject),
            m_streams (Gtk::ListStore::create (columns)),
            m_filtered (Gtk::TreeModelFilter::create (m_streams))
      {
        m_filtered->set_visible_func (sigc::mem_fun (this, &Bmp::Streams::Icecast::visible_func));

        for (unsigned int n = 0; n < G_N_ELEMENTS (icecast_headers); n++)
        {
          Gtk::CellRendererText *cell;
          cell = Gtk::manage (new Gtk::CellRendererText());
          append_column (_(icecast_headers[n]), *cell);
          get_column (n)->add_attribute (*cell, "text", n);
          get_column (n)->set_resizable (true);
          get_column (n)->signal_clicked ().connect
                    (sigc::bind (sigc::mem_fun (this, &Bmp::Streams::Icecast::column_clicked),n));
        }
        get_selection()->set_mode (Gtk::SELECTION_SINGLE);
        set_headers_clickable (true);
        set_model (m_filtered);

        refresh ();
      }
  }
};
