//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#include <config.h>

#include <cstdlib>
#include <cstring>
#include <cctype>

#include <vector>

#include <gtkmm.h>
#include <glib/gstdio.h>
#include <glib/gi18n.h>

#include <boost/algorithm/string.hpp>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "util.hh"
#include "md5.h"

#include "x_vfs.hh"

namespace Bmp
{
  namespace Util
  {

    bool
    str_has_prefix_nocase (const gchar *str, const gchar *prefix)
    {
        if ((!str) || (!prefix)) return false;
        return (strncasecmp (str, prefix, std::strlen (prefix)) == 0);
    }

    bool
    str_has_suffix_nocase (const gchar *str, const gchar *_suffix)
    {
        gchar *suffix_fq;
        gboolean rv;

        if ((!str) || (!_suffix)) return false;

        suffix_fq = g_strconcat (".", _suffix, NULL);
        rv = (strcasecmp (str + std::strlen (str) - std::strlen (suffix_fq), suffix_fq) == 0);
        g_free (suffix_fq);
        return rv;
    }

    bool
    str_has_suffixes_nocase (const gchar * str,
                             char * const *suffixes)
    {
        char *const *suffix;

        g_return_val_if_fail (str != NULL, FALSE);
        g_return_val_if_fail (suffixes != NULL, FALSE);

        for (suffix = suffixes; *suffix; suffix++)
      if (str_has_suffix_nocase (str, *suffix))
          return true;

        return false;
    }

    bool
    match_keys (std::string haystack,
                std::string needle)
    {
        using namespace boost;
        using namespace std;

        typedef vector< string > Fragments;

        to_lower (haystack);
        to_lower (needle);

        Fragments keys;    
        split ( keys, needle, is_any_of(" ") );

        for (Fragments::const_iterator iter = keys.begin (), end = keys.end (); iter != end; ++iter)
        {
      if (!find_first (haystack, (*iter))) return false;
        }

        return true;
    }

    bool
    daemonize ()
    {
      pid_t child_pid;
      int dev_null_fd;

      if (g_chdir ("/") < 0)
        {
          g_message ("Could not change into root directory");
          return FALSE;
        }

      switch ((child_pid = fork ()))
        {
          case -1:
            g_message ("fork failed\n");
            return FALSE;
            break;

          case 0:
            /* silently ignore failures here, if someone
             * doesn't have /dev/null we may as well try
             * to continue anyhow
             */

            dev_null_fd = g_open ("/dev/null", O_RDWR);
            if (dev_null_fd >= 0)
              {
                dup2 (dev_null_fd, 0);
                dup2 (dev_null_fd, 1);
                dup2 (dev_null_fd, 2);
              }

            /* Get a predictable umask */
            umask (022);
            break;

          default:
            exit (0);
            break;
        }

      if (setsid () == -1)
        g_error ("setsid () failed");

      return true;
    }

    char*
    hexify (char *pass, int len)
    {
      char *hash;
      char *bp;
      char hexchars[] = "0123456789abcdef";
      int i;

      hash = static_cast<char *> (g_malloc0 (33));
      bp = hash;

      for(i = 0; i < len; i++)
        {
          *(bp++) = hexchars[(pass[i] >> 4) & 0x0f];
          *(bp++) = hexchars[pass[i] & 0x0f];
        }
      *bp = 0;
      return hash;
    }

    std::string
    md5_hex (char* const data, size_t data_size)
    {
        md5_state_t  md5state;
        char         md5pword[16];

        md5_init (&md5state);
        md5_append (&md5state, (unsigned const char *)data, data_size); 
        md5_finish (&md5state, (unsigned char *)md5pword);
        char *md5_hex = Util::hexify (md5pword, sizeof(md5pword));
        std::string md5_hex_std = md5_hex;
        free (md5_hex);
        return md5_hex_std;
    }

    char*
    get_timestr (time_t t, int gmt)
    {
      struct tm *tm;
      static char buf[30];

      tm = gmt ? gmtime(&t) : localtime(&t);
      snprintf(buf, sizeof(buf), "%d-%.2d-%.2d %.2d:%.2d:%.2d",
          tm->tm_year + 1900,
          tm->tm_mon + 1,
          tm->tm_mday,
          tm->tm_hour,
          tm->tm_min,
          tm->tm_sec);
      return buf;
    }

    std::string 
    get_time ()
    {
      struct tm *tm;
      static char buf[30];

      time_t t = time (NULL);
      tm = gmtime (&t);
      snprintf(buf, sizeof(buf), "%d-%.2d-%.2dT%.2d:%.2d:%.2d",
          tm->tm_year + 1900,
          tm->tm_mon + 1,
          tm->tm_mday,
          tm->tm_hour,
          tm->tm_min,
          tm->tm_sec);
      return std::string(buf);
    }

    Glib::RefPtr<Gdk::Pixbuf> 
    get_image_from_uri (Glib::ustring const& uri)
    {
      Glib::RefPtr<Gdk::Pixbuf> image;
      VFS::Handle handle (uri);
      vfs->read (handle, VFS::TRANSPORT);
          
      if (!handle.get_buffer()) return image;
  
      std::string format;

      if ((str_has_suffix_nocase (uri.c_str(), "jpeg")) || (str_has_suffix_nocase (uri.c_str(), "jpg")))
          format = "jpeg";
      else if (str_has_suffix_nocase (uri.c_str(), "gif"))
          format = "gif";
      else if (str_has_suffix_nocase (uri.c_str(), "png"))
          format = "png";

      try
        {
          Glib::RefPtr<Gdk::PixbufLoader> loader = Gdk::PixbufLoader::create (format);
          loader->write (reinterpret_cast<const guint8*>(handle.get_buffer()), handle.get_buffer_size());
          loader->close ();
          image = loader->get_pixbuf();
        }
      catch (...) {}
        return image;
    }
  }
}

